/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2009 The LS-Universe Team

    Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "Taskbar.h"
#include "TaskbarButton.h"

#include "GhostButton.h"
#include "GhostTaskbar.h"

#include "DropTarget.h"
#include "WindowUtilities.h"

#include "shlwapi.h"
#include <algorithm>

#define GET_X_LPARAM(lp) ((int)(short)LOWORD(lp))
#define GET_Y_LPARAM(lp) ((int)(short)HIWORD(lp))

extern int bitsperpixel;

extern ProcessStatsRequestFunc m_pProcessStatsRequest;

//Alpha Transparency
typedef BOOL (WINAPI *PULW)(HWND hwnd, HDC hdcDst, POINT *pptDst, SIZE *psize, HDC hdcSrc, POINT *pptSrc, COLORREF crKey, BLENDFUNCTION *pblend, DWORD dwFlags);
static PULW pUpdateLayeredWindow = NULL;

FUNC_BOOL__DWORD pAllowSetForegroundWindow = NULL;

#define TIMER_UPDATE        1
#define TIMER_MOUSETRACK    2
#define TIMER_ALPHAFADEIN   3
#define TIMER_ALPHAFADEOUT  4
#define TIMER_DROPTIMEOUT   5
#define TIMER_ACTIVETASK    6
#define TIMER_SCROLL        7
#define TIMER_FLASH         8
#define TIMER_SPECIALUPDATE 9
#define TIMER_ALPHAFADEUPBANG 13
#define TIMER_ALPHAFADEDOWNBANG 14

#define LM_LSSELECT         9185

UINT lsMessages[] = {
                      LM_LSSELECT,
                      LM_REDRAW,
                      LM_WINDOWACTIVATED,
                      LM_WINDOWCREATED,
                      LM_WINDOWDESTROYED,
                      LM_REFRESH,
                      0
                     };

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Taskbar::Taskbar(const string &name) :
    pDropTarget(NULL),
        
    hWnd(NULL),
    hWndOwner(NULL),
    hInstance(NULL),
    box(NULL),
        
    actualx(0),
    actualy(0),
    actualwidth(0),
    actualheight(0),

    bTrueTransparency(false),

    bStartHidden(false),

    bAlwaysOnTop(false),

    bNormalWindowzOrder(false),

    strBehindWindow(""),
    hWndBehindWindow(NULL),

    bFocusOnEnter(false),

    bHideIfEmpty(false),

    bActionsOnDownButton(false),

    iDropTimeOut(0),
    iTimerRefresh(1000),
    iTrackingRefresh(250),
    iFlashRefresh(500),
    iSpecialUpdate(5000),

    iDisplayButtonType(-1),

    iSpecialHoverState(-1),
    
    bNoMinimizeOnClick(false),

    bUseBigNormalIcon(false),
    bUseBigActiveIcon(false),
    bUseBigMinimizedIcon(false),
    bUseBigFlashingIcon(false),
    bUseBigHoverIcon(false),
    bUseBigPressedIcon(false),
    bUseBigGroupedIcon(false),

    bMultiLineUsed(false),
    bUseCustomizedCaption(false),
    strNormalText(""),
    strActiveText(""),
    strMinimizedText(""),
    strFlashingText(""),
    strHoverText(""),
    strPressedText(""),
    strGroupedText(""),
    
    bShowIcon(true),

    bShowText(true),

    iAutoSort(0),
    
    iThisMultiMonitorOnly(-1),

    bUseGroupedFontSettings(false),
    bUseHoverFontSettings(false),
    bUsePressedFontSettings(false),

    bUseGroupedIconSettings(false),
    bUseHoverIconSettings(false),
    bUsePressedIconSettings(false),

    bUseGrouping(false),
    bUseHoverState(false),
    bUsePressedState(false),
    bUseFlashingState(true),

    bScroll(false),
    iScrollPadLength(10),
    iScrollRefresh(100),
    iScrollSpeed(2),

    bVWMCurrentDesktopOnly(false),
    iVWMSpecifiedDesktopOnly(1),

    backupAddList(""),
    backupRemoveList(""),

    iLines_RowsAdded(0),
    iWrapCountAdded(0),

    iStartWithPosition(0),

    hWndLastActiveForSwitch(NULL),
    hWndLastActiveForGrouping(NULL),

    bVisible(false),

    bMinimizeAtAllCosts(false),

    bCreateLayoutLocked(false),

    bMouseInside(false),
    
    bSkipButton(false),
    
    bIsScrolling(false),

    bInCriticalProcess(false),

    iBackupExportedButtonCount(0),

    bAutoHidden(false),

    bParentDestroyNotificationReceived(false),

    bHasOverlayConditional(false),

    oldbackupBMP(NULL),
    backupDC(NULL),
    backupBitmap(NULL),

    pTaskbarTexture(NULL),

    pNormalButtonTexture(NULL),
    pActiveButtonTexture(NULL),
    pMinimizedButtonTexture(NULL),
    pFlashingButtonTexture(NULL),
    pHoverButtonTexture(NULL),
    pPressedButtonTexture(NULL),
    pGroupedButtonTexture(NULL),

    pNormalIcon(NULL),
    pActiveIcon(NULL),
    pMinimizedIcon(NULL),
    pFlashingIcon(NULL),
    pHoverIcon(NULL),
    pPressedIcon(NULL),
    pGroupedIcon(NULL),
    
    pNormalText(NULL),
    pActiveText(NULL),
    pMinimizedText(NULL),
    pFlashingText(NULL),
    pHoverText(NULL),
    pPressedText(NULL),
    pGroupedText(NULL),

    iLeftBorder(0),
    iTopBorder(0),
    iRightBorder(0),
    iBottomBorder(0),

    iLayout(0),
    iDirection(1),
    iWrapDirection(1),
    iWrapCount(-1),
    bUseAdaptiveSize(false),

    iLines_Rows(1),
    iMaxLines_Rows(100),

    iMaxButtonCount(100),

    iXSpacing(0),
    iYSpacing(0),

    iMaxTaskWidth(iScreenWidth),
    iMaxTaskHeight(iScreenHeight),

    bAutoSize(false),
    iMaxTaskbarSize(-1),
    iAutoSizeSteps(-1),
    iAutoSizeDelay(-1),

    bAlphaMap(false),
    bAlphaFade(false),
    iAlphaTransparency(255),

    iUpdateAlpha(0),

    iAlphaFadeStep(2),
    iAlphaFadeDelay(10),

    iAlphaFadeBangEndValue(255),

    bAlphaInitialized(false),
    bLayeredFlag(false),

    pClickCapture(NULL),

    pDropCapture(NULL),

    pMoveCapture(NULL),

    pCapture(NULL),
    
    pActiveTask(NULL),

    pDontPaintAddedTask(NULL),

    iLButtonAction(1),
    iRButtonAction(2),
    iMButtonAction(0),

    pTooltip(NULL),
    hWndToolTip(NULL),

    bShowToolTip(true),

    iTooltipMode(0)
{
    strName = name;

    displayedtaskbuttons.clear();
    taskbuttons.clear();

    listGhostTaskbar.clear();
    listGhostButton.clear();

    AddList.clear();
    RemoveList.clear();
    CustomIconList.clear();
    ChangingIconList.clear();

    AddMatchingList.clear();
    RemoveMatchingList.clear();
    
    pAllowSetForegroundWindow = (FUNC_BOOL__DWORD)GetProcAddress( GetModuleHandle("USER32"),"AllowSetForegroundWindow");
}

Taskbar::~Taskbar()
{
    AddList.clear();
    RemoveList.clear();
    CustomIconList.clear();
    ChangingIconList.clear();

    AddMatchingList.clear();
    RemoveMatchingList.clear();

    if (pDropTarget)
    {
        pDropTarget->Revoke( hWnd );
        delete pDropTarget;
        pDropTarget = NULL;
    }

    if (pTooltip)
    {
        Destroy_xTooltipClass(pTooltip);
        pTooltip = NULL;
    }

    displayedtaskbuttons.clear();

    // release buttons
    for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
    {   
        delete (*tbit);
        (*tbit) = NULL;
    }
    taskbuttons.clear();

    // unregister messages
    SendMessage( GetLitestepWnd(), LM_UNREGISTERMESSAGE, (WPARAM) hWnd, (LPARAM) ::lsMessages );

    if(IsWindow(hWnd))
    {   
        // shutdown timer
        KillTimer(hWnd, TIMER_UPDATE);
        KillTimer(hWnd, TIMER_MOUSETRACK);
        KillTimer(hWnd, TIMER_ALPHAFADEIN);
        KillTimer(hWnd, TIMER_ALPHAFADEOUT);
        KillTimer(hWnd, TIMER_DROPTIMEOUT);
        KillTimer(hWnd, TIMER_ACTIVETASK);
        KillTimer(hWnd, TIMER_SCROLL);
        KillTimer(hWnd, TIMER_FLASH);
        KillTimer(hWnd, TIMER_SPECIALUPDATE);
        KillTimer(hWnd, TIMER_ALPHAFADEUPBANG);
        KillTimer(hWnd, TIMER_ALPHAFADEDOWNBANG);

        DestroyWindow(hWnd);
    }

    DestroyWindow(hWndOwner);

    if ( backupBitmap )
    {
        backupBitmap = (HBITMAP) SelectObject(backupDC, oldbackupBMP);
        DeleteObject(backupBitmap);
    }
    if ( backupDC )
        DeleteDC(backupDC);
    if ( oldbackupBMP )
        DeleteObject(oldbackupBMP);

    backupDC = NULL;
    backupBitmap = NULL;
    oldbackupBMP = NULL;

    for(GhostTaskbarListIterator gtlit = listGhostTaskbar.begin(); gtlit != listGhostTaskbar.end(); gtlit++)
    {
        delete (*gtlit);
        (*gtlit) = NULL;
    }

    listGhostTaskbar.clear();

    for(GhostButtonListIterator glit = listGhostButton.begin(); glit != listGhostButton.end(); glit++)
    {
        delete (*glit);
        (*glit) = NULL;
    }

    listGhostButton.clear();
    
    Destroy_xPaintClassSupport();
}

void Taskbar::Load(HINSTANCE hInstance, HWND setbox)
{
    this->hInstance = hInstance;
    
    if (IsWindow(setbox))
    {
        box = setbox;
    }
    else
    {
        box = NULL;
    }

    if (IsWindow(hWnd)) // window already exists...
        return;

    hWndOwner = CreateWindowEx(WS_EX_TOOLWINDOW,
                                "xTaskbarOwner",
                                strName.c_str(),
                                0,
                                0, 0, 0, 0, 
                                HWND_MESSAGE,
                                0,
                                hInstance,
                                0);

    hWnd = CreateWindowEx( (box ? 0 : WS_EX_TOOLWINDOW),
                            "xTaskbar",
                            strName.c_str(),
                            (box ? WS_CHILD : WS_POPUP),
                            actualx, actualy,       
                            actualwidth, actualheight,
                            (box ? box : hWndOwner), 
                            0,
                            hInstance,
                            this);
        
    if (hWnd)
    {
        PrefixedLSSetVariable(strName, "CurrentHWND", IntToString( (int) hWnd ));

        SendMessage( GetLitestepWnd(), LM_REGISTERMESSAGE, (WPARAM) hWnd, (LPARAM) ::lsMessages );
        
        SetWindowLong(hWnd, GWL_USERDATA, MAGIC_DWORD);

        Reconfigure();

        int nButtonsAddedAtStartup = CheckDisplayedTasks();

        for(int nI = 1; nI <= nButtonsAddedAtStartup; nI++)
        {
            PrefixedLSSetVariable(strName, "CurrentButtonCount", IntToString( nI ));
            if( !taskbarAdd.empty() )
                LSExecute(hWnd, taskbarAdd.c_str(), SW_SHOWNORMAL);

            if ( iWrapCount != -1 )
            {
                int realwrapCount = iWrapCount+iWrapCountAdded;
                int availableButtonCount = nI;

                if ( availableButtonCount/realwrapCount >= iLines_Rows+iLines_RowsAdded && availableButtonCount%realwrapCount == 1 )
                {
                    iLines_RowsAdded++;
                    if ( iLines_Rows+iLines_RowsAdded <= iMaxLines_Rows )
                    {   
                        PrefixedLSSetVariable(strName, "CurrentLines", IntToString( iLines_Rows+iLines_RowsAdded ));
                        if ( !taskbarLineAdd.empty() )
                            LSExecute(hWnd, taskbarLineAdd.c_str(), SW_SHOWNORMAL);
                    }
                    else if ( iLines_Rows+iLines_RowsAdded > iMaxLines_Rows)
                    {
                        iLines_RowsAdded--;
                        iWrapCountAdded++;
                    }
                }
            }
        }
    
        if ( bAutoSize )
            AutoSizeFunction(nButtonsAddedAtStartup);

        iBackupExportedButtonCount = nButtonsAddedAtStartup;

        PrefixedLSSetVariable(strName, "CurrentButtonCount", IntToString( iBackupExportedButtonCount ));
        PrefixedLSSetVariable(strName, "CurrentLines", IntToString( iLines_Rows+iLines_RowsAdded ));

        bCreateLayoutLocked = false;

        CreateLayout();

        //we become a drop target
        pDropTarget = new DropTarget;
        pDropTarget->Register( hWnd );
        pDropTarget->RegisterDrop( hWnd, MOUSEISDRAGGING ); 
        
        if ( !bHideIfEmpty )
        {
            if ( bStartHidden )
                Hide();
            else
                Show();
        }
        else
        {
            Show();
            CreateLayout();
        }   
        
        if ( iBackupExportedButtonCount != 0 && !taskbarFilled.empty() )    
            LSExecute(hWnd, taskbarFilled.c_str(), SW_SHOWNORMAL);
        else if ( iBackupExportedButtonCount == 0 && !taskbarEmpty.empty() )    
            LSExecute(hWnd, taskbarEmpty.c_str(), SW_SHOWNORMAL);

        SetTimer(hWnd, TIMER_UPDATE, iTimerRefresh, NULL);
        if ( bVWMCurrentDesktopOnly || (iRealMonitors > 1 && iThisMultiMonitorOnly != -1) )
            SetTimer(hWnd, TIMER_SPECIALUPDATE, iSpecialUpdate, NULL);

        SetTimer(hWnd, TIMER_ACTIVETASK, 200, NULL);
    }
    else
    {
        DestroyWindow(hWndOwner);

        delete this;
    }
}

void Taskbar::Create_xPaintClassSupport()
{
    bShowIcon = Getx2RCBool(strName.c_str(), "ShowIcon", true);

    bShowText = Getx2RCBool(strName.c_str(), "ShowText", true);

    bUseHoverState = Getx2RCBool(strName.c_str(), "ShowHoverState");
    bUsePressedState = Getx2RCBool(strName.c_str(), "ShowPressedState");
    bUseGrouping = Getx2RCBool(strName.c_str(), "UseGrouping");
    bUseFlashingState = Getx2RCBool(strName.c_str(), "UseFlashing", true);

    bUseHoverIconSettings = Getx2RCBool(strName.c_str(), "UseHoverIconSettings");
    bUsePressedIconSettings = Getx2RCBool(strName.c_str(), "UsePressedIconSettings");
    bUseGroupedIconSettings = Getx2RCBool(strName.c_str(), "UseGroupedIconSettings");

    bUseHoverFontSettings = Getx2RCBool(strName.c_str(), "UseHoverFontSettings");
    bUsePressedFontSettings = Getx2RCBool(strName.c_str(), "UsePressedFontSettings");
    bUseGroupedFontSettings = Getx2RCBool(strName.c_str(), "UseGroupedFontSettings");

    if (!pTaskbarTexture)
        pTaskbarTexture = Create_xTextureClass();

    if (!pNormalButtonTexture)
        pNormalButtonTexture = Create_xTextureClass();
    if (!pActiveButtonTexture)
        pActiveButtonTexture = Create_xTextureClass();
    if (!pMinimizedButtonTexture)
        pMinimizedButtonTexture = Create_xTextureClass();

    if (bUseFlashingState && !pFlashingButtonTexture)
        pFlashingButtonTexture = Create_xTextureClass();

    if (bUseHoverState && !pHoverButtonTexture)
        pHoverButtonTexture = Create_xTextureClass();
    if (bUsePressedState && !pPressedButtonTexture)
        pPressedButtonTexture = Create_xTextureClass();
    if (bUseGrouping && !pGroupedButtonTexture)
        pGroupedButtonTexture = Create_xTextureClass();

    if (!pNormalIcon)
        pNormalIcon = Create_xIconClass();
    if (!pActiveIcon)
        pActiveIcon = Create_xIconClass();
    if (!pMinimizedIcon)
        pMinimizedIcon = Create_xIconClass();

    if (bUseFlashingState && !pFlashingIcon)
        pFlashingIcon = Create_xIconClass();

    if (bUseHoverState && bUseHoverIconSettings && !pHoverIcon)
        pHoverIcon = Create_xIconClass();
    if (bUsePressedState && bUsePressedIconSettings && !pPressedIcon)
        pPressedIcon = Create_xIconClass();
    if (bUseGrouping && bUseGroupedIconSettings && !pGroupedIcon)
        pGroupedIcon = Create_xIconClass();

    if (!pNormalText)
        pNormalText = Create_xTextClass();
    if (!pActiveText)
        pActiveText = Create_xTextClass();
    if (!pMinimizedText)
        pMinimizedText = Create_xTextClass();

    if (bUseFlashingState && !pFlashingText)
        pFlashingText = Create_xTextClass();

    if (bUseHoverState && bUseHoverFontSettings && !pHoverText)
        pHoverText = Create_xTextClass();
    if (bUsePressedState && bUsePressedFontSettings && !pPressedText)
        pPressedText = Create_xTextClass();
    if (bUseGrouping && bUseGroupedFontSettings && !pGroupedText)
        pGroupedText = Create_xTextClass();

    if ( IsOS(OS_2KXP) && Getx2RCBool(strName.c_str(), "AlphaMap") )
    {
        pTaskbarTexture->configure(strName.c_str(), "", true);

        PaintTexture *defaultbuttontexture = Create_xTextureClass();
        defaultbuttontexture->configure(strName.c_str(), "Button", true);
        
        pNormalButtonTexture->configure(strName.c_str(), "NormalButton", true, defaultbuttontexture);
        pActiveButtonTexture->configure(strName.c_str(), "ActiveButton", true, pNormalButtonTexture);
        pMinimizedButtonTexture->configure(strName.c_str(), "MinimizedButton", true, pNormalButtonTexture);
        
        if (pFlashingButtonTexture)
            pFlashingButtonTexture->configure(strName.c_str(), "FlashingButton", true, pNormalButtonTexture);
        
        if (pHoverButtonTexture)
            pHoverButtonTexture->configure(strName.c_str(), "HoverButton", true, pNormalButtonTexture);
        if (pPressedButtonTexture)
            pPressedButtonTexture->configure(strName.c_str(), "PressedButton", true, pNormalButtonTexture);
        if (pGroupedButtonTexture)
            pGroupedButtonTexture->configure(strName.c_str(), "GroupedButton", true, pNormalButtonTexture);
    
        PaintIcon *defaulticon = Create_xIconClass();
        defaulticon->configure(strName.c_str(), "", true, true);

        pNormalIcon->configure(strName.c_str(), "Normal", true, true, defaulticon/*DefaultClass*/);
        pActiveIcon->configure(strName.c_str(), "Active", true, true, pNormalIcon/*DefaultClass*/);
        pMinimizedIcon->configure(strName.c_str(), "Minimized", true, true, pNormalIcon/*DefaultClass*/);
        
        if (pFlashingIcon)
            pFlashingIcon->configure(strName.c_str(), "Flashing", true, true, pNormalIcon/*DefaultClass*/);
        
        if (pHoverIcon)
            pHoverIcon->configure(strName.c_str(), "Hover", true, true, pNormalIcon/*DefaultClass*/);
        if (pPressedIcon)
            pPressedIcon->configure(strName.c_str(), "Pressed", true, true, pNormalIcon/*DefaultClass*/);
        if (pGroupedIcon)
            pGroupedIcon->configure(strName.c_str(), "Grouped", true, true, pNormalIcon/*DefaultClass*/);
        
        PaintText *defaulttext = Create_xTextClass();
        defaulttext->configure(strName.c_str(), "", true);

        pNormalText->configure(strName.c_str(), "Normal", true, defaulttext/*DefaultClass*/);
        pActiveText->configure(strName.c_str(), "Active", true, pNormalText/*DefaultClass*/);
        pMinimizedText->configure(strName.c_str(), "Minimized", true, pNormalText/*DefaultClass*/);
        
        if (pFlashingText)
            pFlashingText->configure(strName.c_str(), "Flashing", true, pNormalText/*DefaultClass*/);
        
        if (pHoverText)
            pHoverText->configure(strName.c_str(), "Hover", true, pNormalText/*DefaultClass*/);
        if (pPressedText)
            pPressedText->configure(strName.c_str(), "Pressed", true, pNormalText/*DefaultClass*/);
        if (pGroupedText)
            pGroupedText->configure(strName.c_str(), "Grouped", true, pNormalText/*DefaultClass*/);

        //CleanUp DefaultButton/Icon/Text
        Destroy_xTextureClass(defaultbuttontexture);
        Destroy_xIconClass(defaulticon);
        Destroy_xTextClass(defaulttext);
    }
    else
    {
        pTaskbarTexture->configure(strName.c_str(), "", false);

        PaintTexture *defaultbuttontexture = Create_xTextureClass();
        defaultbuttontexture->configure(strName.c_str(), "Button", false);
        
        pNormalButtonTexture->configure(strName.c_str(), "NormalButton", false, defaultbuttontexture);
        pActiveButtonTexture->configure(strName.c_str(), "ActiveButton", false, pNormalButtonTexture);
        pMinimizedButtonTexture->configure(strName.c_str(), "MinimizedButton", false, pNormalButtonTexture);
        
        if (pFlashingButtonTexture)
            pFlashingButtonTexture->configure(strName.c_str(), "FlashingButton", false, pNormalButtonTexture);
        
        if (pHoverButtonTexture)
            pHoverButtonTexture->configure(strName.c_str(), "HoverButton", false, pNormalButtonTexture);
        if (pPressedButtonTexture)
            pPressedButtonTexture->configure(strName.c_str(), "PressedButton", false, pNormalButtonTexture);
        if (pGroupedButtonTexture)
            pGroupedButtonTexture->configure(strName.c_str(), "GroupedButton", false, pNormalButtonTexture);
    
        PaintIcon *defaulticon = Create_xIconClass();
        defaulticon->configure(strName.c_str(), "", true, false);

        pNormalIcon->configure(strName.c_str(), "Normal", true, false, defaulticon/*DefaultClass*/);
        pActiveIcon->configure(strName.c_str(), "Active", true, false, pNormalIcon/*DefaultClass*/);
        pMinimizedIcon->configure(strName.c_str(), "Minimized", true, false, pNormalIcon/*DefaultClass*/);
        
        if (pFlashingIcon)
            pFlashingIcon->configure(strName.c_str(), "Flashing", true, false, pNormalIcon/*DefaultClass*/);
        
        if (pHoverIcon)
            pHoverIcon->configure(strName.c_str(), "Hover", true, false, pNormalIcon/*DefaultClass*/);
        if (pPressedIcon)
            pPressedIcon->configure(strName.c_str(), "Pressed", true, false, pNormalIcon/*DefaultClass*/);
        if (pGroupedIcon)
            pGroupedIcon->configure(strName.c_str(), "Grouped", true, false, pNormalIcon/*DefaultClass*/);
    
        PaintText *defaulttext = Create_xTextClass();
        defaulttext->configure(strName.c_str(), "", false);

        pNormalText->configure(strName.c_str(), "Normal", false, defaulttext/*DefaultClass*/);
        pActiveText->configure(strName.c_str(), "Active", false, pNormalText/*DefaultClass*/);
        pMinimizedText->configure(strName.c_str(), "Minimized", false, pNormalText/*DefaultClass*/);
        
        if (pFlashingText)
            pFlashingText->configure(strName.c_str(), "Flashing", false, pNormalText/*DefaultClass*/);
        
        if (pHoverText)
            pHoverText->configure(strName.c_str(), "Hover", false, pNormalText/*DefaultClass*/);
        if (pPressedText)
            pPressedText->configure(strName.c_str(), "Pressed", false, pNormalText/*DefaultClass*/);
        if (pGroupedText)
            pGroupedText->configure(strName.c_str(), "Grouped", false, pNormalText/*DefaultClass*/);

        //CleanUp DefaultButton/Icon/Text
        Destroy_xTextureClass(defaultbuttontexture);
        Destroy_xIconClass(defaulticon);
        Destroy_xTextClass(defaulttext);
    }
}

void Taskbar::Destroy_xPaintClassSupport()
{
    if (pTaskbarTexture)
        Destroy_xTextureClass(pTaskbarTexture);

    if (pNormalButtonTexture)
        Destroy_xTextureClass(pNormalButtonTexture);
    if (pActiveButtonTexture)
        Destroy_xTextureClass(pActiveButtonTexture);
    if (pMinimizedButtonTexture)
        Destroy_xTextureClass(pMinimizedButtonTexture);
    if (pFlashingButtonTexture)
        Destroy_xTextureClass(pFlashingButtonTexture);
    if (pHoverButtonTexture)
        Destroy_xTextureClass(pHoverButtonTexture);
    if (pPressedButtonTexture)
        Destroy_xTextureClass(pPressedButtonTexture);
    if (pGroupedButtonTexture)
        Destroy_xTextureClass(pGroupedButtonTexture);

    if (pNormalIcon)
        Destroy_xIconClass(pNormalIcon);
    if (pActiveIcon)
        Destroy_xIconClass(pActiveIcon);
    if (pMinimizedIcon)
        Destroy_xIconClass(pMinimizedIcon);
    if (pFlashingIcon)
        Destroy_xIconClass(pFlashingIcon);
    if (pHoverIcon)
        Destroy_xIconClass(pHoverIcon);
    if (pPressedIcon)
        Destroy_xIconClass(pPressedIcon);
    if (pGroupedIcon)
        Destroy_xIconClass(pGroupedIcon);

    if (pNormalText)
        Destroy_xTextClass(pNormalText);
    if (pActiveText)
        Destroy_xTextClass(pActiveText);
    if (pMinimizedText)
        Destroy_xTextClass(pMinimizedText);
    if (pFlashingText)
        Destroy_xTextClass(pFlashingText);
    if (pHoverText)
        Destroy_xTextClass(pHoverText);
    if (pPressedText)
        Destroy_xTextClass(pPressedText);
    if (pGroupedText)
        Destroy_xTextClass(pGroupedText);

    pTaskbarTexture = NULL;

    pNormalButtonTexture = NULL;
    pActiveButtonTexture = NULL;
    pMinimizedButtonTexture = NULL;
    pFlashingButtonTexture = NULL;
    pHoverButtonTexture = NULL;
    pPressedButtonTexture = NULL;
    pGroupedButtonTexture = NULL;

    pNormalIcon = NULL;
    pActiveIcon = NULL;
    pMinimizedIcon = NULL;
    pFlashingIcon = NULL;
    pHoverIcon = NULL;
    pPressedIcon = NULL;
    pGroupedIcon = NULL;
    
    pNormalText = NULL;
    pActiveText = NULL;
    pMinimizedText = NULL;
    pFlashingText = NULL;
    pHoverText = NULL;
    pPressedText = NULL;
    pGroupedText = NULL;
}


void Taskbar::Reconfigure(bool refresh)
{
    //Initialize Basic Helper Strings
    int numTokens = 0;
    string one = "";
    string two = "";
    string three = "";
    string four = "";
    string five = "";
    string six = "";
    string seven = "";

    SetWindowRgn(hWnd, NULL, true);

    const NameValuePair buttonActionValues[] = {
            { ".none", 0 },
            { ".select", 1 },
            { ".menu", 2 },
            { ".close", 3 },
            { ".ontop", 4 },
            { "none", 0 },
            { "select", 1 },
            { "menu", 2 },
            { "close", 3 },
            { "ontop", 4 },
            { 0, 0 }
            };

    const NameValuePair directionValues[] = {
            { ".right", 1 },
            { ".left", 2 },
            { ".bottom", 3 },
            { ".top", 4 },
            { "right", 1 },
            { "left", 2 },
            { "bottom", 3 },
            { "top", 4 },
            { 0, 0 }
            };

    const NameValuePair wrapdirectionValues[] = {
            { ".bottom", 1 },
            { ".top", 2 },
            { ".right", 3 },
            { ".left", 4 },
            { "bottom", 1 },
            { "top", 2 },
            { "right", 3 },
            { "left", 4 },
            { 0, 0 }
            };

    const NameValuePair displayValues[] = {
            { "all", -1 },
            { "normal+active+minimized", -1 },
            { "normal", 0 },
            { "active", 1 },
            { "minimized", 2 },
            { "active+normal", 3 },
            { "normal+active", 3 },
            { "minimized+normal", 4 },
            { "normal+minimized", 4 },
            { "active+minimized", 5 },
            { "minimized+active", 5 },
            { "flashing", 6},
            { 0, 0 }
            };

    const NameValuePair layoutValues[] = {
            { ".horizontal", 0 },
            { ".vertical", 1 },
            { "horizontal", 0 },
            { "vertical", 1 },
            { 0, 0 }
            };


    const NameValuePair autoSortValues[] = {
            { ".none", 0 },
            { ".filename", 1 },
            { ".classname", 2 },
            { ".windowname", 3 },
            { ".zorder", -1 },
            { "none", 0 },
            { "filename", 1 },
            { "classname", 2 },
            { "windowname", 3 },
            { "zorder", -1 },
            { 0, 0 }
            };

    const NameValuePair tooltipModeValues[] = {
            { ".prefix-caption-file", 0 },
            { ".caption-file", 1 },
            { ".prefix-caption", 2 },
            { ".caption", 3 },
            { ".prefix-file", 4 },
            { ".file", 5 },
            { "prefix-caption-file", 0 },
            { "caption-file", 1 },
            { "prefix-caption", 2 },
            { "caption", 3 },
            { "prefix-file", 4 },
            { "file", 5 },
            { 0, 0 }
            };

    if ( refresh )
    {
        bCreateLayoutLocked = true;

        // Remove All Visible on Refresh
        if ( !taskbarRemove.empty() )
        {
            int iButtonCount = iBackupExportedButtonCount;

            if ( iButtonCount > 0 )
            {
                for (int i = iButtonCount-1; i >= 0; i--)
                {
                    PrefixedLSSetVariable(strName, "CurrentButtonCount", IntToString( i ));
                    LSExecute(hWnd, taskbarRemove.c_str(), SW_SHOWNORMAL);
                }
            }
        }

        //Remove Added Lines
        if ( !taskbarLineRemove.empty() )
        {
            while ( iLines_RowsAdded > 0 )
            {
                PrefixedLSSetVariable(strName, "CurrentLines", IntToString( iLines_Rows+iLines_RowsAdded-1 ));
                LSExecute(hWnd, taskbarLineRemove.c_str(), SW_SHOWNORMAL);
                iLines_RowsAdded--;
            }
        }

        Destroy_xPaintClassSupport();
    }
    
    iLines_RowsAdded = 0;
    iWrapCountAdded = 0;

    displayedtaskbuttons.clear();

    vector<HWND> backuptaskbuttons;

    if ( refresh )
    {
        for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
        {
            backuptaskbuttons.insert(backuptaskbuttons.end(), (*tbit)->hWndButtonTask);
        }
    }

    // Release buttons
    for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
    {   
        delete (*tbit);
        (*tbit) = NULL;
    }
    taskbuttons.clear();
    
    AddList.clear();
    RemoveList.clear();
    CustomIconList.clear();
    ChangingIconList.clear();

    AddMatchingList.clear();
    RemoveMatchingList.clear();

    //Create/Refresh all Painting Stuff
    Create_xPaintClassSupport();

    iSpecialHoverState = Getx2RCNamedValue(strName.c_str(), "HoverStateDisplay", displayValues, -1);

    GetxRCList("*xTaskbarRemove", RemoveList, true);
    GetxRCList("*xTaskbarAdd", AddList, true);
    GetxRCList("*xTaskbarIcon", CustomIconList, false);
    GetxRCList("*xTaskbarUpdateIcon", ChangingIconList, true);

    iMaxButtonCount = Getx2RCInt(strName.c_str(), "ButtonMaxCount", 100, 1);

    backupAddList = iGetx2RCString(strName, "AddOnlyMatching", "");
    backupRemoveList = iGetx2RCString(strName, "RemoveOnlyMatching", "");
    AddMatchingList = split(backupAddList , ";");
    RemoveMatchingList = split(backupRemoveList , ";");

    bHideIfEmpty = Getx2RCBool(strName.c_str(), "HideIfEmpty");

    iAutoSort = Getx2RCNamedValue(strName.c_str(), "Sort", autoSortValues, 0);

    if ( !refresh )
        iDisplayButtonType = Getx2RCNamedValue(strName.c_str(), "Display", displayValues, -1);

    bActionsOnDownButton = Getx2RCBool(strName.c_str(), "ActionsOnButtonDown");

    iLButtonAction = Getx2RCNamedValue(strName.c_str(), "OnLeftClickButton", buttonActionValues, 1);
    iRButtonAction = Getx2RCNamedValue(strName.c_str(), "OnRightClickButton", buttonActionValues, 2);
    iMButtonAction = Getx2RCNamedValue(strName.c_str(), "OnMiddleClickButton", buttonActionValues, 0);

    int iCrucialIconSize = Getx2RCInt(strName.c_str(), "LargeIconChangeAt", 32, 17);

    bUseBigNormalIcon = (pNormalIcon->getIconSize() >= iCrucialIconSize ? true : false);
    bUseBigActiveIcon = (pActiveIcon->getIconSize() >= iCrucialIconSize ? true : false);
    bUseBigMinimizedIcon = (pMinimizedIcon->getIconSize() >= iCrucialIconSize ? true : false);
    
    if (pFlashingIcon)
        bUseBigFlashingIcon = (pFlashingIcon->getIconSize() >= iCrucialIconSize ? true : false);
    else
        bUseBigFlashingIcon = false;

    if (pHoverIcon)
        bUseBigHoverIcon = (pHoverIcon->getIconSize() >= iCrucialIconSize ? true : false);
    else
        bUseBigHoverIcon = false;
    if (pPressedIcon)
        bUseBigPressedIcon = (pPressedIcon->getIconSize() >= iCrucialIconSize ? true : false);
    else
        bUseBigPressedIcon = false;
    if (pGroupedIcon)
        bUseBigGroupedIcon = (pGroupedIcon->getIconSize() >= iCrucialIconSize ? true : false);
    else
        bUseBigGroupedIcon = false;

    bNoMinimizeOnClick = Getx2RCBool(strName.c_str(), "NoMinimizeOnClick");

    bVWMCurrentDesktopOnly = Getx2RCBool(strName.c_str(), "VWMCurrentDesktopOnly");
    iVWMSpecifiedDesktopOnly = Getx2RCInt(strName.c_str(), "VWMSpecifiedDesktopOnly", -1, -1, 10);
    if ( iVWMSpecifiedDesktopOnly != -1 && !bVWMCurrentDesktopOnly )
        bVWMCurrentDesktopOnly = true;

    if ( iRealMonitors > 1 )
        iThisMultiMonitorOnly = Getx2RCInt(strName.c_str(), "SpecifiedMultiMonitorOnly", -1, -1, 3);
    else
        iThisMultiMonitorOnly = -1;

    if ( Getx2RCBool(strName.c_str(), "NormalButtonText") || Getx2RCBool(strName.c_str(), "ActiveButtonText") || Getx2RCBool(strName.c_str(), "MinimizedButtonText") )
    {   
        bUseCustomizedCaption = true;

        strNormalText = iGetx2RCString(strName, "NormalButtonText", "%caption%");
        strActiveText = iGetx2RCString(strName, "ActiveButtonText", "%caption%");
        strMinimizedText = iGetx2RCString(strName, "MinimizedButtonText", "%caption%");
    }
    else
    {
        bUseCustomizedCaption = false;
    }

    strFlashingText = iGetx2RCString(strName, "FlashingButtonText", "");
    strHoverText = iGetx2RCString(strName, "HoverButtonText", "");
    strPressedText = iGetx2RCString(strName, "PressedButtonText", "");
    strGroupedText = iGetx2RCString(strName, "GroupedButtonText", "");

    //Convert for Multiline support
    CString strReplace;

    if ( bUseCustomizedCaption )
    {
        strReplace = strNormalText.c_str(); 
        if ( !bMultiLineUsed && strReplace.Find("<br>") != -1 )
            bMultiLineUsed = true;
        strReplace.Replace("<br>", "\n");
        strNormalText = strReplace;

        strReplace = strActiveText.c_str();
        if ( !bMultiLineUsed && strReplace.Find("<br>") != -1 )
            bMultiLineUsed = true;
        strReplace.Replace("<br>", "\n");
        strActiveText = strReplace;

        strReplace = strMinimizedText.c_str();
        if ( !bMultiLineUsed && strReplace.Find("<br>") != -1 )
            bMultiLineUsed = true;
        strReplace.Replace("<br>", "\n");
        strMinimizedText = strReplace;
    }
    
    strReplace = strFlashingText.c_str();
    if ( !bMultiLineUsed && strReplace.Find("<br>") != -1 )
        bMultiLineUsed = true;
    strReplace.Replace("<br>", "\n");
    strFlashingText = strReplace;

    strReplace = strHoverText.c_str();
    if ( !bMultiLineUsed && strReplace.Find("<br>") != -1 )
        bMultiLineUsed = true;
    strReplace.Replace("<br>", "\n");
    strHoverText = strReplace;

    strReplace = strPressedText.c_str();
    if ( !bMultiLineUsed && strReplace.Find("<br>") != -1 )
        bMultiLineUsed = true;
    strReplace.Replace("<br>", "\n");
    strPressedText = strReplace;

    strReplace = strGroupedText.c_str();
    if ( !bMultiLineUsed && strReplace.Find("<br>") != -1 )
        bMultiLineUsed = true;
    strReplace.Replace("<br>", "\n");
    strGroupedText = strReplace;
    //------------------------------------------------------------------------------------------

    //Events/Actions
    taskbarLButtonUp = iGetx2RCLine(strName, "OnLeftClickUp", "");
    taskbarLButtonDown =iGetx2RCLine(strName, "OnLeftClickDown", "");
    taskbarMButtonUp = iGetx2RCLine(strName, "OnMiddleClickUp", "");
    taskbarMButtonDown = iGetx2RCLine(strName, "OnMiddleClickDown", "");
    taskbarRButtonUp = iGetx2RCLine(strName, "OnRightClickUp", "");
    taskbarRButtonDown = iGetx2RCLine(strName, "OnRightClickDown", "");
    taskbarMWheelDown = iGetx2RCLine(strName, "OnWheelDown", "");
    taskbarMWheelUp = iGetx2RCLine(strName, "OnWheelUp", "");
    taskbarEnter = iGetx2RCLine(strName, "OnEnter", "");
    taskbarLeave = iGetx2RCLine(strName, "OnLeave", "");
    taskbarMove = iGetx2RCLine(strName, "OnMove", "");
    taskbarResize = iGetx2RCLine(strName, "OnResize", "");
    taskbarAdd = iGetx2RCLine(strName, "OnAdd", "");
    taskbarRemove = iGetx2RCLine(strName, "OnRemove", "");
    taskbarSelect = iGetx2RCLine(strName, "OnSelect", "");

    taskbarFlash = iGetx2RCLine(strName, "OnFlash", "");

    taskbarLineAdd = iGetx2RCLine(strName, "OnLineAdd", "");
    taskbarLineRemove = iGetx2RCLine(strName, "OnLineRemove", "");

    taskbarEmpty = iGetx2RCLine(strName, "BecomesEmpty", "");
    taskbarFilled = iGetx2RCLine(strName, "BecomesFilled", "");

    bShowToolTip = Getx2RCBool(strName.c_str(), "ShowTooltip", true);
    iTooltipMode = Getx2RCNamedValue(strName.c_str(), "TooltipMode", tooltipModeValues, 0);

    bScroll = Getx2RCBool(strName.c_str(), "FontScroll");
    iScrollRefresh = Getx2RCInt(strName.c_str(), "FontScrollInterval", 100, 50);
    iScrollSpeed = Getx2RCInt(strName.c_str(), "FontScrollSpeed", 2, 1);
    iScrollPadLength = Getx2RCInt(strName.c_str(), "FontScrollPad", 10, 0);

    bStartHidden = Getx2RCBool(strName.c_str(), "StartHidden");
    
    bFocusOnEnter = Getx2RCBool(strName.c_str(), "FocusOnEnter");

    string borders = iGetx2RCLine(strName, "Borders", "");
    
    if ( !borders.empty() )
    {
        numTokens = SplitString(borders, one, two, three, four, five, six, seven);
    
        if(numTokens == 1)
        {
            iLeftBorder = atoi(one.c_str());
            iRightBorder = iLeftBorder;
            iTopBorder = iLeftBorder;
            iBottomBorder = iLeftBorder;
        }
        else if (numTokens == 2)
        {
            iLeftBorder = atoi(one.c_str());
            iRightBorder = iLeftBorder;
            iTopBorder = atoi(two.c_str());
            iBottomBorder = iTopBorder;
        }
        else if (numTokens == 4)
        {
            iLeftBorder = atoi(one.c_str());
            iRightBorder = atoi(two.c_str());
            iTopBorder = atoi(three.c_str());
            iBottomBorder = atoi(four.c_str());
        }

        if ( iLeftBorder < 0 )
            iLeftBorder = 0;
        if ( iRightBorder < 0 )
            iRightBorder = 0;
        if ( iTopBorder < 0 )
            iTopBorder = 0;
        if ( iBottomBorder < 0 )
            iBottomBorder = 0;
    }
    else
    {
        iLeftBorder = Getx2RCInt(strName.c_str(), "LeftBorder", 0, 0);
        iTopBorder = Getx2RCInt(strName.c_str(), "TopBorder", 0, 0);
        iRightBorder = Getx2RCInt(strName.c_str(), "RightBorder", 0, 0);
        iBottomBorder = Getx2RCInt(strName.c_str(), "BottomBorder", 0, 0);
    }

    iLines_Rows = Getx2RCInt(strName.c_str(), "Lines", 1, 1);
    iMaxLines_Rows = Getx2RCInt(strName.c_str(), "MaxLines", 100, 1);
    if (iLines_Rows > iMaxLines_Rows)
        iLines_Rows = iMaxLines_Rows;

    iMaxTaskWidth = Getx2RCInt(strName.c_str(), "ButtonMaxWidth", iScreenWidth, 0);
    iMaxTaskHeight = Getx2RCInt(strName.c_str(), "ButtonMaxHeight", iScreenHeight, 0);

    iWrapCount = Getx2RCInt(strName.c_str(), "WrapCount", -1);
    if ( iWrapCount < -1 || iWrapCount == 0 )
        iWrapCount = -1;
    if ( iWrapCount == -1 && iLines_Rows > 1 )
        iWrapCount = 5;

    bUseAdaptiveSize = Getx2RCBool(strName.c_str(), "ButtonAdaptiveSize");

    iLayout = Getx2RCNamedValue(strName.c_str(), "Layout", layoutValues, 0);

    iDirection = Getx2RCNamedValue(strName.c_str(), "ButtonDirection", directionValues, 1);
    iWrapDirection = Getx2RCNamedValue(strName.c_str(), "ButtonWrapDirection", wrapdirectionValues, 1);

    if ( iLayout == 0 )
    {
        if ( iDirection == 3 )
            iDirection = 1;
        else if ( iDirection == 4 )
            iDirection = 2;
        if ( iWrapDirection == 3 )
            iWrapDirection = 1;
        else if ( iWrapDirection == 4 )
            iWrapDirection = 2;
    }
    else if ( iLayout == 1 )
    {
        if ( iDirection == 1 )
            iDirection = 3;
        else if ( iDirection == 2 )
            iDirection = 4;
        if ( iWrapDirection == 1 )
            iWrapDirection = 3;
        else if ( iWrapDirection == 2 )
            iWrapDirection = 4;
    }

    iXSpacing = Getx2RCInt(strName.c_str(), "ButtonxSpacing", 0);
    iYSpacing = Getx2RCInt(strName.c_str(), "ButtonySpacing", 0);

    //AutoSizeMode Initialisation
    string automode = iGetx2RCLine(strName, "AutoSize", "");
    if ( !automode.empty() )
    {
        numTokens = SplitString(automode, one, two, three, four, five, six, seven);
        
        if ( numTokens >= 1 )
        {
            if ( _stricmp(one.c_str(), "true") == NULL || _stricmp(one.c_str(), "on") == NULL )
            {
                MessageBox(NULL, "You have to specify a MaxWidth for xTaskbar after xTaskbarAutoSize!\nAutoSize is DISABLED!", "Themer Information", MB_OK | MB_SETFOREGROUND | MB_ICONEXCLAMATION);
                bAutoSize = false;
            }
            else
            {
                bAutoSize = true;
                iMaxTaskbarSize = atoi(one.c_str());

                if ( iMaxTaskbarSize <= 0 )
                {
                    MessageBox(NULL, "You have to specify a realistic MaxWidth for xTaskbar after xTaskbarAutoSize!\nAutoSize is DISABLED!", "Themer Information", MB_OK | MB_SETFOREGROUND | MB_ICONEXCLAMATION);
                    bAutoSize = false;
                }
                else
                {
                    if ( numTokens == 3)
                    {
                        iAutoSizeSteps = atoi(two.c_str());
                        iAutoSizeDelay = atoi(three.c_str());
                    }
                    
                    if ( iMaxTaskWidth == iScreenWidth || iMaxTaskHeight == iScreenHeight)
                        MessageBox(NULL, "You have to specify a xTaskbarButtonMaxWidth and xTaskbarButtonMaxHeight, when you want to use xTaskbarAutoSize!", "Themer Information", MB_OK | MB_SETFOREGROUND | MB_ICONEXCLAMATION);
                
                    iLines_Rows = 1;

                    iWrapCount = Getx2RCInt(strName.c_str(), "WrapCount", -1);
                    if ( iWrapCount < -1 || iWrapCount == 0 )
                        iWrapCount = -1;
                }
            }
        }
    }
    else
    {
        if ( Getx2RCBool(strName.c_str(), "AutoSize") )
            MessageBox(NULL, "You have to specify a MaxWidth for xTaskbar after xTaskbarAutoSize!\nAutoSize is DISABLED!", "Themer Information", MB_OK | MB_SETFOREGROUND | MB_ICONEXCLAMATION);

        bAutoSize = false;
    }

    bTrueTransparency = Getx2RCBool(strName.c_str(), "TrueTransparency");
    
    //Extra GhostImage Painting Background
    //Erase Existing on Refresh
    if ( refresh )
    {
        for(GhostTaskbarListIterator gtlit = listGhostTaskbar.begin(); gtlit != listGhostTaskbar.end(); gtlit++)
        {
            delete (*gtlit);
            (*gtlit) = NULL;
        }

        listGhostTaskbar.clear();

        for(GhostButtonListIterator glit = listGhostButton.begin(); glit != listGhostButton.end(); glit++)
        {
            delete (*glit);
            (*glit) = NULL;
        }

        listGhostButton.clear();
    }

    StringList extraPainting;
    extraPainting.clear();
    //Create Extra Painting
    extraPainting = GetxRCLineList( ("*" + strName + "OverlayTaskbar") );
    
    StringListIterator sit;
    for(sit = extraPainting.begin(); sit != extraPainting.end(); sit++)
    {
        numTokens = SplitString((*sit), one, two, three, four, five, six, seven);
        
        if ( numTokens >= 5 )
        {
            GhostTaskbar* newtaskbar = new GhostTaskbar();
        
            newtaskbar->strOrgX = one;
            newtaskbar->strOrgY = two;

            newtaskbar->strOrgWidth = three;
            newtaskbar->strOrgHeight = four;

            if ( IsOS( OS_2KXP ) && (strchr(six.c_str(), 'a') != 0 || strchr(six.c_str(), 'A') != 0) )
            {
                if ( strchr(six.c_str(), 't') != 0 || strchr(six.c_str(), 'T') != 0 )
                    newtaskbar->bOnTop = true;

                if ( !seven.empty() )
                {
                    newtaskbar->strConditional = seven;
                    bHasOverlayConditional = true;
                }
                
                newtaskbar->pTaskbarTexture->configure(five.c_str(), "", TRUE);
            }
            else 
            {
                if ( strchr(six.c_str(), 't') != 0 || strchr(six.c_str(), 'T') != 0 )
                    newtaskbar->bOnTop = true;

                if ( !seven.empty() )
                {
                    newtaskbar->strConditional = seven;
                    bHasOverlayConditional = true;
                }

                newtaskbar->pTaskbarTexture->configure(five.c_str(), "", FALSE);
            }
    
            if ( newtaskbar->pTaskbarTexture->getPaintingMode() != -1 )
                listGhostTaskbar.insert(listGhostTaskbar.end(), newtaskbar);
            else
                delete newtaskbar;
        }
    }

    extraPainting.clear();
    //Create Extra Painting
    extraPainting = GetxRCLineList( ("*" + strName + "OverlayButton") );
    
    for(sit = extraPainting.begin(); sit != extraPainting.end(); sit++)
    {
        //Use Split Helper Strings
        numTokens = SplitString((*sit), one, two, three, four, five, six, seven);
        
        if ( numTokens >= 5 )
        {
            GhostButton* newbutton = new GhostButton();
        
            newbutton->strOrgX = one;
            newbutton->strOrgY = two;

            newbutton->strOrgWidth = three;
            newbutton->strOrgHeight = four;
            
            PaintTexture *defaultbuttontexture = Create_xTextureClass();

            bool useNormal = false;
            bool useActive = false;
            bool useMinimized = false;
            bool useFlashing = false;
            bool useHover = false;
            bool usePressed = false;
            bool useGrouped = false;
        
            if ( IsOS( OS_2KXP ) && (strchr(six.c_str(), 'a') != 0 || strchr(six.c_str(), 'A') != 0) )
            {
                defaultbuttontexture->configure(five.c_str(), "Button", TRUE);
                
                if ( strchr(six.c_str(), 't') != 0 || strchr(six.c_str(), 'T') != 0 )
                    newbutton->bOnTop = true;

                if ( !seven.empty() )
                {
                    newbutton->strConditional = seven;
                    bHasOverlayConditional = true;
                }

                useNormal = newbutton->pNormalButtonTexture->configure(five.c_str(), "NormalButton", TRUE, defaultbuttontexture);
                useActive = newbutton->pActiveButtonTexture->configure(five.c_str(), "ActiveButton", TRUE, newbutton->pNormalButtonTexture);
                useMinimized = newbutton->pMinimizedButtonTexture->configure(five.c_str(), "MinimizedButton", TRUE, newbutton->pNormalButtonTexture);
                if ( bUseFlashingState )
                    useFlashing = newbutton->pFlashingButtonTexture->configure(five.c_str(), "FlashingButton", TRUE, newbutton->pNormalButtonTexture);
                if ( bUseHoverState )
                    useHover = newbutton->pHoverButtonTexture->configure(five.c_str(), "HoverButton", TRUE, newbutton->pNormalButtonTexture);
                if ( bUsePressedState )
                    usePressed = newbutton->pPressedButtonTexture->configure(five.c_str(), "PressedButton", TRUE, newbutton->pNormalButtonTexture);
                if ( bUseGrouping )
                    useGrouped = newbutton->pGroupedButtonTexture->configure(five.c_str(), "GroupedButton", TRUE, newbutton->pNormalButtonTexture);
            }
            else
            {
                defaultbuttontexture->configure(five.c_str(), "Button", FALSE);
                
                if ( strchr(six.c_str(), 't') != 0 || strchr(six.c_str(), 'T') != 0 )
                    newbutton->bOnTop = true;

                if ( !seven.empty() )
                {
                    newbutton->strConditional = seven;
                    bHasOverlayConditional = true;
                }

                useNormal = newbutton->pNormalButtonTexture->configure(five.c_str(), "NormalButton", FALSE, defaultbuttontexture);
                useActive = newbutton->pActiveButtonTexture->configure(five.c_str(), "ActiveButton", FALSE, newbutton->pNormalButtonTexture);
                useMinimized = newbutton->pMinimizedButtonTexture->configure(five.c_str(), "MinimizedButton", FALSE, newbutton->pNormalButtonTexture);
                if ( bUseFlashingState )
                    useFlashing = newbutton->pFlashingButtonTexture->configure(five.c_str(), "FlashingButton", FALSE, newbutton->pNormalButtonTexture);
                if ( bUseHoverState )
                    useHover = newbutton->pHoverButtonTexture->configure(five.c_str(), "HoverButton", FALSE, newbutton->pNormalButtonTexture);
                if ( bUsePressedState )
                    usePressed = newbutton->pPressedButtonTexture->configure(five.c_str(), "PressedButton", FALSE, newbutton->pNormalButtonTexture);
                if ( bUseGrouping )
                    useGrouped = newbutton->pGroupedButtonTexture->configure(five.c_str(), "GroupedButton", FALSE, newbutton->pNormalButtonTexture);
            }

            Destroy_xTextureClass(defaultbuttontexture);

            if (!useNormal)
            {
                Destroy_xTextureClass(newbutton->pNormalButtonTexture);
                newbutton->pNormalButtonTexture = NULL;
            }
            if (!useActive)
            {
                Destroy_xTextureClass(newbutton->pActiveButtonTexture);
                newbutton->pActiveButtonTexture = NULL;
            }
            if (!useMinimized)
            {
                Destroy_xTextureClass(newbutton->pMinimizedButtonTexture);
                newbutton->pMinimizedButtonTexture = NULL;
            }
            if (!useFlashing && bUseFlashingState)
            {
                Destroy_xTextureClass(newbutton->pFlashingButtonTexture);
                newbutton->pFlashingButtonTexture = NULL;
            }
            if (!useHover && bUseHoverState)
            {
                Destroy_xTextureClass(newbutton->pHoverButtonTexture);
                newbutton->pHoverButtonTexture = NULL;
            }
            else
            {
                newbutton->iSpecialHoverState = Getx2RCNamedValue(five.c_str(), "HoverStateDisplay", displayValues, -1);
            }
            if (!usePressed && bUsePressedState)
            {
                Destroy_xTextureClass(newbutton->pPressedButtonTexture);
                newbutton->pPressedButtonTexture = NULL;
            }
            if (!useGrouped && bUseGrouping)
            {
                Destroy_xTextureClass(newbutton->pGroupedButtonTexture);
                newbutton->pGroupedButtonTexture = NULL;
            }

            if ( useNormal || useActive || useMinimized || useFlashing || useHover || usePressed || useGrouped )
                listGhostButton.insert(listGhostButton.end(), newbutton);
            else
                delete newbutton;
        }
    }

    iDropTimeOut = Getx2RCInt(strName.c_str(), "DropDelay", 0, 0, 5000);
    iTimerRefresh = Getx2RCInt(strName.c_str(), "UpdateInterval", 1000, 250);
    iTrackingRefresh = Getx2RCInt(strName.c_str(), "TrackingInterval", 250, 50, 1000);
    iFlashRefresh = Getx2RCInt(strName.c_str(), "FlashInterval", 500, 250);
    iSpecialUpdate = Getx2RCInt(strName.c_str(), "SpecialUpdateInterval", 5000, 1000);

    if (!refresh)
    {
        int maxX = iScreenWidth;
        int maxY = iScreenHeight;
        if ( box )
        {
            RECT rc;
            GetClientRect(box, &rc);
            maxX = rc.right;
            maxY = rc.bottom;
        }

        int iStartx = Getx2RCCoord(strName.c_str(), "X", 0, maxX, false);
        int iStarty = Getx2RCCoord(strName.c_str(), "Y", 0, maxY, true);
        int iStartWidth = 0;
        int iStartHeight = 0;

        if ( !bAutoSize )
        {
            iStartWidth = Getx2RCDimen(strName.c_str(), "Width", maxX, maxX, false);
            iStartHeight = Getx2RCDimen(strName.c_str(), "Height", 32, maxY, true);
        }   
        else
        {
            iStartWidth = 0;
            iStartHeight = 0;
        }
        
        Reposition(3, iStartx, iStarty, iStartWidth, iStartHeight);

        if ( bShowToolTip && !pTooltip )
        {
            pTooltip = Create_xTooltipClass(hWnd, hInstance);
            hWndToolTip = pTooltip->configure(strName.c_str(), "");
        
            pTooltip->apply("", true);
        }
    }
    else
    {
        if (pTooltip)
        {
            Destroy_xTooltipClass(pTooltip);
            pTooltip = 0;
        }

        if ( bShowToolTip && !pTooltip )
        {
            pTooltip = Create_xTooltipClass(hWnd, hInstance);
            hWndToolTip = pTooltip->configure(strName.c_str(), "");
        
            pTooltip->apply("", true);
        }
    }

    iAlphaTransparency = Getx2RCInt(strName.c_str(), "AlphaTransparency", 255, 0, 255);

    bAlphaMap = Getx2RCBool(strName.c_str(), "AlphaMap");

    string fade = iGetx2RCLine(strName, "CustomAlphaFade", "");
    numTokens = SplitString(fade, one, two, three, four, five, six, seven);

    if (fade.empty())
    {
        bAlphaFade = false;
    }
    else if (numTokens >= 2)
    {
        iAlphaFadeStep = atoi( one.c_str() );
        if (iAlphaFadeStep < 1 || iAlphaFadeStep > 255)
            iAlphaFadeStep = 2;

        iAlphaFadeDelay = atoi( two.c_str() );
        if (iAlphaFadeDelay < 1 || iAlphaFadeDelay > 500)
            iAlphaFadeDelay = 10;

        bAlphaFade = true;
    }

    if (!bAlphaFade)
        bAlphaFade = Getx2RCBool(strName.c_str(), "AlphaFade");

    if ( IsOS(OS_2KXP) && !refresh && (iAlphaTransparency != 255 || bAlphaFade || bAlphaMap) )
    {
        ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);
        bLayeredFlag = true;

        if (bAlphaFade)
            UpdateWindowTransparency(iUpdateAlpha);
        else
            UpdateWindowTransparency(iAlphaTransparency);
    }
    else if ( !IsOS(OS_2KXP) && !refresh)
    {
        iAlphaTransparency = 255;
        bAlphaFade = false;
        bAlphaMap = false;
    }

    bNormalWindowzOrder = Getx2RCBool(strName.c_str(), "WindowzOrder");

    strBehindWindow = iGetx2RCLine(strName, "BehindWindow", "");

    EnumWindows(Taskbar::EnumWindowsProc, (LPARAM) this);

    if (!refresh)
    {
        SetAlwaysOnTop(Getx2RCBool(strName.c_str(), "AlwaysOnTop"), true);
    }
    else
    {
        if ( iAutoSort == 0 && backuptaskbuttons.size() == taskbuttons.size() )
        {
            displayedtaskbuttons.clear();

            for (vector<HWND>::iterator hit = backuptaskbuttons.begin(); hit != backuptaskbuttons.end(); ++hit)
            {
                for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
                {   
                    if ( (*tbit)->hWndButtonTask == (*hit) || (bUseGrouping && (*tbit)->SearchInGroup((*hit)) ) )
                    {
                        displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                        break;
                    }
                }
            }

            if ( displayedtaskbuttons.size() == taskbuttons.size() )
            {
                taskbuttons.clear();

                for (TaskbarButtonIterator tbit = displayedtaskbuttons.begin(); tbit != displayedtaskbuttons.end(); ++tbit)
                {
                    taskbuttons.insert(taskbuttons.end(), (*tbit));
                }
            }

            displayedtaskbuttons.clear();
            backuptaskbuttons.clear();
        }

        int nButtonsAddedAtStartup = CheckDisplayedTasks();

        for(int nI = 1; nI <= nButtonsAddedAtStartup; nI++)
        {
            PrefixedLSSetVariable(strName, "CurrentButtonCount", IntToString( nI ));
            if( !taskbarAdd.empty() )
                LSExecute(hWnd, taskbarAdd.c_str(), SW_SHOWNORMAL);

            if ( iWrapCount != -1 )
            {
                int realwrapCount = iWrapCount+iWrapCountAdded;
                int availableButtonCount = nI;

                if ( availableButtonCount/realwrapCount >= iLines_Rows+iLines_RowsAdded && availableButtonCount%realwrapCount == 1 )
                {
                    iLines_RowsAdded++;
                    if ( iLines_Rows+iLines_RowsAdded <= iMaxLines_Rows )
                    {   
                        PrefixedLSSetVariable(strName, "CurrentLines", IntToString(iLines_Rows+iLines_RowsAdded));
                        if ( !taskbarLineAdd.empty() )
                            LSExecute(hWnd, taskbarLineAdd.c_str(), SW_SHOWNORMAL);
                    }
                    else if ( iLines_Rows+iLines_RowsAdded > iMaxLines_Rows)
                    {
                        iLines_RowsAdded--;
                        iWrapCountAdded++;
                    }
                }
            }
        }
            
        if ( bAutoSize )
            AutoSizeFunction(nButtonsAddedAtStartup);

        iBackupExportedButtonCount = nButtonsAddedAtStartup;

        bCreateLayoutLocked = false;

        CreateLayout();

        if ( bVWMCurrentDesktopOnly || (iRealMonitors > 1 && iThisMultiMonitorOnly != -1) )
            SetTimer(hWnd, TIMER_SPECIALUPDATE, iSpecialUpdate, NULL);
        else
            KillTimer(hWnd, TIMER_SPECIALUPDATE);
    }
}

void Taskbar::SwitchToButton( LPCSTR direction )
{
    HWND prev = NULL;
    HWND next = NULL;

    if ( displayedtaskbuttons.size() <= 1 )
        return;

    for (TaskbarButtonIterator tbit = displayedtaskbuttons.begin(); tbit != displayedtaskbuttons.end(); ++tbit)
    {
        if((*tbit)->GetCurrentState() == STATE_SELECTED)
        {
            if ( _stricmp(direction, "next") == NULL )
            {
                ++tbit;
                if(tbit != displayedtaskbuttons.end())
                {
                    next = (*tbit)->hWndButtonTask;
                }
                else
                {
                    next = (*displayedtaskbuttons.begin())->hWndButtonTask; 
                }
                
                if ( next != NULL )
                {
                    hWndLastActiveForSwitch = next;
                    SelectWindow(next, (GetState(next) == STATE_MINIMIZED ? true : false));
                }
            }
            else /*if ( _stricmp(direction, "prev") == NULL )*/
            {
                if (tbit == displayedtaskbuttons.begin())
                {
                    prev = (*(displayedtaskbuttons.end()-1))->hWndButtonTask;
                }
                else
                {
                    --tbit;
                    prev = (*tbit)->hWndButtonTask;
                }
                
                if ( prev != NULL )
                {
                    hWndLastActiveForSwitch = prev;
                    SelectWindow(prev, (GetState(prev) == STATE_MINIMIZED ? true : false));
                }
            }
            return;
        }
    }

    if ( hWndLastActiveForSwitch == NULL || !IsWindow( hWndLastActiveForSwitch) )
        hWndLastActiveForSwitch = (*displayedtaskbuttons.begin())->hWndButtonTask;

    for (tbit = displayedtaskbuttons.begin(); tbit != displayedtaskbuttons.end(); ++tbit)
    {
        if((*tbit)->hWndButtonTask == hWndLastActiveForSwitch || (bUseGrouping && (*tbit)->SearchInGroup(hWndLastActiveForSwitch)) )
        {
            if ( _stricmp(direction, "next") == NULL )
            {
                ++tbit;
                if(tbit != displayedtaskbuttons.end())
                {
                    next = (*tbit)->hWndButtonTask;
                }
                else
                {
                    next = (*displayedtaskbuttons.begin())->hWndButtonTask; 
                }
                
                if ( next != NULL )
                {
                    hWndLastActiveForSwitch = next;
                    SelectWindow(next, (GetState(next) == STATE_MINIMIZED ? true : false));
                }
            }
            else /*if ( _stricmp(direction, "prev") == NULL )*/
            {
                if (tbit == displayedtaskbuttons.begin())
                {
                    prev = (*(displayedtaskbuttons.end()-1))->hWndButtonTask;
                }
                else
                {
                    --tbit;
                    prev = (*tbit)->hWndButtonTask;
                }
                
                if ( prev != NULL )
                {
                    hWndLastActiveForSwitch = prev;
                    SelectWindow(prev, (GetState(prev) == STATE_MINIMIZED ? true : false));
                }
            }
            return;
        }
    }
}

void Taskbar::ScrollButtons(bool next, int amount)
{
    if ( displayedtaskbuttons.size() > iMaxButtonCount )
    {
        if (next)
        {
            if ( iStartWithPosition+amount < displayedtaskbuttons.size()-iMaxButtonCount )
                iStartWithPosition = iStartWithPosition+amount;
            else
                iStartWithPosition = displayedtaskbuttons.size()-iMaxButtonCount;
        }
        else
        {
            if ( iStartWithPosition-amount < 0 )
                iStartWithPosition = 0;
            else
                iStartWithPosition = iStartWithPosition-amount;
        }

        CreateLayout();
    }
}

void Taskbar::SetAlwaysOnTop(bool bNewAlwaysOnTop, bool bFirstTime)
{
    if (box)
        return;
    
    bool bRealChange = (bAlwaysOnTop != bNewAlwaysOnTop ? true : false);

    bAlwaysOnTop = bNewAlwaysOnTop;

    if ( (bFirstTime || bRealChange) && !bNormalWindowzOrder )
    {
        ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
        SetParent(hWnd, (bAlwaysOnTop || bNormalWindowzOrder || iRealMonitors > 1 ) ? 0 : GetLitestepDesktop());
        ModifyStyle(hWnd, WS_CHILD, WS_POPUP);
    }

    if ( bAlwaysOnTop )
    {   
        SetWindowPos(hWnd, HWND_TOPMOST,
                            -1, -1, -1, -1,
                            SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
    }
    else
    {
        SetWindowPos(hWnd, HWND_NOTOPMOST,
                            -1, -1, -1, -1,
                            SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
    }
    
    if ( !bNormalWindowzOrder )
    {
        if ( !strBehindWindow.empty() && hWndBehindWindow == NULL )
        {
            string one = "";
            string two = "";
            string three = "";
            string four = "";
            string five = "";
            string six = "";
            string seven = "";

            SplitString(strBehindWindow, one, two, three, four, five, six, seven);
        
            if ( _stricmp(one.c_str(), "HWND_TOP") == NULL )
                hWndBehindWindow = HWND_TOP;
            else
                hWndBehindWindow = FindWindowEx2(one.c_str(), two.c_str());

            if ( hWndBehindWindow != NULL )
            {
                SetWindowPos(hWnd, hWndBehindWindow,
                                    -1, -1, -1, -1,
                                    SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
            }
        }
        else if ( hWndBehindWindow != NULL )
        {
            SetWindowPos(hWnd, hWndBehindWindow,
                                -1, -1, -1, -1,
                                SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
        }
    }
}

void Taskbar::SetAlpha(int newAlphaTransparency)
{
    if (box)
        return;

    if ( newAlphaTransparency >= 0 && newAlphaTransparency <= 255 && IsOS(OS_2KXP) )
    {
        if (!bLayeredFlag)
        {
            ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
            SetParent(hWnd, 0);
            ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

            ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);

            ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
            SetParent(hWnd, (bAlwaysOnTop || bNormalWindowzOrder || iRealMonitors > 1 ) ? 0 : GetLitestepDesktop());
            ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

            bLayeredFlag = true;
        }

        iAlphaTransparency = newAlphaTransparency;

        UpdateWindowTransparency(newAlphaTransparency);
    }
}

void Taskbar::SetAlphaFade(int newAlphaTransparency, int step, int delay)
{
    if (box)
        return;

    if (newAlphaTransparency >= 0 && newAlphaTransparency <= 255 && IsOS(OS_2KXP) )
    {
        if (!bLayeredFlag)
        {
            ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
            SetParent(hWnd, 0);
            ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

            ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);

            ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
            SetParent(hWnd, (bAlwaysOnTop || bNormalWindowzOrder || iRealMonitors > 1 ) ? 0 : GetLitestepDesktop());
            ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

            bLayeredFlag = true;
        }
        
        iUpdateAlpha = iAlphaTransparency;
        
        iAlphaFadeBangEndValue = newAlphaTransparency;

        if ( iUpdateAlpha > iAlphaFadeBangEndValue )
        {
            KillTimer(hWnd, TIMER_ALPHAFADEUPBANG);
            SetTimer(hWnd, TIMER_ALPHAFADEDOWNBANG, delay, NULL);
        }
        else if ( iUpdateAlpha < iAlphaFadeBangEndValue )
        {
            KillTimer(hWnd, TIMER_ALPHAFADEDOWNBANG);
            SetTimer(hWnd, TIMER_ALPHAFADEUPBANG, delay, NULL);
        }
    }
}

void Taskbar::UpdateWindowTransparency(int newAlphaTransparency)
{
    /* First, see if we can get the API call we need. If we've tried
    * once, we don't need to try again. */
    if (!bAlphaInitialized)
    {
        HMODULE hDLL = LoadLibrary ("user32");

        pUpdateLayeredWindow = 
                    (PULW) GetProcAddress(hDLL, "UpdateLayeredWindow");

        bAlphaInitialized = true;
   
        if (pUpdateLayeredWindow == NULL)
        {
            ModifyStyleEx(hWnd, WS_EX_LAYERED, 0 );

            iAlphaTransparency = 255;
            bAlphaMap = false;
            bAlphaFade = false;

            bLayeredFlag = false;

            return;
        }
        else
        {
            bLayeredFlag = true;

            SetWindowRgn(hWnd, NULL, true);

            if ( backupBitmap )
            {
                backupBitmap = (HBITMAP) SelectObject(backupDC, oldbackupBMP);
                DeleteObject(backupBitmap);
            }
            if ( backupDC )
                DeleteDC(backupDC);
            if ( oldbackupBMP )
                DeleteObject(oldbackupBMP);

            backupDC = NULL;
            backupBitmap = NULL;
            oldbackupBMP = NULL;
        }
    }

    /* Now, we need to set the 'layered window attributes'. This
     * is where the alpha values get set. */
    //------------------------------------------------------------------
    RECT rectPosition;
    GetWindowRect(hWnd, &rectPosition);
        
    int width = (rectPosition.right-rectPosition.left);
    int height = (rectPosition.bottom-rectPosition.top);

    if ( width <= 0 || height <= 0 )
        return;

    POINT boxConverted = {rectPosition.left, rectPosition.top};
    if (box)
        ScreenToClient(box, &boxConverted);
    int x = boxConverted.x;
    int y = boxConverted.y;

    HDC backgroundDC = CreateCompatibleDC(NULL);
            
    BITMAPINFO alpha;
    ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
    alpha.bmiHeader.biWidth=width;      // Set size you need
    alpha.bmiHeader.biHeight=height;    // Set size you need
    alpha.bmiHeader.biPlanes=1;
    alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
    alpha.bmiHeader.biSizeImage=0;
    alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
    alpha.bmiHeader.biClrUsed= 0;
    alpha.bmiHeader.biClrImportant= 0;
    VOID *pvBits;

    HBITMAP backgroundBitmap = CreateDIBSection( backgroundDC,
                                                    &alpha,
                                                    DIB_RGB_COLORS,
                                                    &pvBits,
                                                    NULL,
                                                    0 );

    HBITMAP oldBMP = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);

    // Use Complete Backup for WindowsInternal Repaint (Covering Windows, ...)
    if ( backupBitmap != NULL && backupDC != NULL )
    {
        BitBlt(backgroundDC, 0, 0, width, height, backupDC, 0, 0, SRCCOPY);
    }
    else
    //1----------------------------------------------------------------
    {
    //1----------------------------------------------------------------

        if ( !bAlphaMap && !bTrueTransparency )
        {
            //  =======================================================================
            //  Paint desktop on display DC and then into background buffer
            //  =======================================================================
            PaintDesktopEx(backgroundDC, 0, 0, width, height, x, y, false);
        }
        
        BITMAP bm;
        GetObject(backgroundBitmap, sizeof(BITMAP), &bm);

        BYTE *pPixel = (LPBYTE)bm.bmBits;

        int iSize = width * height;

        int j = 0;

        for (j = 0; j < iSize; ++j)
        {
            if ( !bAlphaMap && bTrueTransparency )
            {
                pPixel[0] = 255;
                pPixel[1] = 0;
                pPixel[2] = 255;
                pPixel[3] = 255;
            }
            else if ( !bAlphaMap && !bTrueTransparency )
            {
                pPixel[3] = 255;
            }
            else
            {
                pPixel[0] = 0;
                pPixel[1] = 0;
                pPixel[2] = 0;
                pPixel[3] = 0;
            }

            pPixel+= 4;
        }

        //Normal Background Painting
        pTaskbarTexture->apply(backgroundDC, 0, 0, width, height);

        if ( !listGhostTaskbar.empty() )
        {
            for(GhostTaskbarListIterator glit = listGhostTaskbar.begin(); glit != listGhostTaskbar.end(); glit++)
            {
                if ( *glit != NULL )
                {
                    if ( (*glit)->bOnTop )
                        continue;

                    if ( !(*glit)->bDoPaint )
                        continue;

                    int gtx = xParseCoord((*glit)->strOrgX.c_str(), 0, width, false);
                    int gty = xParseCoord((*glit)->strOrgY.c_str(), 0, height, true);
                    int gtwidth = xParseDimen((*glit)->strOrgWidth.c_str(), width, width, false);
                    int gtheight = xParseDimen((*glit)->strOrgHeight.c_str(), height, height, true);

                    (*glit)->pTaskbarTexture->apply(backgroundDC, gtx, gty, gtwidth, gtheight);
                }
                else
                {
                    break;
                }
            }
        }
        
        for (TaskbarButtonIterator tbit = displayedtaskbuttons.begin(); tbit != displayedtaskbuttons.end(); ++tbit)
        {
            if ( (*tbit) != pDontPaintAddedTask )
                (*tbit)->Paint( backgroundDC );
        }
        
        if ( !listGhostTaskbar.empty() )
        {
            for(GhostTaskbarListIterator glit = listGhostTaskbar.begin(); glit != listGhostTaskbar.end(); glit++)
            {
                if ( *glit != NULL )
                {
                    if ( !(*glit)->bOnTop )
                        continue;

                    if ( !(*glit)->bDoPaint )
                        continue;

                    int gtx = xParseCoord((*glit)->strOrgX.c_str(), 0, width, false);
                    int gty = xParseCoord((*glit)->strOrgY.c_str(), 0, height, true);
                    int gtwidth = xParseDimen((*glit)->strOrgWidth.c_str(), width, width, false);
                    int gtheight = xParseDimen((*glit)->strOrgHeight.c_str(), height, height, true);

                    (*glit)->pTaskbarTexture->apply(backgroundDC, gtx, gty, gtwidth, gtheight);
                }
                else
                {
                    break;
                }
            }
        }

        //Prepare Combined Image
        GetObject(backgroundBitmap, sizeof(BITMAP), &bm);

        pPixel = (LPBYTE)bm.bmBits;

        for (j = 0; j < iSize; ++j)
        {
            if (pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255)
            {
                pPixel[0] = 0;
                pPixel[1] = 0;
                pPixel[2] = 0;
                pPixel[3] = 0;
            }
            else if ( !bAlphaMap )
            {
                pPixel[3] = 255;
            }
                
            pPixel += 4;
        }
        
        //Create BackupBitmap, if it doesn't exist!
        //----------------------------------------------------------------------
        if ( backupBitmap == NULL && backupDC == NULL )
        {
            backupDC = CreateCompatibleDC(NULL);
            
            BITMAPINFO alpha;
            ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
            alpha.bmiHeader.biWidth=width;      // Set size you need
            alpha.bmiHeader.biHeight=height;    // Set size you need
            alpha.bmiHeader.biPlanes=1;
            alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
            alpha.bmiHeader.biSizeImage=0;
            alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
            alpha.bmiHeader.biClrUsed= 0;
            alpha.bmiHeader.biClrImportant= 0;
            VOID *pvBits;

            backupBitmap = CreateDIBSection( backupDC,
                                                    &alpha,
                                                    DIB_RGB_COLORS,
                                                    &pvBits,
                                                    NULL,
                                                    0 );

            oldbackupBMP = (HBITMAP) SelectObject(backupDC, backupBitmap);

            BitBlt(backupDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);
        }
        //----------------------------------------------------------------------
    //1
    }
    //1

    POINT ptWindowScreenPosition = {x, y};
    SIZE szWindow = {width, height};

    POINT ptSrc = {0,0};

    BLENDFUNCTION blendPixelFunction = { AC_SRC_OVER, 0, newAlphaTransparency, AC_SRC_ALPHA };

    pUpdateLayeredWindow(hWnd, NULL, &ptWindowScreenPosition, &szWindow, backgroundDC, &ptSrc, 0, &blendPixelFunction, ULW_ALPHA );
    //------------------------------------------------------------------*/

    backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, oldBMP);
    DeleteDC(backgroundDC);
    DeleteObject(backgroundBitmap);
    DeleteObject(oldBMP);
}

//------------------------------------------------------------
//Show/Hide
//------------------------------------------------------------
void Taskbar::Show()
{
    if ( bParentDestroyNotificationReceived )
        return;

    if (bAlphaFade && !bVisible)
    {
        KillTimer(hWnd, TIMER_ALPHAFADEOUT);
        SetTimer(hWnd, TIMER_ALPHAFADEIN, iAlphaFadeDelay, NULL);
    }
    else if ( !bVisible )
    {
        ShowWindow(hWnd, SW_SHOWNOACTIVATE);
    }

    bVisible = true;

    Repaint();
}

void Taskbar::Hide()
{
    if (bAlphaFade && bVisible)
    {
        iUpdateAlpha = iAlphaTransparency;

        KillTimer(hWnd, TIMER_ALPHAFADEIN);
        SetTimer(hWnd, TIMER_ALPHAFADEOUT, iAlphaFadeDelay, NULL);
    }
    else if ( bVisible )
    {
        ShowWindow(hWnd, SW_HIDE);
    }

    bVisible = false;
}

//------------------------------------------------------------
//Move/Resize/Reposition
//------------------------------------------------------------
//Type == 1 -> Move
//Type == 2 -> Resize
//Type == 3 -> Reposition

void Taskbar::Reposition(int type, int newx, int newy, int newwidth, int newheight, int steps, int time, bool repositionBy)
{
    RECT rectPosition;
    GetWindowRect(hWnd, &rectPosition);
        
    int width = (rectPosition.right-rectPosition.left);
    int height = (rectPosition.bottom-rectPosition.top);

    POINT boxConverted;
    boxConverted.x = rectPosition.left;
    boxConverted.y = rectPosition.top;

    if (box)
        ScreenToClient(box, &boxConverted);

    int x = boxConverted.x;
    int y = boxConverted.y;

    if (steps > 0 && time > 0)
    {
        if (time > 50)
            time = 50;

        if (repositionBy)
        {
            if (type != 2)
            {
                newx = x + newx;
                newy = y + newy;
            }
            if (type != 1)
            {
                newwidth = width + newwidth;
                newheight = height + newheight;
            }
        }

        bool dynamicSizeSkip = false;

        int counter = 0;
        int deltax = newx - x;
        int deltay = newy - y;
        int deltasizex = newwidth - width;
        int deltasizey = newheight - height;

        bool growx;
        bool growy;
        bool growsizex;
        bool growsizey;
        
        if (deltax < 0)
            growx = false;
        else
            growx = true;
        if (deltay < 0)
            growy = false;
        else
            growy = true;
        
        if (deltasizex < 0)
            growsizex = false;
        else
            growsizex = true;
        if (deltasizey < 0)
            growsizey = false;
        else
            growsizey = true;

        int restx = 0;
        int resty = 0;
        int restw = 0;
        int resth = 0;

        div_t div_result;
        div_result = div( deltax, steps );
        deltax = div_result.quot;
        restx = div_result.rem;
        div_result = div( deltay, steps );
        deltay = div_result.quot;
        resty = div_result.rem;

        div_result = div( deltasizex, steps );
        deltasizex = div_result.quot;
        restw = div_result.rem;
        div_result = div( deltasizey, steps );
        deltasizey = div_result.quot;
        resth = div_result.rem;

        if (deltax == 0 && growx)
            deltax = 1;
        else if (deltax == 0 && !growx)
            deltax = -1;
        if (deltay == 0 && growy)
            deltay = 1;
        else if (deltay == 0 && !growy)
            deltay = -1;

        if (deltasizex == 0 && growsizex)
            deltasizex = 1;
        else if (deltasizex == 0 && !growsizex)
            deltasizex = -1;
        if (deltasizey == 0 && growsizey)
            deltasizey = 1;
        else if (deltasizey == 0 && !growsizey)
            deltasizey = -1;

        int dynx = x;
        int dyny = y;
        int dynwidth = width;
        int dynheight = height;

        int stopper = 0;
        while (counter <= steps)
        {
            stopper = 0; // Stop if x, y, width and height are finished, stopper = 4
            if ( restx > 0 && deltax > 0 )
            {
                dynx = dynx+deltax+1;
                restx--;
            }
            else if ( restx < 0 && deltax < 0 )
            {
                dynx = dynx+deltax-1;
                restx++;
            }
            else
                dynx = dynx+deltax;

            if (growx && dynx > newx)
            {
                dynx = newx;
                stopper++;
            }
            else if (!growx && dynx < newx)
            {
                dynx = newx;
                stopper++;
            }

            if ( resty > 0 && deltay > 0 )
            {
                dyny = dyny+deltay+1;
                resty--;
            }
            else if ( resty < 0 && deltay < 0 )
            {
                dyny = dyny+deltay-1;
                resty++;
            }
            else
                dyny = dyny+deltay;

            if (growy && dyny > newy)
            {
                dyny = newy;
                stopper++;
            }
            else if (!growy && dyny < newy)
            {
                dyny = newy;
                stopper++;
            }

            if ( restw > 0 && deltasizex > 0 )
            {
                dynwidth = dynwidth+deltasizex+1;
                restw--;
            }
            else if ( restw < 0 && deltasizex < 0 )
            {
                dynwidth = dynwidth+deltasizex-1;
                restw++;
            }
            else
                dynwidth = dynwidth+deltasizex;

            if (growsizex && dynwidth > newwidth)
            {
                dynwidth = newwidth;
                stopper++;
            }
            else if (!growsizex && dynwidth < newwidth)
            {
                dynwidth = newwidth;
                stopper++;
            }

            if ( resth > 0 && deltasizey > 0 )
            {
                dynheight = dynheight+deltasizey+1;
                resth--;
            }
            else if ( resth < 0 && deltasizey < 0 )
            {
                dynheight = dynheight+deltasizey-1;
                resth++;
            }
            else
                dynheight = dynheight+deltasizey;

            if (growsizey && dynheight > newheight)
            {
                dynheight = newheight;
                stopper++;
            }
            else if (!growsizey && dynheight < newheight)
            {
                dynheight = newheight;
                stopper++;
            }
                
            if ( (type == 2 || type == 3) && !dynamicSizeSkip )
            {
                if ( (iLayout == 0 && iDirection != 2 && (iWrapCount == -1 || (iWrapCount != -1 && iWrapDirection == 1)))
                    || (iLayout == 1 && iDirection != 4 && (iWrapCount == -1 || (iWrapCount != -1 && iWrapDirection == 3))) )
                {
                    CreateLayout(newwidth, newheight);
                    dynamicSizeSkip = true;
                }
                else
                {
                    CreateLayout(dynwidth, dynheight);
                }
            }

            SetWindowPos(hWnd, 0, dynx, dyny, dynwidth, dynheight,
                    SWP_NOACTIVATE | (type == 2 ? SWP_NOMOVE : 0) | (type == 1 ? SWP_NOSIZE : 0 ) | SWP_NOZORDER);
    
            if (box != NULL)
                RedrawWindow(box, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);

            RedrawWindow(hWnd, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);

            if (stopper == 4)
                counter = steps;

            counter++;

            Sleep(time);
        }

        //Paint Added task with the last Resize Action
        pDontPaintAddedTask = NULL;

        if ( type == 1 )
        {
            SetWindowPos(hWnd, 0, newx, newy, 0, 0,
                    SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
        }
        else if ( type == 2 )
        {
            CreateLayout(newwidth, newheight);

            SetWindowPos(hWnd, 0, 0, 0, newwidth, newheight,
                    SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);
        }
        else if ( type == 3 )
        {
            CreateLayout(newwidth, newheight);

            SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
                    SWP_NOACTIVATE | SWP_NOZORDER);
        }
    }
    else
    {
        if (repositionBy)
        {
            if (type != 2)
            {
                newx = x + newx;
                newy = y + newy;
            }
            if (type != 1)
            {
                newwidth = width + newwidth;
                newheight = height + newheight;
            }
        }

        //Paint Added task with the last Resize Action
        pDontPaintAddedTask = NULL;

        if ( type == 1 )
        {
            SetWindowPos(hWnd, 0, newx, newy, 0, 0,
                    SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
        }
        else if ( type == 2 )
        {
            CreateLayout(newwidth, newheight);

            SetWindowPos(hWnd, 0, 0, 0, newwidth, newheight,
                    SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);
        }
        else if ( type == 3 )
        {
            CreateLayout(newwidth, newheight);

            SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
                    SWP_NOACTIVATE | SWP_NOZORDER);
        }
    }
}

void Taskbar::UpdateRegion()
{   
    RECT rectPosition;
    GetClientRect(hWnd, &rectPosition);

    if ( rectPosition.right-rectPosition.left <= 0 || rectPosition.bottom-rectPosition.top <= 0 )
        return;

    HDC tmpDC = CreateCompatibleDC(NULL);
    HBITMAP tmpBmp = CreateBitmap(rectPosition.right, rectPosition.bottom, 1, bitsperpixel, NULL);
    HBITMAP oldBMP = (HBITMAP) SelectObject(tmpDC, tmpBmp);

    onPaint(tmpDC);
                                
    tmpBmp = (HBITMAP) SelectObject(tmpDC, oldBMP);

    HRGN region = BitmapToRegion(tmpBmp, RGB(255, 0, 255), 0, 0, 0);
    HRGN windowRgn = CreateRectRgn(0, 0, 0, 0);
    CombineRgn(windowRgn, region, NULL, RGN_COPY);

    SetWindowRgn(hWnd, windowRgn, true);    

    DeleteObject(windowRgn);
    DeleteObject(region);

    DeleteDC(tmpDC);
    DeleteObject(tmpBmp);
    DeleteObject(oldBMP);
}

//------------------------------------------------------------
//Mouse Related Actions (Clicks, Enter, Leave, Move, ...)
//------------------------------------------------------------

void Taskbar::onLButtonDown(int x, int y)
{
    TaskbarButton *button = ButtonAtPoint( x, y );

    if ( button != NULL )
    {
        if ( bActionsOnDownButton )
        {
            if ( button->iCurrentState == STATE_SELECTED )
            {
                hWndLastActiveForGrouping = button->hWndButtonTask;
                button->ButtonClick( x, y, iLButtonAction, true);
            }
            else
            {
                button->ButtonClick( x, y, iLButtonAction);
            }
        }
        else
        {
            if ( button->iCurrentState == STATE_SELECTED )
            {
                hWndLastActiveForGrouping = button->hWndButtonTask;
                bMinimizeAtAllCosts = true;
            }
        }

        if ( bUsePressedState )
        {
            button->bIsPressed = true;
            Repaint();
        }

        pMoveCapture = button;
    }
    else if ( button == NULL && !taskbarLButtonDown.empty() )
    {
        LSExecute(hWnd, taskbarLButtonDown.c_str(), SW_SHOWNORMAL);
    }
}

void Taskbar::onLButtonUp(int x, int y)
{   
    TaskbarButton *button = ButtonAtPoint( x, y );

    if ( button != NULL && pMoveCapture != NULL && button == pMoveCapture )
    {
        if ( !bActionsOnDownButton )
            button->ButtonClick( x, y, iLButtonAction, bMinimizeAtAllCosts );
                
        if ( bUsePressedState )
        {
            button->bIsPressed = false;
            Repaint();
        }
    }
    //Drag 'n Drop Positioning
    else if ( iAutoSort == 0 && button != NULL && pMoveCapture != NULL && pMoveCapture != button && iDisplayButtonType < 0 )
    {   
        vector<TaskbarButton *> tmptaskbuttons;
        
        for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
        {
            if ( *tbit == button )
            {
                if ( button->iPositionInTaskbar < pMoveCapture->iPositionInTaskbar )
                {
                    tmptaskbuttons.insert( tmptaskbuttons.end(), (pMoveCapture) );
                    tmptaskbuttons.insert( tmptaskbuttons.end(), (button) );
                }
                else if ( button->iPositionInTaskbar > pMoveCapture->iPositionInTaskbar )
                {
                    tmptaskbuttons.insert( tmptaskbuttons.end(), (button) );
                    tmptaskbuttons.insert( tmptaskbuttons.end(), (pMoveCapture) );
                }
            }
            else if ( *tbit != pMoveCapture )
            {
                tmptaskbuttons.insert( tmptaskbuttons.end(), (*tbit) );
            }
        }

        taskbuttons = tmptaskbuttons;
        tmptaskbuttons.clear();

        if ( bUsePressedState )
            pMoveCapture->bIsPressed = false;
        
        pMoveCapture = NULL;

        CreateLayout();
    }
    else if ( button == NULL && !taskbarLButtonUp.empty() )
    {
        LSExecute(hWnd, taskbarLButtonUp.c_str(), SW_SHOWNORMAL);
    }

    if ( bUsePressedState )
    {
        if ( pMoveCapture != NULL )
            pMoveCapture->bIsPressed = false;

        Repaint();
    }

    pMoveCapture = NULL;

    bMinimizeAtAllCosts = false;
}

void Taskbar::onMButtonDown(int x, int y)
{
    TaskbarButton *button = ButtonAtPoint( x, y );

    if ( button != NULL )
    {
        if ( bActionsOnDownButton )
        {
            button->ButtonClick( x, y, iMButtonAction );
        }
        else
        {
            if ( button->iCurrentState == STATE_SELECTED )
                bMinimizeAtAllCosts = true;
        }
        
        if ( bUsePressedState )
        {
            button->bIsPressed = true;
            Repaint();
        }

        pClickCapture = button;
    }
    else if ( button == NULL && !taskbarMButtonDown.empty() )
    {
        LSExecute(hWnd, taskbarMButtonDown.c_str(), SW_SHOWNORMAL);
    }
}

void Taskbar::onMButtonUp(int x, int y)
{
    TaskbarButton *button = ButtonAtPoint( x, y );

    if ( button != NULL && pClickCapture != NULL && button == pClickCapture )
    {
        if ( !bActionsOnDownButton )
            button->ButtonClick( x, y, iMButtonAction, bMinimizeAtAllCosts );
        
        if ( bUsePressedState )
        {
            button->bIsPressed = false;
            Repaint();
        }
    }
    else if ( button == NULL && !taskbarMButtonUp.empty() )
    {
        LSExecute(hWnd, taskbarMButtonUp.c_str(), SW_SHOWNORMAL);
    }

    bMinimizeAtAllCosts = false;
    pClickCapture = NULL;
}

void Taskbar::onRButtonDown(int x, int y)
{
    TaskbarButton *button = ButtonAtPoint( x, y );

    if ( button != NULL )
    {
        if ( bActionsOnDownButton )
        {
            button->ButtonClick( x, y, iRButtonAction );
        }
        else
        {
            if ( button->iCurrentState == STATE_SELECTED )
                bMinimizeAtAllCosts = true;
        }

        if ( bUsePressedState )
        {
            button->bIsPressed = true;
            Repaint();
        }

        pClickCapture = button;
    }
    else if ( button == NULL && !taskbarRButtonDown.empty() )
    {
        LSExecute(hWnd, taskbarRButtonDown.c_str(), SW_SHOWNORMAL);
    }
}

void Taskbar::onRButtonUp(int x, int y)
{
    TaskbarButton *button = ButtonAtPoint( x, y );

    if ( button != NULL && pClickCapture != NULL && button == pClickCapture )
    {
        if ( !bActionsOnDownButton )
            button->ButtonClick( x, y, iRButtonAction, bMinimizeAtAllCosts );
        
        if ( bUsePressedState )
        {
            button->bIsPressed = false;
            Repaint();
        }
    }
    else if ( button == NULL && !taskbarRButtonUp.empty() )
    {
        LSExecute(hWnd, taskbarRButtonUp.c_str(), SW_SHOWNORMAL);
    }
    
    bMinimizeAtAllCosts = false;
    pClickCapture = NULL;
}

void Taskbar::onWheelDown(int x, int y)
{
    if(!taskbarMWheelDown.empty())
        LSExecute(hWnd, taskbarMWheelDown.c_str(), SW_SHOWNORMAL);
}

void Taskbar::onWheelUp(int x, int y)
{
    if(!taskbarMWheelUp.empty())
        LSExecute(hWnd, taskbarMWheelUp.c_str(), SW_SHOWNORMAL);
}

void Taskbar::onMouseEnter()
{
    if(!taskbarEnter.empty())
        LSExecute(hWnd, taskbarEnter.c_str(), SW_SHOWNORMAL);

    if (bFocusOnEnter)
    {   
        bSkipButton = true;
        mouse_event( MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_MIDDLEDOWN,
                        NULL, NULL, NULL, NULL );
        mouse_event( MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_MIDDLEUP,
                        NULL, NULL, NULL, NULL );
    }
}

void Taskbar::onMouseLeave()
{
    if (!taskbarLeave.empty())
        LSExecute(hWnd, taskbarLeave.c_str(), SW_SHOWNORMAL);
}

void Taskbar::onMouseMove(int x, int y)
{
    if (!bMouseInside)
    {
        bMouseInside = true;
        onMouseEnter();
        SetTimer(hWnd, TIMER_MOUSETRACK, iTrackingRefresh, 0);
    }
}

void Taskbar::Repaint()
{   
    if ( backupBitmap )
    {
        backupBitmap = (HBITMAP) SelectObject(backupDC, oldbackupBMP);
        DeleteObject(backupBitmap);
    }
    if ( backupDC )
        DeleteDC(backupDC);
    if ( oldbackupBMP )
        DeleteObject(oldbackupBMP);

    backupDC = NULL;
    backupBitmap = NULL;
    oldbackupBMP = NULL;

    //Attention!! Background MUST exist
    if ( bLayeredFlag && IsOS(OS_2KXP) )
    {
        UpdateWindowTransparency(iAlphaTransparency);
    }
    else
    {
        RedrawWindow(hWnd, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);
            
        if ( bTrueTransparency )
            UpdateRegion();
    }
}

void Taskbar::onPaint(HDC hDC)
{   
    RECT rectPosition;
    GetWindowRect(hWnd, &rectPosition);

    int width = (rectPosition.right-rectPosition.left);
    int height = (rectPosition.bottom-rectPosition.top);
    
    if ( width <= 0 || height <= 0 )
        return;

    POINT boxConverted = {rectPosition.left, rectPosition.top};
    if (box)
        ScreenToClient(box, &boxConverted);
    int x = boxConverted.x;
    int y = boxConverted.y;

    rectPosition.left = 0;
    rectPosition.top = 0;
    rectPosition.right = width;
    rectPosition.bottom = height;

    HDC backgroundDC = 0;
    HBITMAP backgroundBitmap = 0;
    HBITMAP oldBMP = 0;

    // Use Complete Backup for WindowsInternal Repaint (Covering Windows, ...)
    if ( backupBitmap != NULL && backupDC != NULL )
    {
        BitBlt(hDC, 0, 0, width, height, backupDC, 0, 0, SRCCOPY);
    }
    else
    //1----------------------------------------------------------------
    {
    //1----------------------------------------------------------------

        backgroundDC = CreateCompatibleDC(hDC);
        backgroundBitmap = CreateCompatibleBitmap(hDC, width, height);
        oldBMP = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);

        //True Transparency
        if (bTrueTransparency)
        {
            ::SetBkColor(backgroundDC, RGB(255, 0, 255));
            ::ExtTextOut(backgroundDC, 0, 0, ETO_OPAQUE, &rectPosition, NULL, 0, NULL);
        }
        //PrePaint Background (Fake Transparency)
        else
        {
            //  =======================================================================
            //  Paint desktop on display DC and then into background buffer
            //  =======================================================================
            PaintDesktopEx(backgroundDC, 0, 0, width, height, x, y, false);
        }
            
        //Normal Background Painting
        pTaskbarTexture->apply(backgroundDC, 0, 0, width, height);

        if ( !listGhostTaskbar.empty() )
        {
            for(GhostTaskbarListIterator glit = listGhostTaskbar.begin(); glit != listGhostTaskbar.end(); glit++)
            {
                if ( *glit != NULL )
                {
                    if ( (*glit)->bOnTop )
                        continue;

                    if ( !(*glit)->bDoPaint )
                        continue;

                    int gtx = xParseCoord((*glit)->strOrgX.c_str(), 0, width, false);
                    int gty = xParseCoord((*glit)->strOrgY.c_str(), 0, height, true);
                    int gtwidth = xParseDimen((*glit)->strOrgWidth.c_str(), width, width, false);
                    int gtheight = xParseDimen((*glit)->strOrgHeight.c_str(), height, height, true);

                    (*glit)->pTaskbarTexture->apply(backgroundDC, gtx, gty, gtwidth, gtheight);
                }
                else
                {
                    break;
                }
            }
        }

        for (TaskbarButtonIterator tbit = displayedtaskbuttons.begin(); tbit != displayedtaskbuttons.end(); ++tbit)
        {
            if ( (*tbit) != pDontPaintAddedTask )
                (*tbit)->Paint( backgroundDC );
        }
    
        if ( !listGhostTaskbar.empty() )
        {
            for(GhostTaskbarListIterator glit = listGhostTaskbar.begin(); glit != listGhostTaskbar.end(); glit++)
            {
                if ( *glit != NULL )
                {
                    if ( !(*glit)->bOnTop )
                        continue;

                    if ( !(*glit)->bDoPaint )
                        continue;

                    int gtx = xParseCoord((*glit)->strOrgX.c_str(), 0, width, false);
                    int gty = xParseCoord((*glit)->strOrgY.c_str(), 0, height, true);
                    int gtwidth = xParseDimen((*glit)->strOrgWidth.c_str(), width, width, false);
                    int gtheight = xParseDimen((*glit)->strOrgHeight.c_str(), height, height, true);

                    (*glit)->pTaskbarTexture->apply(backgroundDC, gtx, gty, gtwidth, gtheight);
                }
                else
                {
                    break;
                }
            }
        }
        
        // blt the double buffer to the display
        BitBlt(hDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);

        //Create BackupBitmap, if it doesn't exist!
        //----------------------------------------------------------------------
        if ( backupBitmap == NULL && backupDC == NULL )
        {
            backupDC = CreateCompatibleDC(hDC);
            backupBitmap = CreateCompatibleBitmap(hDC, width, height);
            oldbackupBMP = (HBITMAP) SelectObject(backupDC, backupBitmap);

            BitBlt(backupDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);
        }
        //----------------------------------------------------------------------

    //1----------------------------------------------------------------
    }
    //1----------------------------------------------------------------

    if ( backgroundBitmap != NULL && backgroundDC != NULL )
    {
        backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, oldBMP);
        DeleteDC(backgroundDC);
        DeleteObject(backgroundBitmap);
        DeleteObject(oldBMP);
    }
}

void Taskbar::Update(bool needrepaint)
{
    if (!bInCriticalProcess && bVisible)
    {
        for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
        {
            if ( !IsWindow( (*tbit)->hWndButtonTask ) )
            {
                //Grouping
                if (bUseGrouping)
                {
                    if ( (*tbit)->listhWndGroup.size() > 1 && (*tbit)->RemoveFromGroup( (*tbit)->hWndButtonTask ) )
                    {
                        if ( (*tbit)->IsFlashing() )
                        {
                            KillTimer(hWnd, TIMER_FLASH);
                            (*tbit)->SetFlashState(0);
                        }

                        Repaint();

                        return;
                    }
                }
                
                if ( (*tbit)->IsFlashing() )
                {
                    KillTimer(hWnd, TIMER_FLASH);
                }
                delete (*tbit);
                taskbuttons.erase( tbit );
                
                AmountOfTasksChanged();

                CreateLayout();

                return;
            }

            int iOldState = (*tbit)->iCurrentState;

            if ( (*tbit)->Update(false) || (*tbit)->GetCurrentState() != iOldState )
                needrepaint = true;
        }

        if ( bHasOverlayConditional )
        {
            for(GhostButtonListIterator glit = listGhostButton.begin(); glit != listGhostButton.end(); glit++)
            {
                if ( *glit != NULL )
                {
                    if ( !(*glit)->strConditional.empty() )
                    {
                        if ( m_pProcessStatsRequest )
                        {
                            char tmp[4096] = {0};
                            m_pProcessStatsRequest( tmp, (*glit)->strConditional.c_str() );

                            if ( _stricmp(tmp, "0") == NULL || _stricmp(tmp, "false") == NULL || _stricmp(tmp, "off") == NULL )
                            {
                                if ( (*glit)->bDoPaint )
                                    needrepaint = true;

                                (*glit)->bDoPaint = false;
                            }
                            else
                            {
                                if ( !(*glit)->bDoPaint )
                                    needrepaint = true;

                                (*glit)->bDoPaint = true;
                            }
                        }
                    }
                }
            }

            for(GhostTaskbarListIterator gtlit = listGhostTaskbar.begin(); gtlit != listGhostTaskbar.end(); gtlit++)
            {
                if ( *gtlit != NULL )
                {
                    if ( !(*gtlit)->strConditional.empty() )
                    {
                        if ( m_pProcessStatsRequest )
                        {
                            char tmp[4096] = {0};
                            m_pProcessStatsRequest( tmp, (*gtlit)->strConditional.c_str() );

                            if ( _stricmp(tmp, "0") == NULL || _stricmp(tmp, "false") == NULL || _stricmp(tmp, "off") == NULL )
                            {
                                if ( (*gtlit)->bDoPaint )
                                    needrepaint = true;

                                (*gtlit)->bDoPaint = false;
                            }
                            else
                            {
                                if ( !(*gtlit)->bDoPaint )
                                    needrepaint = true;

                                (*gtlit)->bDoPaint = true;
                            }
                        }
                    }
                }
            }
        }

        /*if ( bVWMCurrentDesktopOnly || (iRealMonitors > 1 && iThisMultiMonitorOnly != -1) )
        {
            AmountOfTasksChanged();
            
            CreateLayout();
        }
        else*/ if ( needrepaint )
        {
            Repaint();
        }
    }
}

void Taskbar::onTimer(int timerID)
{
    switch(timerID)
    {   
        case TIMER_UPDATE:
        {
            Update();
        }
        break;

        case TIMER_SPECIALUPDATE:
        {
            if ( bVisible )
            {
                AmountOfTasksChanged();
            
                CreateLayout();
            }
        }
        break;

        case TIMER_MOUSETRACK:
        {
            POINT pt;
            GetCursorPos(&pt);
            
            RECT rc;
            GetWindowRect(hWnd, &rc);

            if ( bTrueTransparency )
            {
                HRGN hrgn = CreateRectRgn(0,0,0,0);
                int regionType = GetWindowRgn(hWnd, hrgn);
                if (regionType != ERROR) 
                { 
                    /* hrgn contains window region */ 
                    ScreenToClient(hWnd, &pt);
                    if(!PtInRegion(hrgn, pt.x, pt.y))
                    {
                        KillTimer(hWnd, TIMER_MOUSETRACK);

                        bMouseInside = false;
                        onMouseLeave(); 

                        if (pTooltip)
                            pTooltip->apply("");

                        PrefixedLSSetVariable(strName, "HoverTask", "");

                        if (bUseHoverState || bUsePressedState || bScroll)
                        {
                            for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
                            {
                                if ( bUseHoverState )
                                    (*tbit)->bIsHover = false;

                                if ( bUsePressedState )
                                    (*tbit)->bIsPressed = false;

                                if ( bScroll && (*tbit)->bIsScrolling )
                                {
                                    (*tbit)->bIsScrolling = false;
                                    bIsScrolling = false;
                                    (*tbit)->iScrollPosition = 0;
                                    KillTimer(hWnd, TIMER_SCROLL);
                                }
                            }

                            Repaint();
                        }

                        pCapture = NULL;
                    
                        DeleteObject(hrgn); /* finished with region */

                        break;
                    }
                }

                DeleteObject(hrgn); /* finished with region */
            }

            GetCursorPos(&pt);

            if(!PtInRect(&rc, pt))
            {
                KillTimer(hWnd, TIMER_MOUSETRACK);

                bMouseInside = false;
                onMouseLeave(); 

                if (pTooltip)
                    pTooltip->apply("");

                PrefixedLSSetVariable(strName, "HoverTask", "");

                if (bUseHoverState || bUsePressedState || bScroll)
                {
                    for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
                    {
                        if ( bUseHoverState )
                            (*tbit)->bIsHover = false;

                        if ( bUsePressedState )
                            (*tbit)->bIsPressed = false;

                        if ( bScroll && (*tbit)->bIsScrolling )
                        {
                            (*tbit)->bIsScrolling = false;
                            bIsScrolling = false;
                            (*tbit)->iScrollPosition = 0;
                            KillTimer(hWnd, TIMER_SCROLL);
                        }
                    }

                    Repaint();
                }

                pCapture = NULL;

                break;
            }

            ScreenToClient(hWnd, &pt);
            TaskbarButton *button = ButtonAtPoint( pt.x, pt.y );

            if ( button && button != pCapture )
            {
                if (bShowToolTip && pTooltip)
                    pTooltip->apply(button->strTooltip.c_str());

                PrefixedLSSetVariable(strName, "HoverTask", ("\"" + button->GetWindowCaption() +"\""));

                if (bUseHoverState || bScroll)
                {
                    for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
                    {
                        if ( bUseHoverState )
                            (*tbit)->bIsHover = false;

                        if ( bScroll && (*tbit)->bIsScrolling )
                        {
                            (*tbit)->bIsScrolling = false;
                            bIsScrolling = false;
                            (*tbit)->iScrollPosition = 0;
                            KillTimer(hWnd, TIMER_SCROLL);
                        }
                    }

                    if ( bUseHoverState )
                        button->bIsHover = true;

                    if ( bScroll )
                    {
                        button->bIsScrolling = true;
                        bIsScrolling = true;
                        SetTimer(hWnd, TIMER_SCROLL, iScrollRefresh, NULL);
                    }

                    Repaint();  
                }
            }
            else if (button == NULL && pCapture != NULL)
            {
                if (bShowToolTip && pTooltip)
                    pTooltip->apply("");

                PrefixedLSSetVariable(strName, "HoverTask", "");

                if (bUseHoverState || bScroll)
                {
                    for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
                    {
                        if ( bUseHoverState )
                            (*tbit)->bIsHover = false;

                        if ( bScroll && (*tbit)->bIsScrolling )
                        {
                            (*tbit)->bIsScrolling = false;
                            bIsScrolling = false;
                            (*tbit)->iScrollPosition = 0;
                            KillTimer(hWnd, TIMER_SCROLL);
                        }
                    }

                    Repaint();
                }
            }

            pCapture = button;
        }
        break;

        case TIMER_ALPHAFADEIN:
        {
            if (iUpdateAlpha <= iAlphaTransparency-iAlphaFadeStep)
            {
                iUpdateAlpha += iAlphaFadeStep;
                UpdateWindowTransparency(iUpdateAlpha);
                ShowWindow(hWnd, SW_SHOWNOACTIVATE);
            }
            else
            {
                KillTimer(hWnd, TIMER_ALPHAFADEIN);
                if (iUpdateAlpha != iAlphaTransparency)
                    UpdateWindowTransparency(iAlphaTransparency);
            }
        }
        break;

        case TIMER_ALPHAFADEOUT:
        {
            if (iUpdateAlpha >= iAlphaFadeStep)
            {
                iUpdateAlpha -= iAlphaFadeStep;
                UpdateWindowTransparency(iUpdateAlpha);
            }
            else
            {
                KillTimer(hWnd, TIMER_ALPHAFADEOUT);
                iUpdateAlpha = 0;
                ShowWindow(hWnd, SW_HIDE);
            }
        }
        break;

        case TIMER_ALPHAFADEUPBANG:
        {
            if (iUpdateAlpha < iAlphaFadeBangEndValue)
            {
                iUpdateAlpha += iAlphaFadeStep;
                UpdateWindowTransparency(iUpdateAlpha);
            }
            else
            {
                KillTimer(hWnd, TIMER_ALPHAFADEUPBANG);
                if (iUpdateAlpha != iAlphaFadeBangEndValue)
                    UpdateWindowTransparency(iAlphaFadeBangEndValue);

                iAlphaTransparency = iAlphaFadeBangEndValue;
            }
        }
        
        break;

        case TIMER_ALPHAFADEDOWNBANG:
        {
            if (iUpdateAlpha > iAlphaFadeBangEndValue)
            {
                iUpdateAlpha -= iAlphaFadeStep;
                UpdateWindowTransparency(iUpdateAlpha);
            }
            else
            {
                KillTimer(hWnd, TIMER_ALPHAFADEDOWNBANG);
                if (iUpdateAlpha != iAlphaFadeBangEndValue)
                    UpdateWindowTransparency(iAlphaFadeBangEndValue);

                iAlphaTransparency = iAlphaFadeBangEndValue;
            }
        }

        break;

        case TIMER_DROPTIMEOUT:
        {
            KillTimer(hWnd, TIMER_DROPTIMEOUT);

            POINT pt;
            GetCursorPos( &pt );
            ScreenToClient( hWnd, &pt );

            TaskbarButton* button = ButtonAtPoint( pt.x, pt.y );
            if( button && button == pDropCapture )
            {
                if (button->iCurrentState != STATE_SELECTED)
                {
                    if ( button->iCurrentState == STATE_MINIMIZED )
                        SelectWindow(button->hWndButtonTask, true);
                    else
                        SelectWindow(button->hWndButtonTask);
                }
            }

            pDropCapture = NULL;
        }
        break;

        case TIMER_ACTIVETASK:
        {
            if ( !bInCriticalProcess && bVisible && pActiveTask != NULL )
            {
                int iOldState = pActiveTask->iCurrentState;

                if ( pActiveTask->GetCurrentState() != iOldState )
                    Repaint();
            }
        }
        break;

        case TIMER_SCROLL:
        {
            if (!bInCriticalProcess && bVisible)
            {
                if ( bScroll && bIsScrolling )
                {
                    for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
                    {
                        if ( (*tbit)->bIsScrolling )
                        {
                            (*tbit)->iScrollPosition -= iScrollSpeed;
                            break;
                        }
                    }

                    Repaint();
                }
            }
        }
        break;

        case TIMER_FLASH:
        {
            if (!bInCriticalProcess && bVisible)
            {
                ToggleFlashingButton();
                Repaint();
            }
        }
        break;
    }
}

void Taskbar::onWindowPosChanging(WINDOWPOS *windowPos)
{
    if ( !bNormalWindowzOrder && !(windowPos->x == -1 && windowPos->y == -1 && windowPos->cx == -1 && windowPos->cy == -1) && (windowPos->flags & SWP_NOZORDER) == 0 )
    {
        if ( !bAlwaysOnTop ) 
        {
            windowPos->flags = windowPos->flags | SWP_NOZORDER;
        }
        else if ( !strBehindWindow.empty() && hWndBehindWindow == NULL )
        {
            string one = "";
            string two = "";
            string three = "";
            string four = "";
            string five = "";
            string six = "";
            string seven = "";

            SplitString(strBehindWindow, one, two, three, four, five, six, seven);
        
            if ( _stricmp(one.c_str(), "HWND_TOP") == NULL )
                hWndBehindWindow = HWND_TOP;
            else
                hWndBehindWindow = FindWindowEx2(one.c_str(), two.c_str());

            if ( hWndBehindWindow != NULL )
            {
                windowPos->hwndInsertAfter = hWndBehindWindow;
            }
        }
        else if ( hWndBehindWindow != NULL )
        {
            windowPos->hwndInsertAfter = hWndBehindWindow;
        }
    }
}

void Taskbar::onMove(int x, int y)
{
    PrefixedLSSetVariable(strName, "CurrentX", IntToString( x ));
    PrefixedLSSetVariable(strName, "CurrentY", IntToString( y ));
    
    actualx = x;
    actualy = y;

    if ( !bTrueTransparency && !bAlphaMap )
        Repaint();

    if( !taskbarMove.empty() )
        LSExecute(hWnd, taskbarMove.c_str(), SW_SHOWNORMAL);
}

void Taskbar::onSize(int width, int height)
{
    PrefixedLSSetVariable(strName, "CurrentWidth", IntToString( width ));
    PrefixedLSSetVariable(strName, "CurrentHeight", IntToString( height ));

    actualwidth = width;
    actualheight = height;

    Repaint();

    //Set new Rect for ToolTip Tracker
    if ( bShowToolTip && pTooltip )
        pTooltip->apply("", true);
            
    if( !taskbarResize.empty() )
        LSExecute(hWnd, taskbarResize.c_str(), SW_SHOWNORMAL);
}

BOOL Taskbar::onWindowMessage( UINT nMessage, WPARAM wParam, LPARAM lParam, LRESULT &lResult )
{
    lResult = 0;

    switch ( nMessage )
    {
        case WM_CLOSE:
        {
            lResult = 0;
            return true;
        }

        case WM_DESTROY:
        {
            if (box != NULL)
            {
                SetParent(hWnd, NULL);
                box = NULL;

                if (pTooltip)
                {
                    Destroy_xTooltipClass(pTooltip);
                    pTooltip = NULL;
                }

                displayedtaskbuttons.clear();

                // release buttons
                for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
                {   
                    delete (*tbit);
                    (*tbit) = NULL;
                }
                taskbuttons.clear();

                if (hWnd)
                {
                    KillTimer(hWnd, TIMER_MOUSETRACK);
                    KillTimer(hWnd, TIMER_UPDATE);
                    KillTimer(hWnd, TIMER_ALPHAFADEIN);
                    KillTimer(hWnd, TIMER_ALPHAFADEOUT);
                    KillTimer(hWnd, TIMER_DROPTIMEOUT);
                    KillTimer(hWnd, TIMER_ACTIVETASK);
                    KillTimer(hWnd, TIMER_SCROLL);
                    KillTimer(hWnd, TIMER_FLASH);
                    KillTimer(hWnd, TIMER_SPECIALUPDATE);
                    KillTimer(hWnd, TIMER_ALPHAFADEUPBANG);
                    KillTimer(hWnd, TIMER_ALPHAFADEDOWNBANG);
                }
            
                Hide();
                
                bParentDestroyNotificationReceived = true;
            }           
            
            return false;
        }

        case WM_LBUTTONDOWN:
        {
            onLButtonDown( LOWORD(lParam), HIWORD(lParam));
            return true;
        }

        case WM_LBUTTONUP:
        {
            onLButtonUp( LOWORD(lParam), HIWORD(lParam));
            return true;
        }

        case WM_MBUTTONDOWN:
        {
            if ( bSkipButton )
                return true;

            onMButtonDown( LOWORD(lParam), HIWORD(lParam));
            return true;
        }

        case WM_MBUTTONUP:
        {
            if ( bSkipButton )
            {
                bSkipButton = false;
                return true;
            }

            onMButtonUp( LOWORD(lParam), HIWORD(lParam));
            return true;
        }

        case WM_RBUTTONDOWN:
        {
            onRButtonDown( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
            return true;
        }

        case WM_RBUTTONUP:
        {
            onRButtonUp( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
            return true;
        }

        case WM_MOUSEWHEEL:
        {
            if ((short)(HIWORD(wParam)) < 0)
            {
                onWheelDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
                return true;
            }
            else
            {
                onWheelUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
                return true;
            }
        }

        case WM_MOUSEMOVE:
        {
            if ( hWndToolTip )
            {
                MSG hintMessage = {hWnd, nMessage, wParam, lParam};
                SendMessage( hWndToolTip, (UINT) TTM_RELAYEVENT, 0, (LPARAM) (LPMSG) &hintMessage ); 
            }

            onMouseMove( LOWORD( lParam ), HIWORD( lParam ) );
            return true;
        }

        case LM_LSSELECT:
        {
            if ( bParentDestroyNotificationReceived )
                return true;

            onWindowActivated( (HWND) wParam, FALSE );
            return true;
        }

        case LM_REDRAW:
        {
            if ( bParentDestroyNotificationReceived )
                return true;

            onRedraw( (HWND) wParam, (BOOL) lParam ); 
            return true;
        }

        case LM_WINDOWACTIVATED:
        {
            if ( bParentDestroyNotificationReceived )
                return true;

            onWindowActivated( (HWND) wParam, ((lParam > 0) ? true : false) );
            return true;
        }

        case LM_WINDOWCREATED:
        {
            if ( bParentDestroyNotificationReceived )
                return true;

            const HWND changedhWnd = (HWND) wParam;

            if ( IsAppWindow(changedhWnd) || IsOnTaskbarList(changedhWnd, AddList, false) )
            {
                //Grouping
                if ( bUseGrouping )
                {
                    TaskbarButton *searchbutton = SearchForButtonByClass( changedhWnd );
                    if ( searchbutton != NULL )
                    {
                        // Not on RemoveList and on MatchingList (default is true, if empty)
                        if ( !IsOnTaskbarList(changedhWnd, RemoveList, false) && !IsOnTaskbarList(changedhWnd, RemoveMatchingList, false) && IsOnTaskbarList(changedhWnd, AddMatchingList, true) )
                        {
                            searchbutton->AddToGroup( changedhWnd );
                            pActiveTask = searchbutton;
                            Repaint();
                            return true;
                        }
                    }
                }
                
                //DoubleCheck, if not already there
                if ( SearchForButtonByHandle(changedhWnd) == NULL )
                {
                    // Not on RemoveList and on MatchingList (default is true, if empty)
                    if ( !IsOnTaskbarList(changedhWnd, RemoveList, false) && !IsOnTaskbarList(changedhWnd, RemoveMatchingList, false) && IsOnTaskbarList(changedhWnd, AddMatchingList, true) )
                    {
                        TaskbarButton *button = new TaskbarButton( this, changedhWnd );
                        taskbuttons.insert(taskbuttons.end(), button);

                        button->Update();

                        pActiveTask = button;
                        
                        AmountOfTasksChanged(button);

                        CreateLayout();

                        return true;
                    }
                }
            }

            return true;
        }

        case LM_WINDOWDESTROYED:
        {
            if ( bParentDestroyNotificationReceived )
                return true;

            const HWND changedhWnd = (HWND) wParam;
    
            //Grouping
            if (bUseGrouping)
            {
                for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
                {
                    if ( (*tbit)->listhWndGroup.size() > 1 && (*tbit)->RemoveFromGroup( changedhWnd ) )
                    {
                        if ( (*tbit)->IsFlashing() )
                        {
                            KillTimer(hWnd, TIMER_FLASH);
                            (*tbit)->SetFlashState(0);
                        }
                        Repaint();
                        return true;
                    }
                }
            }

            bool bDeleted = false;

            for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit )
            {   
                if ( (*tbit)->hWndButtonTask == changedhWnd )
                {
                    if ( (*tbit)->IsFlashing() )
                    {
                        KillTimer(hWnd, TIMER_FLASH);
                    }
                    delete (*tbit);
                    taskbuttons.erase( tbit );
                    
                    bDeleted = true;

                    break;
                }
            }

            if ( bDeleted )
            {
                AmountOfTasksChanged();

                CreateLayout();

                return true;
            }
                        
            return true;
        }

        case WM_PAINT:
        {
            PAINTSTRUCT ps;

            HDC hDC = BeginPaint( hWnd, &ps );

            onPaint( hDC );
            
            EndPaint( hWnd, &ps );

            return true;
        }

        case WM_WINDOWPOSCHANGING:
        {
            onWindowPosChanging( (WINDOWPOS *) lParam );
            return false;
        }

        case WM_MOVE:
        {
            onMove( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
            return true;
        }

        case WM_SIZE:
        {
            onSize( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
            return true;
        }

        case WM_TIMER:
        {
            onTimer((int) wParam);
            return true;
        }

        case MOUSEISDRAGGING:
        {
            if ( pDropCapture == NULL )
            {
                POINT pt;
                pt.x = GET_X_LPARAM( lParam );
                pt.y = GET_Y_LPARAM( lParam );

                ScreenToClient( hWnd, &pt );

                TaskbarButton* button = ButtonAtPoint( pt.x, pt.y );
                if( button )
                {
                    if ( iDropTimeOut != 0 )
                    {
                        pDropCapture = button;
                        SetTimer(hWnd, TIMER_DROPTIMEOUT, iDropTimeOut, 0);
                    }
                    else
                    {
                        if (button->iCurrentState != STATE_SELECTED)
                        {
                            if ( button->iCurrentState == STATE_MINIMIZED )
                                SelectWindow(button->hWndButtonTask, true);
                            else
                                SelectWindow(button->hWndButtonTask);
                        }
                    }
                }
            }
        }
    }

    return false;
}

TaskbarButton *Taskbar::ButtonAtPoint( int x, int y )
{
    for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
    {
        if ( (*tbit)->HasPoint(x, y) )
            return (*tbit);
    }

    return NULL;
}

void Taskbar::ToggleFlashingButton()
{
    for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
    {
        if ( (*tbit)->IsFlashing() )
            (*tbit)->SetFlashState(-1/*toggle*/);
    }
}

TaskbarButton *Taskbar::SearchForButtonByHandle( HWND hWndSearch )
{
    for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
    {
        if ( bUseGrouping )
        {
            for ( hWndGroupListIterator hglit = (*tbit)->listhWndGroup.begin(); hglit != (*tbit)->listhWndGroup.end(); ++hglit )
            {
                if ( hWndSearch == (*hglit) )
                    return (*tbit);
            }
                
        }
        else
        {
            if ( hWndSearch == (*tbit)->hWndButtonTask  )
                return (*tbit);
        }
    }
    
    return NULL;
}

//Actually first tries to get Exe-Filename before using Class
TaskbarButton *Taskbar::SearchForButtonByClass( HWND hWndSearch )
{
    for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
    {
        if ( _stricmp( (GetTaskFileName( hWndSearch )).c_str(), (*tbit)->strClassName.c_str()) == NULL )
            return (*tbit);
    }

    return NULL;
}

BOOL WINAPI Taskbar::EnumWindowsProc(HWND hWnd, LPARAM lParam)
{
    Taskbar *taskbar = (Taskbar *) lParam;

    if ( taskbar && ( IsAppWindow(hWnd) || IsOnTaskbarList(hWnd, taskbar->AddList, false) ) )
    {
        bool normalMode = true;
        //Grouping
        if ( taskbar->bUseGrouping )
        {
            TaskbarButton *searchbutton = taskbar->SearchForButtonByClass( hWnd );
            if ( searchbutton != NULL )
            {
                searchbutton->AddToGroup( hWnd );
                if ( taskbar->pActiveTask == NULL )
                    taskbar->pActiveTask = searchbutton;

                normalMode = false;
            }
        }

        if ( normalMode) 
        {
            //DoubleCheck, if not already there
            if ( taskbar->SearchForButtonByHandle(hWnd) == NULL )
            {
                // Not on general RemoveList, not on custom RemoveList and on MatchingList (default is true)
                if ( !IsOnTaskbarList(hWnd, taskbar->RemoveList, false) && !IsOnTaskbarList(hWnd, taskbar->RemoveMatchingList, false) && IsOnTaskbarList(hWnd, taskbar->AddMatchingList, true) )
                {
                    TaskbarButton *button = new TaskbarButton( taskbar, hWnd );
                    taskbar->taskbuttons.insert(taskbar->taskbuttons.end(), button);

                    button->Update();

                    if ( taskbar->pActiveTask == NULL )
                        taskbar->pActiveTask = button;
                }
            }
        }
    }
    return TRUE;
}

LRESULT WINAPI Taskbar::WindowProcedure(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    Taskbar *taskbar = (Taskbar *) GetWindowLong(hWnd, 0);

    if (message == WM_NCCREATE)
    {
        LPCREATESTRUCT pcs = (LPCREATESTRUCT) lParam;
        taskbar = (Taskbar *) pcs->lpCreateParams;

        taskbar->hWnd = hWnd;
        SetWindowLong(hWnd, 0, (LONG) taskbar);
    }

    if (taskbar)
    {
        LRESULT lResult;

        if (taskbar->onWindowMessage(message, wParam, lParam, lResult))
            return lResult;
    }

    return DefWindowProc(hWnd, message, wParam, lParam);
}

bool Taskbar::AutoSizeFunction(int buttons)
{   
    if ( buttons < 0 )
        buttons = 0;

    RECT rectPosition;
    GetWindowRect(hWnd, &rectPosition);
        
    int width = (rectPosition.right-rectPosition.left);
    int height = (rectPosition.bottom-rectPosition.top);

    int hcounter = 0;
    int vcounter = 0;

    POINT boxConverted = {rectPosition.left, rectPosition.top};
    if (box)
        ScreenToClient(box, &boxConverted);

    int autoX = boxConverted.x;
    int autoY = boxConverted.y;
    int autoWidth = width;
    int autoHeight = height;

    int realwrapCount = iWrapCount+iWrapCountAdded;

    if (iLayout == 0)
    {
        if ( iWrapCount != -1 )
        {
            hcounter = buttons%realwrapCount;   
            vcounter = min(buttons/realwrapCount, iMaxLines_Rows);
        }
        else
        {
            hcounter = buttons;
        }
        
        if (iWrapCount == -1 || ( iWrapCount != -1 && buttons <= realwrapCount ) )
        {
            autoWidth = iLeftBorder+(buttons-1)*iXSpacing+buttons*iMaxTaskWidth+iRightBorder;
            autoHeight = iTopBorder+iMaxTaskHeight+iBottomBorder;
        }
        else if ( vcounter >= 1 )
        {
            autoWidth = iLeftBorder+(realwrapCount-1)*iXSpacing+realwrapCount*iMaxTaskWidth+iRightBorder;
            
            if ( buttons%realwrapCount == 0 || vcounter >= iMaxLines_Rows )
            {
                autoHeight = iTopBorder+(vcounter-1)*iYSpacing+vcounter*iMaxTaskHeight+iBottomBorder;
            }
            else
            {
                autoHeight = iTopBorder+vcounter*iYSpacing+(vcounter+1)*iMaxTaskHeight+iBottomBorder;
            }
        }

        autoWidth = min(autoWidth, iMaxTaskbarSize);

        if ( iDirection == 2 )
        {
            autoX = actualx-(autoWidth-width);
        }
        if ( iWrapDirection == 2 )
        {
            autoY = actualy-(autoHeight-height);
        }
            
        if ( actualx != autoX || actualy != autoY || actualwidth != autoWidth || actualheight != autoHeight )
        {   
            Reposition(3, autoX, autoY, autoWidth, autoHeight, iAutoSizeSteps, iAutoSizeDelay);
        }
        else if ( actualwidth != autoWidth || actualheight != autoHeight )
        {   
            Reposition(2, 0, 0, autoWidth, autoHeight, iAutoSizeSteps, iAutoSizeDelay);
        }
        else
        {
            pDontPaintAddedTask = NULL;
            CreateLayout();
            return false;
        }
    }
    else if (iLayout == 1)
    {
        if ( iWrapCount != -1 )
        {
            vcounter = buttons%realwrapCount;
            hcounter = min(buttons/realwrapCount, iMaxLines_Rows);
        }
        else
        {
            vcounter = buttons;
        }

        if (iWrapCount == -1 || ( iWrapCount != -1 && buttons <= realwrapCount ) )
        {
            autoWidth = iLeftBorder+iMaxTaskWidth+iRightBorder;
            autoHeight = iTopBorder+(buttons-1)*iYSpacing+buttons*iMaxTaskHeight+iBottomBorder;
        }
        else if ( hcounter >= 1 )
        {
            if ( buttons%realwrapCount == 0 || hcounter >= iMaxLines_Rows)
            {
                autoWidth = iLeftBorder+(hcounter-1)*iXSpacing+hcounter*iMaxTaskWidth+iRightBorder;
            }
            else
            {
                autoWidth = iLeftBorder+hcounter*iXSpacing+(hcounter+1)*iMaxTaskWidth+iRightBorder;
            }
            autoHeight = iTopBorder+(realwrapCount-1)*iYSpacing+realwrapCount*iMaxTaskHeight+iBottomBorder;
        }

        autoHeight = min(autoHeight, iMaxTaskbarSize);

        if ( iDirection == 4 )
        {
            autoY = actualy-(autoHeight-height);
        }
        if ( iWrapDirection == 4 )
        {
            autoX = actualx-(autoWidth-width);
        }

        if ( actualx != autoX || actualy != autoY || actualwidth != autoWidth || actualheight != autoHeight )
        {   
            Reposition(3, autoX, autoY, autoWidth, autoHeight, iAutoSizeSteps, iAutoSizeDelay);
        }
        else if ( actualwidth != autoWidth || actualheight != autoHeight )
        {   
            Reposition(2, 0, 0, autoWidth, autoHeight, iAutoSizeSteps, iAutoSizeDelay);
        }
        else
        {
            pDontPaintAddedTask = NULL;
            CreateLayout();
            return false;
        }
    }

    return true;
}

void Taskbar::AmountOfTasksChanged(TaskbarButton *button)
{
    int iVisibleTasks = CheckDisplayedTasks();

    int iOldBackupExportedButtonCount = iBackupExportedButtonCount;
    int iVisibleTaskChanged = iVisibleTasks - iBackupExportedButtonCount;

    if ( iVisibleTaskChanged > 0 )
    {
        if ( button != NULL 
            && ( (!taskbarAdd.empty() && (StrStrI(taskbarAdd.c_str(), "xtaskbarresize") != NULL || StrStrI(taskbarAdd.c_str(), "xtaskbarreposition") != NULL)) )
                    || (!taskbarLineAdd.empty() && (StrStrI(taskbarLineAdd.c_str(), "xtaskbarresize") != NULL || StrStrI(taskbarLineAdd.c_str(), "xtaskbarreposition") != NULL)) 
                    || bAutoSize )
            pDontPaintAddedTask = button;

        while ( iVisibleTaskChanged > 0 )
        {
            iOldBackupExportedButtonCount++;
            PrefixedLSSetVariable(strName, "CurrentButtonCount", IntToString( iOldBackupExportedButtonCount ));
            if ( !taskbarAdd.empty() )
                LSExecute(hWnd, taskbarAdd.c_str(), SW_SHOWNORMAL);

            if ( iWrapCount != -1 )
            {
                int realwrapCount = iWrapCount+iWrapCountAdded;
                int availableButtonCount = iOldBackupExportedButtonCount;

                if ( availableButtonCount/realwrapCount >= iLines_Rows+iLines_RowsAdded && availableButtonCount%realwrapCount == 1 )
                {
                    iLines_RowsAdded++;
                    if ( iLines_Rows+iLines_RowsAdded <= iMaxLines_Rows )
                    {   
                        PrefixedLSSetVariable(strName, "CurrentLines", IntToString( iLines_Rows+iLines_RowsAdded ));
                        if ( !taskbarLineAdd.empty() )
                            LSExecute(hWnd, taskbarLineAdd.c_str(), SW_SHOWNORMAL);
                    }
                    else if ( iLines_Rows+iLines_RowsAdded > iMaxLines_Rows)
                    {
                        iLines_RowsAdded--;
                        iWrapCountAdded++;
                    }
                }
            }

            iVisibleTaskChanged--;
        }
    }
    else if ( iVisibleTaskChanged < 0 )
    {
        while ( iVisibleTaskChanged < 0 )
        {
            iOldBackupExportedButtonCount--;
            PrefixedLSSetVariable(strName, "CurrentButtonCount", IntToString( iOldBackupExportedButtonCount ));
            if ( !taskbarRemove.empty() )
                LSExecute(hWnd, taskbarRemove.c_str(), SW_SHOWNORMAL);

            if ( iWrapCount != -1 )
            {
                int realwrapCount = iWrapCount+iWrapCountAdded;
                int availableButtonCount = iOldBackupExportedButtonCount;

                ///Check for added iWrapCountAdded
                if ( iWrapCountAdded > 0 )
                {
                    if ( availableButtonCount/iWrapCount >= iMaxLines_Rows && availableButtonCount%(iWrapCount+iWrapCountAdded-1) == 0 && (iLines_Rows+iLines_RowsAdded)*(iWrapCount+iWrapCountAdded-1) == availableButtonCount )
                    {
                        iWrapCountAdded--;
                        iVisibleTaskChanged++;
                        continue;
                    }
                }

                if ( iLines_RowsAdded >= 1 && availableButtonCount/realwrapCount >= iLines_Rows && availableButtonCount%realwrapCount == 0 )
                {
                    iLines_RowsAdded--;

                    PrefixedLSSetVariable(strName, "CurrentLines", IntToString( min(iLines_Rows+iLines_RowsAdded, iMaxLines_Rows) ));
                    if ( !taskbarLineRemove.empty() )
                        LSExecute(hWnd, taskbarLineRemove.c_str(), SW_SHOWNORMAL);
                }
            }

            iVisibleTaskChanged++;
        }
    }

    iBackupExportedButtonCount = iVisibleTasks;

    PrefixedLSSetVariable(strName, "CurrentButtonCount", IntToString( iBackupExportedButtonCount ));
    PrefixedLSSetVariable(strName, "CurrentLines", IntToString( min(iLines_Rows+iLines_RowsAdded, iMaxLines_Rows) ));   

    if ( bAutoSize )
        AutoSizeFunction(iBackupExportedButtonCount);
}

int Taskbar::CheckDisplayedTasks()
{
    int buttonCount = 0;

    displayedtaskbuttons.clear();

    if ( iDisplayButtonType == -1 && !bVWMCurrentDesktopOnly && !(iRealMonitors > 1 && iThisMultiMonitorOnly != -1) )
    {
        for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
        {
            buttonCount++;
            (*tbit)->iPositionInTaskbar = buttonCount;
            displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
        }
    }
    else
    {
        for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
        {
            (*tbit)->GetCurrentState();

            if ( iDisplayButtonType == 6 )
            {
                if ( (*tbit)->IsFlashing() )
                {
                    if ( bVWMCurrentDesktopOnly )
                    {
                        if ( IsValidWindow((*tbit)->hWndButtonTask, iVWMSpecifiedDesktopOnly) )
                        {
                            if ( iRealMonitors > 1 && iThisMultiMonitorOnly != -1 )
                            {
                                if ( (*tbit)->bMonitorHidden )
                                {
                                    (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                                }
                                else
                                {
                                    buttonCount++;
                                    (*tbit)->iPositionInTaskbar = buttonCount;
                                    displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                                }
                            }
                            else
                            {
                                buttonCount++;
                                (*tbit)->iPositionInTaskbar = buttonCount;
                                displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                            }
                        }
                        else
                        {
                            (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                        }
                    }
                    else if ( iRealMonitors > 1 && iThisMultiMonitorOnly != -1 )
                    {
                        if ( (*tbit)->bMonitorHidden )
                        {
                            (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                        }
                        else
                        {
                            buttonCount++;
                            (*tbit)->iPositionInTaskbar = buttonCount;
                            displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                        }
                    }
                    else
                    {
                        buttonCount++;
                        (*tbit)->iPositionInTaskbar = buttonCount;
                        displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                    }
                }
                else
                {
                    (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                }
            }
            else if ((*tbit)->iCurrentState == STATE_NORMAL 
                    && (iDisplayButtonType < 0 || iDisplayButtonType == 0 || iDisplayButtonType == 3 || iDisplayButtonType == 4))
            {
                if ( bVWMCurrentDesktopOnly )
                {
                    if ( IsValidWindow((*tbit)->hWndButtonTask, iVWMSpecifiedDesktopOnly) )
                    {
                        if ( iRealMonitors > 1 && iThisMultiMonitorOnly != -1 )
                        {
                            if ( (*tbit)->bMonitorHidden )
                            {
                                (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                            }
                            else
                            {
                                buttonCount++;
                                (*tbit)->iPositionInTaskbar = buttonCount;
                                displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                            }
                        }
                        else
                        {
                            buttonCount++;
                            (*tbit)->iPositionInTaskbar = buttonCount;
                            displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                        }
                    }
                    else
                    {
                        (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                    }
                }
                else if ( iRealMonitors > 1 && iThisMultiMonitorOnly != -1 )
                {
                    if ( (*tbit)->bMonitorHidden )
                    {
                        (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                    }
                    else
                    {
                        buttonCount++;
                        (*tbit)->iPositionInTaskbar = buttonCount;
                        displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                    }
                }
                else
                {
                    buttonCount++;
                    (*tbit)->iPositionInTaskbar = buttonCount;
                    displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                }
            }
            else if ((*tbit)->iCurrentState == STATE_SELECTED 
                && (iDisplayButtonType < 0 || iDisplayButtonType == 1 || iDisplayButtonType == 3 || iDisplayButtonType == 5))
            {
                if ( bVWMCurrentDesktopOnly )
                {
                    if ( IsValidWindow((*tbit)->hWndButtonTask, iVWMSpecifiedDesktopOnly) )
                    {
                        if ( iRealMonitors > 1 && iThisMultiMonitorOnly != -1 )
                        {
                            if ( (*tbit)->bMonitorHidden )
                            {
                                (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                            }
                            else
                            {
                                buttonCount++;
                                (*tbit)->iPositionInTaskbar = buttonCount;
                                displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                            }
                        }
                        else
                        {
                            buttonCount++;
                            (*tbit)->iPositionInTaskbar = buttonCount;
                            displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                        }
                    }
                    else
                    {
                        (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                    }
                }
                else if ( iRealMonitors > 1 && iThisMultiMonitorOnly != -1 )
                {
                    if ( (*tbit)->bMonitorHidden )
                    {
                        (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                    }
                    else
                    {
                        buttonCount++;
                        (*tbit)->iPositionInTaskbar = buttonCount;
                        displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                    }
                }
                else
                {
                    buttonCount++;
                    (*tbit)->iPositionInTaskbar = buttonCount;
                    displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                }
            }
            else if ((*tbit)->iCurrentState == STATE_MINIMIZED 
                    && (iDisplayButtonType < 0 || iDisplayButtonType == 2 || iDisplayButtonType == 4 || iDisplayButtonType == 5))
            {
                if ( bVWMCurrentDesktopOnly )
                {
                    if ( IsValidWindow((*tbit)->hWndButtonTask, iVWMSpecifiedDesktopOnly) )
                    {
                        if ( iRealMonitors > 1 && iThisMultiMonitorOnly != -1 )
                        {
                            if ( (*tbit)->bMonitorHidden )
                            {
                                (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                            }
                            else
                            {
                                buttonCount++;
                                (*tbit)->iPositionInTaskbar = buttonCount;
                                displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                            }
                        }
                        else
                        {
                            buttonCount++;
                            (*tbit)->iPositionInTaskbar = buttonCount;
                            displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                        }
                    }
                    else
                    {
                        (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                    }
                }
                else if ( iRealMonitors > 1 && iThisMultiMonitorOnly != -1 )
                {
                    if ( (*tbit)->bMonitorHidden )
                    {
                        (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                    }
                    else
                    {
                        buttonCount++;
                        (*tbit)->iPositionInTaskbar = buttonCount;
                        displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                    }
                }
                else
                {
                    buttonCount++;
                    (*tbit)->iPositionInTaskbar = buttonCount;
                    displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
                }
            }
            else
            {
                (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
            }
        }
    }

    return min(buttonCount, iMaxButtonCount);
}


void Taskbar::CreateLayout(int width, int height)
{
    if ( bCreateLayoutLocked || bInCriticalProcess )
        return;

    bInCriticalProcess = true;

    if (width == -1 && height == -1)
    {
        RECT rectPosition;
        GetClientRect(hWnd, &rectPosition);
            
        width = rectPosition.right;
        height = rectPosition.bottom;
    }

    int buttonCount = CheckDisplayedTasks();

    PrefixedLSSetVariable(strName, "CurrentButtonCount", IntToString( buttonCount ));
    PrefixedLSSetVariable(strName, "CurrentLines", IntToString( iLines_Rows+iLines_RowsAdded ));

    if (buttonCount <= 0)
    {
        buttonCount = 0;
        iLines_RowsAdded = 0;
        iWrapCountAdded = 0;

        displayedtaskbuttons.clear();

        bInCriticalProcess = false;

        Repaint();

        if ( bHideIfEmpty && !bAutoHidden && bVisible )
        {
            bAutoHidden = true;
            Hide();
        }

        if ( !taskbarEmpty.empty() )    
            LSExecute(hWnd, taskbarEmpty.c_str(), SW_SHOWNORMAL);

        iBackupExportedButtonCount = buttonCount;

        return;
    }
    else if ( bHideIfEmpty && bAutoHidden && !bVisible )
    {
        bAutoHidden = false;
        Show();

        if ( !taskbarFilled.empty() )   
            LSExecute(hWnd, taskbarFilled.c_str(), SW_SHOWNORMAL);
    }
    else if ( !taskbarFilled.empty() )
    {
        LSExecute(hWnd, taskbarFilled.c_str(), SW_SHOWNORMAL);
    }

    if ( iAutoSort > 0 )
        SortTasks(displayedtaskbuttons);

    if (iLayout == 0)
    {
        int realwrapCount = iWrapCount+iWrapCountAdded;
    
        int availableButtonCount = buttonCount;
        //Start Horizontal Version
        //-----------------------------------------------------------------------
        int maxavailableWidth = max( width-iLeftBorder-iRightBorder, 0 );

        int realLines = min(iLines_Rows+iLines_RowsAdded, iMaxLines_Rows);

        if ( availableButtonCount >= 1 )
        {
            if ( iWrapCount != -1 && buttonCount < realwrapCount )
                maxavailableWidth = maxavailableWidth - (buttonCount-1) * iXSpacing;
            else if ( iWrapCount != -1 )
                maxavailableWidth = maxavailableWidth - (realwrapCount-1) * iXSpacing;
            else
                maxavailableWidth = maxavailableWidth - (availableButtonCount-1) * iXSpacing;
        }

        int buttonWidth = 0;
        if ( iWrapCount != -1 )
        {
            if ( buttonCount < realwrapCount )
            {
                if ( maxavailableWidth%buttonCount == 0 )
                    buttonWidth = maxavailableWidth/buttonCount;
                else
                    buttonWidth = (maxavailableWidth/buttonCount)+1;
            }
            else
            {
                if ( maxavailableWidth%realwrapCount == 0 )
                    buttonWidth = maxavailableWidth/realwrapCount;
                else
                    buttonWidth = (maxavailableWidth/realwrapCount)+1;
            }

            if ( !bUseAdaptiveSize || bAutoSize )
                buttonWidth = min(buttonWidth, iMaxTaskWidth);
        }
        else
        {
            if ( maxavailableWidth%availableButtonCount == 0 )
                buttonWidth = maxavailableWidth/availableButtonCount;
            else
                buttonWidth = (maxavailableWidth/availableButtonCount)+1;

            buttonWidth = min(buttonWidth, iMaxTaskWidth);
        }
        
        int buttonHeight = min( (height - iTopBorder - iBottomBorder - (realLines - 1) * iYSpacing)/realLines, iMaxTaskHeight);
    
        PrefixedLSSetVariable(strName, "CurrentButtonWidth", IntToString( buttonWidth ));
        PrefixedLSSetVariable(strName, "CurrentButtonHeight", IntToString( buttonHeight ));

        int x = 0;
        int y = 0;

        int buttoncounter = 1;
        int hcounter = 1;
        int vcounter = 1;

        int scrollcounter = 0;
        // set placement for each button
        for (TaskbarButtonIterator tbit = displayedtaskbuttons.begin(); tbit != displayedtaskbuttons.end(); ++tbit)
        {
            if ( buttonWidth > 0 && buttoncounter <= iMaxButtonCount )
            {
                if ( displayedtaskbuttons.size() > iMaxButtonCount && iStartWithPosition > scrollcounter )
                {
                    scrollcounter++;
                    (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                    continue;
                }
                else
                {
                    scrollcounter++;
                }

                if ( iDirection == 1 ) //Right
                    x = iLeftBorder + (hcounter-1) * (buttonWidth + iXSpacing);
                else if ( iDirection == 2 ) // Left
                    x = (width - iRightBorder) - hcounter * buttonWidth - (hcounter-1) * iXSpacing;
                
                if ( iWrapDirection == 1 ) //Bottom
                    y = iTopBorder + (vcounter-1) * (buttonHeight + iYSpacing);
                else if ( iWrapDirection == 2 ) //Top
                    y = (height - iBottomBorder) - vcounter * buttonHeight - (vcounter-1) * iYSpacing;

                //Resize Button, if required
                if ( x+buttonWidth > width-iRightBorder )
                    (*tbit)->SetButtonLayout( x, y, width-iRightBorder-x, buttonHeight );
                else if ( x < iLeftBorder )
                    (*tbit)->SetButtonLayout( iLeftBorder, y, buttonWidth-(iLeftBorder-x), buttonHeight );
                else
                    (*tbit)->SetButtonLayout( x, y, buttonWidth, buttonHeight );

                if ( iWrapCount != -1 && buttoncounter%realwrapCount == 0 )
                {
                    hcounter = 0;
                    vcounter++;
                }
                hcounter++;
                buttoncounter++;
            }
            else
            {
                (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
            }
        }
        //-----------------------------------------------------------------------
        //End Horizontal Version
    }
    else if (iLayout == 1)
    {
        int realwrapCount = iWrapCount+iWrapCountAdded;

        int availableButtonCount = buttonCount;
        //Start Vertical Version
        //-----------------------------------------------------------------------
        int maxavailableHeight = max( height-iTopBorder-iBottomBorder, 0 );
    
        int realLines = min(iLines_Rows+iLines_RowsAdded, iMaxLines_Rows);

        if ( availableButtonCount >= 1 )
        {
            if ( iWrapCount != -1 && buttonCount < realwrapCount )
                maxavailableHeight = maxavailableHeight - (buttonCount-1) * iYSpacing;
            else if ( iWrapCount != -1 )
                maxavailableHeight = maxavailableHeight - (realwrapCount-1) * iYSpacing;
            else
                maxavailableHeight = maxavailableHeight - (availableButtonCount-1) * iYSpacing;
        }

        int buttonHeight = 0;
        if ( iWrapCount != -1 )
        {
            if ( buttonCount < realwrapCount )
            {
                if ( maxavailableHeight%buttonCount == 0 )
                    buttonHeight = maxavailableHeight/buttonCount;
                else
                    buttonHeight = (maxavailableHeight/buttonCount)+1;
            }
            else
            {
                if ( maxavailableHeight%realwrapCount == 0 )
                    buttonHeight = maxavailableHeight/realwrapCount;
                else
                    buttonHeight = (maxavailableHeight/realwrapCount)+1;
            }

            if ( !bUseAdaptiveSize || bAutoSize )
                buttonHeight = min(buttonHeight, iMaxTaskHeight);
        }
        else
        {
            if ( maxavailableHeight%availableButtonCount == 0 )
                buttonHeight = maxavailableHeight/availableButtonCount;
            else
                buttonHeight = (maxavailableHeight/availableButtonCount)+1;

            buttonHeight = min(buttonHeight, iMaxTaskHeight);
        }
    
        int buttonWidth = min( (width - iLeftBorder - iRightBorder - (realLines - 1) * iXSpacing)/realLines, iMaxTaskWidth);

        PrefixedLSSetVariable(strName, "CurrentButtonWidth", IntToString( buttonWidth ));
        PrefixedLSSetVariable(strName, "CurrentButtonHeight", IntToString( buttonHeight ));

        int x = 0;
        int y = 0;

        int buttoncounter = 1;
        int hcounter = 1;
        int vcounter = 1;

        int scrollcounter = 0;
        // set placement for each button
        for (TaskbarButtonIterator tbit = displayedtaskbuttons.begin(); tbit != displayedtaskbuttons.end(); ++tbit)
        {   
            if ( buttonWidth > 0 && buttoncounter <= iMaxButtonCount )
            {
                if ( displayedtaskbuttons.size() > iMaxButtonCount && iStartWithPosition > scrollcounter )
                {
                    scrollcounter++;
                    (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
                    continue;
                }
                else
                {
                    scrollcounter++;
                }

                if ( iDirection == 3 ) //bottom
                    y = iTopBorder + (vcounter-1) * (buttonHeight + iYSpacing);
                else if ( iDirection == 4 )  //top
                    y = (height - iBottomBorder) - vcounter * buttonHeight - (vcounter-1) * iYSpacing;
        
                if ( iWrapDirection == 3 ) //right
                    x = iLeftBorder + (hcounter-1) * (buttonWidth + iXSpacing);
                else if ( iWrapDirection == 4 ) //left
                    x = (width - iRightBorder) - hcounter * buttonWidth - (hcounter-1) * iXSpacing;

                //Make Smaller if required
                if ( y+buttonHeight > height-iBottomBorder )
                    (*tbit)->SetButtonLayout( x, y, buttonWidth, height-iBottomBorder-y );
                else if ( y < iTopBorder )
                    (*tbit)->SetButtonLayout( x, iTopBorder, buttonWidth, buttonHeight-(iTopBorder-y) );
                else
                    (*tbit)->SetButtonLayout( x, y, buttonWidth, buttonHeight );

                if ( iWrapCount != -1 && buttoncounter%realwrapCount == 0 )
                {
                    hcounter++;
                    vcounter = 0;
                }
                vcounter++;
                buttoncounter++;
            }
            else
            {
                (*tbit)->SetButtonLayout( -1, -1, -1, -1 );
            }
        }
        //-----------------------------------------------------------------------
        //End Vertical Version
    }

    bInCriticalProcess = false;

    Repaint();

    iBackupExportedButtonCount = buttonCount;
}

void Taskbar::onRedraw( HWND hApplication, bool fFlash )
{
    TaskbarButton *searchbutton = SearchForButtonByHandle( hApplication );
    if ( searchbutton != NULL )
    {
        if( GetActiveWindow() != searchbutton->hWndButtonTask )
        {
            if (fFlash)
            {
                if ( !taskbarFlash.empty() )
                    LSExecute(hWnd, taskbarFlash.c_str(), SW_SHOWNORMAL);
                
                if ( bUseFlashingState || iDisplayButtonType == 6 )
                    searchbutton->SetFlashState(1);

                if ( iDisplayButtonType == 6 )
                {
                    AmountOfTasksChanged();

                    CreateLayout();
                }
                else if ( bUseFlashingState )
                {
                    Repaint();
                }

                if ( bUseFlashingState )
                    SetTimer(hWnd, TIMER_FLASH, iFlashRefresh, NULL);   
            }
        }
    }
}

void Taskbar::onWindowActivated( HWND hApplication, BOOL fFullScreen )
{
    if ( iAutoSort == -1 )
    {
        displayedtaskbuttons.clear();

        // release buttons
        for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
        {   
            delete (*tbit);
            (*tbit) = NULL;
        }
        taskbuttons.clear();

        EnumWindows(Taskbar::EnumWindowsProc, (LPARAM) this);
    }

    TaskbarButton *searchbutton = SearchForButtonByHandle( hApplication );
    if ( searchbutton != NULL )
    {
        //Special In UseGrouping
        if ( bUseGrouping )
            searchbutton->hWndButtonTask = hApplication;

        if ( searchbutton->IsFlashing() )
        {
            KillTimer(hWnd, TIMER_FLASH);
            searchbutton->SetFlashState(0);
        }

        searchbutton->iCurrentState = GetState(searchbutton->hWndButtonTask);

        pActiveTask = searchbutton;
    }
    
    if ( iAutoSort == -1 || iDisplayButtonType >= 0 )
    {
        if ( iDisplayButtonType >= 0 )
            AmountOfTasksChanged();

        CreateLayout();
    }
    else
    {
        Repaint();
    }
}

void Taskbar::SortTasks(vector<TaskbarButton *> &sorttaskbuttons)
{
    std::sort( sorttaskbuttons.begin(), sorttaskbuttons.end(), CompareTasks );
}
