/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2009 The LS-Universe Team

    Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#if !defined(AFX_TASKBAR_H__EBF1ABD2_1EA7_40AF_93CF_D4F2B4ECC8EE__INCLUDED_)
#define AFX_TASKBAR_H__EBF1ABD2_1EA7_40AF_93CF_D4F2B4ECC8EE__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "TaskbarButton.h"

class DropTarget;

class TaskbarButton;

class GhostButton;
class GhostTaskbar;

//Alpha Transparency
#ifndef WS_EX_LAYERED
    #define WS_EX_LAYERED           0x00080000L
    #define LWA_COLORKEY            0x00000001L
    #define ULW_COLORKEY            0x00000001L
    #define LWA_ALPHA               0x00000002L
    #define ULW_ALPHA               0x00000002L
    #define ULW_OPAQUE              0x00000003L
#endif // WS_EX_LAYERED

#ifndef AC_SRC_ALPHA
    #define AC_SRC_ALPHA            0x00000001L
#endif

#ifndef WM_MOUSEWHEEL
    #define WM_MOUSEWHEEL 0x020A
#endif

typedef WINUSERAPI BOOL (WINAPI *FUNC_BOOL__DWORD)(DWORD);
#ifndef ASFW_ANY
    #define ASFW_ANY ((DWORD)-1)
#endif

typedef vector<TaskbarButton *>::iterator TaskbarButtonIterator;

typedef vector<GhostButton *>GhostButtonList;
typedef vector<GhostButton *>::iterator GhostButtonListIterator;

typedef vector<GhostTaskbar *>GhostTaskbarList;
typedef vector<GhostTaskbar *>::iterator GhostTaskbarListIterator;

class Taskbar  
{

public:
    Taskbar(const string &name);
    virtual ~Taskbar();

    friend class TaskbarButton;

    void Load(HINSTANCE hInstance, HWND setbox = 0);
    void Reconfigure(bool refresh = false);

    const HWND &GethWnd() const { return hWnd; }
    const string &GetName() const { return strName; }
    const HWND &GetBox() const { return box; }
    bool IsAlwaysOnTop() const { return bAlwaysOnTop; }
    bool IsVisible() const { return bVisible; }
    bool IsStartHidden() const { return bStartHidden; }
    bool IsAlphaMap() const { return bAlphaMap; }
    bool IsLayoutLocked() const { return bCreateLayoutLocked; }
    bool IsShowNotAllowed() const { return (bHideIfEmpty && iBackupExportedButtonCount == 0 ? true : false); }
    bool IsHideNotAllowed() const { return (bHideIfEmpty && iBackupExportedButtonCount != 0 ? true : false); }

    void Reposition(int type, int newx, int newy, int newwidth, int newheight, int steps = 0, int time = 0, bool repositionBy = false);

    void SetAlwaysOnTop(bool bNewAlwaysOnTop, bool bFirstTime = false);

    void SetAlpha(int newAlphaTransparency);
    void SetAlphaFade(int newAlphaTransparency, int step, int delay);

    void SetDisplayButtonType(int setiDisplayButtonType) { iDisplayButtonType = setiDisplayButtonType; }

    void SwitchToButton( LPCSTR direction );
    void ScrollButtons(bool next, int amount);

    void Update(bool needrepaint = false);
    void Repaint();
    void UpdateRegion();

    void Show();    
    void Hide();

    void CreateLayout(int width = -1, int height = -1);
        
private:

    DropTarget* pDropTarget;

    StringList AddList;
    StringList RemoveList;
    StringList CustomIconList;
    StringList ChangingIconList;

    StringList AddMatchingList;
    StringList RemoveMatchingList;

    //-----------------------------------------------
    // Essential Settings
    //-----------------------------------------------

    string strName;

    HWND hWnd;
    HWND hWndOwner;
    HINSTANCE hInstance;
    HWND box;

    int actualx;
    int actualy;
    int actualwidth;
    int actualheight;

    bool bTrueTransparency;

    bool bStartHidden;

    bool bAlwaysOnTop;

    bool bNormalWindowzOrder;

    string strBehindWindow;
    HWND hWndBehindWindow;

    //-----------------------------------------------
    // Settings
    //-----------------------------------------------

    bool bFocusOnEnter;

    bool bHideIfEmpty;

    bool bActionsOnDownButton;

    int iDropTimeOut;
    int iTimerRefresh;
    int iTrackingRefresh;
    int iFlashRefresh;
    int iSpecialUpdate;

    int iDisplayButtonType;

    int iSpecialHoverState;
    
    bool bNoMinimizeOnClick;

    bool bUseBigNormalIcon;
    bool bUseBigActiveIcon;
    bool bUseBigMinimizedIcon;
    bool bUseBigFlashingIcon;
    bool bUseBigHoverIcon;
    bool bUseBigPressedIcon;
    bool bUseBigGroupedIcon;

    bool bMultiLineUsed;
    bool bUseCustomizedCaption;
    string strNormalText;
    string strActiveText;
    string strMinimizedText;
    string strFlashingText;
    string strHoverText;
    string strPressedText;
    string strGroupedText;
    
    bool bShowIcon;

    bool bShowText;

    int iAutoSort;
    
    int iThisMultiMonitorOnly;

    //Special Font and Icon Settings
    bool bUseGroupedFontSettings;
    bool bUseHoverFontSettings;
    bool bUsePressedFontSettings;

    bool bUseGroupedIconSettings;
    bool bUseHoverIconSettings;
    bool bUsePressedIconSettings;

    bool bUseGrouping;
    bool bUseHoverState;
    bool bUsePressedState;
    bool bUseFlashingState;

    bool bScroll;
    int iScrollPadLength;
    int iScrollRefresh;
    int iScrollSpeed;

    bool bVWMCurrentDesktopOnly;
    int iVWMSpecifiedDesktopOnly;
    
    //-----------------------------------------------
    // Internal Switches and Settings
    //-----------------------------------------------
    
    string backupAddList;
    string backupRemoveList;

    int iLines_RowsAdded;
    int iWrapCountAdded;

    int iStartWithPosition;

    HWND hWndLastActiveForSwitch;
    HWND hWndLastActiveForGrouping;

    bool bVisible;

    bool bMinimizeAtAllCosts;

    bool bCreateLayoutLocked;

    bool bMouseInside;
    
    bool bSkipButton;
    
    bool bIsScrolling;

    bool bInCriticalProcess;

    int iBackupExportedButtonCount;

    bool bAutoHidden;

    bool bParentDestroyNotificationReceived;

    //-----------------------------------------------
    // Lists (Overlays)
    //-----------------------------------------------

    GhostTaskbarList listGhostTaskbar;

    bool bHasOverlayConditional;

    GhostButtonList listGhostButton;

    vector<TaskbarButton *> taskbuttons;
    vector<TaskbarButton *> displayedtaskbuttons;

    //-----------------------------------------------
    // Painting/Layout
    //-----------------------------------------------
    
    HDC backupDC;
    HBITMAP backupBitmap;
    HBITMAP oldbackupBMP;

    void Create_xPaintClassSupport();
    void Destroy_xPaintClassSupport();

    PaintTexture *pTaskbarTexture;

    PaintTexture *pNormalButtonTexture;
    PaintTexture *pActiveButtonTexture;
    PaintTexture *pMinimizedButtonTexture;
    PaintTexture *pFlashingButtonTexture;
    PaintTexture *pHoverButtonTexture;
    PaintTexture *pPressedButtonTexture;
    PaintTexture *pGroupedButtonTexture;

    PaintIcon *pNormalIcon;
    PaintIcon *pActiveIcon;
    PaintIcon *pMinimizedIcon;
    PaintIcon *pFlashingIcon;
    PaintIcon *pHoverIcon;
    PaintIcon *pPressedIcon;
    PaintIcon *pGroupedIcon;

    PaintText *pNormalText;
    PaintText *pActiveText;
    PaintText *pMinimizedText;
    PaintText *pFlashingText;
    PaintText *pHoverText;
    PaintText *pPressedText;
    PaintText *pGroupedText;

    int iLeftBorder;
    int iTopBorder;
    int iRightBorder;
    int iBottomBorder;

    int iLayout;
    int iDirection;
    int iWrapDirection;
    int iWrapCount;
    bool bUseAdaptiveSize;

    int iLines_Rows;
    int iMaxLines_Rows;

    int iMaxButtonCount;

    //General Button Settings
    int iXSpacing;
    int iYSpacing;

    int iMaxTaskWidth;
    int iMaxTaskHeight;

    //AutoSize
    bool bAutoSize;
    int iMaxTaskbarSize;
    int iAutoSizeDelay;
    int iAutoSizeSteps;

    bool AutoSizeFunction(int buttons);

    //-----------------------------------------------
    // Alpha
    //-----------------------------------------------

    bool bAlphaMap;
    bool bAlphaFade;
    int iAlphaTransparency;

    int iUpdateAlpha;

    int iAlphaFadeStep;
    int iAlphaFadeDelay;
    int iAlphaFadeBangEndValue;

    bool bAlphaInitialized;
    bool bLayeredFlag;

    //-----------------------------------------------
    // Temporary Help Pointers
    //-----------------------------------------------
    
    //For fixing Multiple Executions
    TaskbarButton *pClickCapture;

    TaskbarButton *pDropCapture;

    //For ReposButtons
    TaskbarButton *pMoveCapture;

    //For Tooltips
    TaskbarButton *pCapture;
    
    //For Better Deactivation Detection 
    TaskbarButton *pActiveTask;

    TaskbarButton *pDontPaintAddedTask;

    //----------------------------------------------------
    // Actions
    //----------------------------------------------------

    int iLButtonAction;
    int iRButtonAction;
    int iMButtonAction;

    string taskbarLButtonDown;
    string taskbarLButtonUp;
    string taskbarRButtonDown;
    string taskbarRButtonUp;
    string taskbarMButtonDown;
    string taskbarMButtonUp;
    string taskbarMWheelDown;
    string taskbarMWheelUp;
    string taskbarEnter;
    string taskbarLeave;
    string taskbarAdd;
    string taskbarRemove;
    string taskbarMove;
    string taskbarResize;

    string taskbarFlash;

    string taskbarSelect;

    string taskbarLineAdd;
    string taskbarLineRemove;

    string taskbarEmpty;
    string taskbarFilled;

    //----------------------------------------------------
    // Sorting
    //----------------------------------------------------

    void SortTasks(vector<TaskbarButton *> &sorttaskbuttons);

    static bool CompareTasks(TaskbarButton *b1, TaskbarButton *b2)
    {
        string one;
        string two;

        if ( b1->taskbar->iAutoSort == 1 )
        {
            CString help;
            help = b1->strClassName.c_str();
            one = (help.Right(help.GetLength()-help.ReverseFind('\\')-1));
            one.append( b1->strCaption );
            help = b2->strClassName.c_str();
            two = (help.Right(help.GetLength()-help.ReverseFind('\\')-1));
            two.append( b2->strCaption );
        }
        else if ( b1->taskbar->iAutoSort == 2 )
        {
            one = GetTaskClass( b1->hWndButtonTask );
            one.append( b1->strCaption );
            two = GetTaskClass( b2->hWndButtonTask );
            two.append( b2->strCaption );
        }
        else if ( b1->taskbar->iAutoSort == 3 )
        {
            one = b1->strCaption;
            two = b2->strCaption;
        }
        
        if (lstrcmpi(one.c_str(), two.c_str()) < 0)
            return true;
        else
            return false;
    };

    //----------------------------------------------------
    // ToolTips
    //----------------------------------------------------
    
    PaintTooltip *pTooltip;
    HWND hWndToolTip;

    bool bShowToolTip;

    int iTooltipMode;

private:

    void UpdateWindowTransparency(int newAlphaTransparency);

    TaskbarButton *ButtonAtPoint( int x, int y );
    TaskbarButton *SearchForButtonByHandle( HWND hWndSearch );
    TaskbarButton *SearchForButtonByClass( HWND hWndSearch );
    
    void ToggleFlashingButton();

    static BOOL WINAPI EnumWindowsProc( HWND hWnd, LPARAM lParam );

    void onRedraw( HWND hApplication, bool fFlash );
    void onWindowActivated( HWND hApplication, BOOL fFullScreen);

    int CheckDisplayedTasks();

    void AmountOfTasksChanged(TaskbarButton *button = NULL);

protected:

    virtual void onLButtonDown(int x, int y);
    virtual void onLButtonUp(int x, int y);
    virtual void onMButtonDown(int x, int y);
    virtual void onMButtonUp(int x, int y);
    virtual void onRButtonDown(int x, int y);
    virtual void onRButtonUp(int x, int y); 
    virtual void onWheelDown(int x, int y);
    virtual void onWheelUp(int x, int y);
    virtual void onMouseEnter();
    virtual void onMouseLeave();
    virtual void onMouseMove(int x, int y);
    virtual void onPaint(HDC hDC);
    virtual void onWindowPosChanging(WINDOWPOS *windowPos);
    virtual void onMove(int x, int y);
    virtual void onSize(int width, int height);
    virtual void onTimer(int timerID);

private:

    BOOL onWindowMessage( UINT message, WPARAM wParam, LPARAM lParam, LRESULT &lResult );

public:

    static LRESULT WINAPI WindowProcedure( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );

};

#endif // !defined(AFX_TASKBAR_H__EBF1ABD2_1EA7_40AF_93CF_D4F2B4ECC8EE__INCLUDED_)
