/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2009 The LS-Universe Team

    Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "TaskbarButton.h"
#include "Taskbar.h"
#include "WindowUtilities.h"

#include "GhostButton.h"

extern ProcessStatsRequestFunc m_pProcessStatsRequest;

//Grouping
#define X_GROUP_MINIMIZE    10
#define X_GROUP_RESTORE     11
#define X_GROUP_MAXIMIZE    12
#define X_GROUP_CLOSE       13

#define X_GROUP_SEPARATOR   66

TaskbarButton::TaskbarButton( Taskbar *settaskbar, HWND hWnd ) : 
    taskbar(NULL),

    iButtonX(0),
    iButtonY(0),
    iButtonWidth(0),
    iButtonHeight(0),
    
    bFlashState(false),
    bIsFlashing(false),
    iFlashCounter(-1),
    bIsHover(false),
    bIsPressed(false),
    bIsScrolling(false),
    iScrollPosition(0),

    iCurrentState(STATE_NORMAL),
    
    iPositionInTaskbar(-1),

    //hIcon(NULL),

    bMonitorHidden(false),
    hMonitorCurrentMonitor(NULL)
{
    taskbar = settaskbar;

    hWndButtonTask = hWnd;

    strClassName = GetTaskFileName( hWnd );

    listhWndGroup.insert(listhWndGroup.end(), hWnd);

    if ( taskbar )
    {
        if (taskbar->bShowText || taskbar->bShowToolTip)
            strCaption = GetWindowCaption();
        else
            strCaption = "";
        
        if (taskbar->bShowToolTip)
            CreateTooltip();
        else
            strTooltip = "";
    }
}

TaskbarButton::~TaskbarButton()
{
    taskbar = NULL;

    listhWndGroup.clear();

    //if (hIcon)
    //  DestroyIcon(hIcon);
}

bool TaskbarButton::HasPoint( int xP, int yP )
{
    if ( xP < (iButtonX + iButtonWidth) && xP >= iButtonX )
    {
        if ( yP < (iButtonY + iButtonHeight) && yP >= iButtonY )
            return TRUE;
    }

    return FALSE;
}


void TaskbarButton::CreateTooltip(bool grouped)
{
    if ( !taskbar )
        return;

    if ( grouped )
    {
        if ( taskbar->iTooltipMode == 0 )
        {
            strTooltip = ( "Grouping:\t" + IntToString(listhWndGroup.size()) + " Grouped Tasks\n");
            strTooltip.append( "Active Caption:\t" + strCaption + "\nExecutable:\t" + strClassName );
        }
        else if ( taskbar->iTooltipMode == 1 )
        {
            strTooltip = ( IntToString(listhWndGroup.size()) + " Grouped Tasks\n");
            strTooltip.append( strCaption + "\n" + strClassName );
        }
        else if ( taskbar->iTooltipMode == 2 )
        {
            strTooltip = ( "Grouping:\t" + IntToString(listhWndGroup.size()) + " Grouped Tasks\n");
            strTooltip.append( "Active Caption:\t" + strCaption );
        }
        else if ( taskbar->iTooltipMode == 3 )
        {
            strTooltip = ( IntToString(listhWndGroup.size()) + " Grouped Tasks\n");
            strTooltip.append( strCaption );
        }
        else if ( taskbar->iTooltipMode == 4 )
        {
            strTooltip = ( "Grouping:\t" + IntToString(listhWndGroup.size()) + " Grouped Tasks\n");
            strTooltip.append( "Executable:\t" + strClassName );
        }
        else if ( taskbar->iTooltipMode == 5 )
        {
            strTooltip = ( IntToString(listhWndGroup.size()) + " Grouped Tasks\n");
            strTooltip.append( strClassName );
        }
    }
    else
    {
        if ( taskbar->iTooltipMode == 0 )
        {
            strTooltip = ( "Caption:\t" + strCaption + "\nExecutable:\t" + strClassName );
        }
        else if ( taskbar->iTooltipMode == 1 )
        {
            strTooltip = ( strCaption + "\n" + strClassName );
        }
        else if ( taskbar->iTooltipMode == 2 )
        {
            strTooltip = ( "Caption:  " + strCaption );
        }
        else if ( taskbar->iTooltipMode == 3 )
        {
            strTooltip = ( strCaption );
        }
        else if ( taskbar->iTooltipMode == 4 )
        {
            strTooltip = ( "Executable:  " + strClassName );
        }
        else if ( taskbar->iTooltipMode == 5 )
        {
            strTooltip = ( strClassName );
        }
    }
}

bool TaskbarButton::Update(bool needRedraw)
{
    if ( !taskbar )
        return false;

    if ( iRealMonitors > 1 && taskbar->iThisMultiMonitorOnly != -1 )
    {
        if ( taskbar->iThisMultiMonitorOnly == 0 )
        {
            if ( MonitorFromWindow( taskbar->GethWnd(), MONITOR_DEFAULTTOPRIMARY ) != MonitorFromWindow( hWndButtonTask, MONITOR_DEFAULTTOPRIMARY ) )
                bMonitorHidden = true;
            else
                bMonitorHidden = false;
        }
        else if ( taskbar->iThisMultiMonitorOnly == 1 )
        {
            if ( pMon != MonitorFromWindow( hWndButtonTask, MONITOR_DEFAULTTOPRIMARY ) )
                bMonitorHidden = true;
            else
                bMonitorHidden = false;
        }
        else if ( taskbar->iThisMultiMonitorOnly == 2 )
        {
            if ( sMon != MonitorFromWindow( hWndButtonTask, MONITOR_DEFAULTTOPRIMARY ) )
                bMonitorHidden = true;
            else
                bMonitorHidden = false;
        }
        else if ( taskbar->iThisMultiMonitorOnly == 3 )
        {
            if ( tMon != MonitorFromWindow( hWndButtonTask, MONITOR_DEFAULTTOPRIMARY ) )
                bMonitorHidden = true;
            else
                bMonitorHidden = false;
        }
        else
        {
            bMonitorHidden = false;
        }

        needRedraw = true;
    }

    if (taskbar->bShowText || taskbar->bShowToolTip)
    {
        string temp1 = strCaption;
        string temp2 = GetWindowCaption();

        if ( m_pProcessStatsRequest && taskbar->bUseCustomizedCaption )
        {
            CString strCurrentCaption = "";
            CString strTmp = temp2.c_str();
            
            switch(GetCurrentState())
            {
                case STATE_SELECTED:

                    strCurrentCaption = taskbar->strActiveText.c_str();
            
                break;

                case STATE_MINIMIZED:

                    strCurrentCaption = taskbar->strMinimizedText.c_str();

                break;

                default:

                    strCurrentCaption = taskbar->strNormalText.c_str();
                
                break;
            }

            strTmp.Replace("[", "[[");
            strTmp.Replace("]", "]]");
            strCurrentCaption.Replace("%caption%", strTmp);
            strCurrentCaption.Replace("%executable%", strClassName.c_str());
        
            char tmp[4096] = {0};
            m_pProcessStatsRequest( tmp, strCurrentCaption );

            temp2 = tmp;
        }

        if ( _stricmp( temp1.c_str(), temp2.c_str() ) != NULL && !temp2.empty() )
        {
            strCaption = temp2;
            if (taskbar->bShowToolTip && listhWndGroup.size() > 1)
            {
                CreateTooltip(true);
            }
            else if (taskbar->bShowToolTip)
            {   
                CreateTooltip();
            }

            if (taskbar->bShowText)
                needRedraw = true;
        }
    }

    if (!needRedraw && taskbar->bShowIcon)
    {
        if (IsOnTaskbarList(hWndButtonTask, taskbar->ChangingIconList, false))
            needRedraw = true;
    }

    return needRedraw;  
}

string TaskbarButton::GetWindowCaption( HWND check_hTask )
{
    if (check_hTask == NULL)
        check_hTask = hWndButtonTask;
    
    return GetTaskWindowText(check_hTask);
}

void TaskbarButton::ButtonClick( int x, int y, int action, bool minimize )
{
    switch (action)
    {
        case 1:
            
            Select(minimize);
            break;

        case 2:

            Menu();
            break;

        case 3:

            Close(hWndButtonTask);
            break;

        case 4:

            OnTopToggle(hWndButtonTask);
            break;
    }
}

void TaskbarButton::Select(bool minimize)
{
    if ( !taskbar )
        return;

    if( bIsFlashing )
    {
        KillTimer(taskbar->GethWnd(), 9/*TIMER_FLASH*/);
        SetFlashState(0);
        GetCurrentState();
    }

    if (listhWndGroup.size() > 1 && taskbar->bUseGrouping)
    {
        GroupMenu();
        return;
    }

    if ( minimize )
    {
        if (!taskbar->bNoMinimizeOnClick)
        {
            iCurrentState = STATE_MINIMIZED;
            MinimizeWindow(hWndButtonTask);
            Sleep(10);
            taskbar->Update(true);
        }
        else
        {
            SelectWindow(hWndButtonTask);
            if ( !taskbar->taskbarSelect.empty() )
                LSExecute(taskbar->GethWnd(), taskbar->taskbarSelect.c_str(), SW_SHOWNORMAL);
        }
    }
    else
    {
        switch(iCurrentState)
        {
            case STATE_SELECTED:

                if (!taskbar->bNoMinimizeOnClick)
                {
                    iCurrentState = STATE_MINIMIZED;
                    MinimizeWindow(hWndButtonTask);
                    Sleep(10);
                    taskbar->Update(true);
                }
                else
                {
                    SelectWindow(hWndButtonTask);
                    if ( !taskbar->taskbarSelect.empty() )
                        LSExecute(taskbar->GethWnd(), taskbar->taskbarSelect.c_str(), SW_SHOWNORMAL);
                }

            break;

            case STATE_MINIMIZED:

                SelectWindow(hWndButtonTask, true);
                if ( !taskbar->taskbarSelect.empty() )
                    LSExecute(taskbar->GethWnd(), taskbar->taskbarSelect.c_str(), SW_SHOWNORMAL);

            break;

            default:

                SelectWindow(hWndButtonTask);
                if ( !taskbar->taskbarSelect.empty() )
                    LSExecute(taskbar->GethWnd(), taskbar->taskbarSelect.c_str(), SW_SHOWNORMAL);

            break;
        }
    }
}

void TaskbarButton::Menu()
{
    if ( !taskbar )
        return;

    HMENU systemMenu;

    systemMenu = GetSystemMenu( hWndButtonTask, FALSE );

    WINDOWPLACEMENT wp;
    wp.length = sizeof(WINDOWPLACEMENT);
    GetWindowPlacement( hWndButtonTask, &wp );

    // restore is enabled only if minimized or maximized (not normal)
    EnableMenuItem( systemMenu, SC_RESTORE, MF_BYCOMMAND | (wp.showCmd != SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

    // move is enabled only if normal
    EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND | (wp.showCmd == SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

    // size is enabled only if normal
    EnableMenuItem( systemMenu, SC_SIZE, MF_BYCOMMAND | (wp.showCmd == SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

    // minimize is enabled only if not minimized
    EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND | (wp.showCmd != SW_SHOWMINIMIZED ? MF_ENABLED : MF_GRAYED) );

    // maximize is enabled only if not maximized
    EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND | (wp.showCmd != SW_SHOWMAXIMIZED ? MF_ENABLED : MF_GRAYED) );

    int counter = 0;

    if (listhWndGroup.size() > 1 && taskbar->bUseGrouping)
    {   
        AppendMenu( systemMenu, MF_SEPARATOR, X_GROUP_SEPARATOR, "" );

        AppendMenu( systemMenu, MF_STRING, X_GROUP_MINIMIZE, "Minimize All In Group" );
        AppendMenu( systemMenu, MF_STRING, X_GROUP_RESTORE, "Restore All In Group" );
        AppendMenu( systemMenu, MF_STRING, X_GROUP_MAXIMIZE, "Maximize All In Group" );
        AppendMenu( systemMenu, MF_STRING, X_GROUP_CLOSE, "Close All In Group" );

        AppendMenu( systemMenu, MF_SEPARATOR, X_GROUP_SEPARATOR, "" );

        counter = 1;
        for(hWndGroupListIterator hglit = listhWndGroup.begin(); hglit != listhWndGroup.end(); ++hglit)
        {
            AppendMenu( systemMenu, MF_STRING , counter, ("Close " + GetWindowCaption(*hglit)).c_str() );
            counter++;
        }
        counter--;
        //END Grouping
    }

    // let application modify menu
    PostMessage( hWndButtonTask, WM_INITMENUPOPUP, (WPARAM) systemMenu, MAKELPARAM( 0, TRUE ) );
    PostMessage( hWndButtonTask, WM_INITMENU, (WPARAM) systemMenu, 0 );

    // display the menu
    POINT pt;
    GetCursorPos( &pt );

    int command = TrackPopupMenu( systemMenu,
                                  TPM_RETURNCMD | TPM_RIGHTBUTTON,
                                  pt.x, pt.y, 0,
                                  taskbar->GethWnd(),
                                  NULL );

    if (listhWndGroup.size() > 1 && taskbar->bUseGrouping)
    {
        DeleteMenu( systemMenu, X_GROUP_SEPARATOR, MF_SEPARATOR );

        DeleteMenu( systemMenu, X_GROUP_SEPARATOR, MF_BYCOMMAND );
        DeleteMenu( systemMenu, X_GROUP_MINIMIZE, MF_BYCOMMAND );
        DeleteMenu( systemMenu, X_GROUP_RESTORE, MF_BYCOMMAND );
        DeleteMenu( systemMenu, X_GROUP_MAXIMIZE, MF_BYCOMMAND );
        DeleteMenu( systemMenu, X_GROUP_CLOSE, MF_BYCOMMAND );

        DeleteMenu( systemMenu, X_GROUP_SEPARATOR, MF_SEPARATOR );

        while ( counter >= 1 )
        {
            DeleteMenu( systemMenu, counter, MF_BYCOMMAND );
            counter--;
        }
    }

    if ( command )
    {
        if (listhWndGroup.size() > 1 && taskbar->bUseGrouping)
        {
            switch( command )
            {
                case X_GROUP_MINIMIZE:
                {
                    for ( hWndGroupListIterator hglit = listhWndGroup.begin(); hglit != listhWndGroup.end(); ++hglit )
                        MinimizeWindow( (*hglit) );
                    
                    return;
                }

                case X_GROUP_RESTORE:
                {
                    for ( hWndGroupListIterator hglit = listhWndGroup.begin(); hglit != listhWndGroup.end(); ++hglit )
                        RestoreWindow( (*hglit) );

                    return;
                }

                case X_GROUP_MAXIMIZE:
                {
                    for ( hWndGroupListIterator hglit = listhWndGroup.begin(); hglit != listhWndGroup.end(); ++hglit )
                        MaximizeWindow( (*hglit) );

                    return;
                }

                case X_GROUP_CLOSE:
                {
                    for ( hWndGroupListIterator hglit = listhWndGroup.begin(); hglit != listhWndGroup.end(); ++hglit )
                        Close( (*hglit) );
                        
                    return;
                }
            }

            hWndGroupListIterator hglit = listhWndGroup.begin();
            counter = 1; 
            while (counter < command)
            {
                hglit++;
                counter++;
            }

            if ( command <= listhWndGroup.size() )
            {
                Close( (*hglit) );
                hWndButtonTask = (*hglit);
                return;
            }
        }
        
        if ( command != SC_MINIMIZE && command != SC_CLOSE )
        {
            SelectWindow(hWndButtonTask);
            if ( !taskbar->taskbarSelect.empty() )
                LSExecute(taskbar->GethWnd(), taskbar->taskbarSelect.c_str(), SW_SHOWNORMAL);
        }

        PostMessage( hWndButtonTask, WM_SYSCOMMAND, (WPARAM) command, MAKELPARAM( pt.x, pt.y ) );
    }
}

void TaskbarButton::GroupMenu()
{
    if ( !taskbar )
        return;

    HMENU GroupMenu = CreatePopupMenu();

    int counter = 1;
    for(hWndGroupListIterator hglit = listhWndGroup.begin(); hglit != listhWndGroup.end(); ++hglit)
    {
        AppendMenu( GroupMenu, MF_STRING, counter, GetWindowCaption((*hglit)).c_str() );
        counter++;
    }

    // display the menu
    POINT pt;
    GetCursorPos( &pt );

    int command = TrackPopupMenu( GroupMenu,
                                  TPM_RETURNCMD | TPM_RIGHTBUTTON,
                                  pt.x, pt.y, 0,
                                  taskbar->GethWnd(),
                                  NULL );

    if ( command )
    {
        hWndGroupListIterator hglit = listhWndGroup.begin();
        counter = 1; 
        while (counter < command)
        {
            hglit++;
            counter++;
        }

        if ( counter <= listhWndGroup.size() )
        {
            if (!taskbar->bNoMinimizeOnClick && taskbar->hWndLastActiveForGrouping == (*hglit))
            {
                MinimizeWindow((*hglit));
            }
            else
            {
                SelectWindow((*hglit), IsIconic((*hglit)));
                if ( !taskbar->taskbarSelect.empty() )
                    LSExecute(taskbar->GethWnd(), taskbar->taskbarSelect.c_str(), SW_SHOWNORMAL);

                hWndButtonTask = (*hglit);
            }
        }
    }
    
    if (GroupMenu)
        DestroyMenu(GroupMenu);
}

int TaskbarButton::GetCurrentState()
{
    iCurrentState = GetState(hWndButtonTask);

    return iCurrentState;
}

void TaskbarButton::SetButtonLayout( int x, int y, int width, int height )
{
    iButtonX = x;
    iButtonY = y;
    iButtonWidth = width;   
    iButtonHeight = height;
}

void TaskbarButton::Paint( HDC desthDC )
{
    if ( !taskbar || iButtonWidth <= 0 || iButtonHeight <= 0 )
        return;

    //Ampersand elimination
    CString strUsedCaption = strCaption.c_str();
    strUsedCaption.Replace("&", "&&");

    int x = 0;
    int y = 0;
    int width = 0;
    int height = 0;

    if ( !bIsFlashing || !taskbar->bUseFlashingState )
        GetCurrentState();
    else if ( bIsFlashing && !bFlashState && iFlashCounter <= 20 )
        iCurrentState = STATE_NORMAL;
    else if ( bIsFlashing && (bFlashState || iFlashCounter > 20) )
        iCurrentState = STATE_FLASHING;
    
    //Update Icon - Sure is Sure
    /*if (hIcon)
    {
        DestroyIcon(hIcon);
        hIcon = NULL;
    }*/

    HICON hIcon = NULL;

    //Button Background Painting with Icon and Text 
    if (iCurrentState == STATE_FLASHING && taskbar->bUseFlashingState)
    {
        taskbar->pFlashingButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
        
        OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 3/*flashing*/, true/*PaintBelow*/);

        if (taskbar->bShowIcon)
        {
            /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigFlashingIcon); /*)*/
                //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

            taskbar->pFlashingIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
        }

        if (taskbar->bShowText)
        {
            if ( m_pProcessStatsRequest && !taskbar->strFlashingText.empty() )
            {
                CString strCurrentCaption = taskbar->strFlashingText.c_str();
                CString strTmp = strUsedCaption;

                strTmp.Replace("[", "[[");
                strTmp.Replace("]", "]]");
                strCurrentCaption.Replace("%caption%", strTmp);
                strCurrentCaption.Replace("%executable%", strClassName.c_str());
            
                char tmp[4096] = {0};
                m_pProcessStatsRequest( tmp, strCurrentCaption );

                strUsedCaption = tmp;
            }

            if ( bIsScrolling )
            {
                ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strUsedCaption, 3);
            }
            else
            {
                taskbar->pFlashingText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strUsedCaption, taskbar->bMultiLineUsed);
            }
        }

        OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 3/*flashing*/, false/*PaintBelow*/);
    }
    else if (taskbar->bUseGrouping && listhWndGroup.size() > 1 && !bIsHover && !bIsPressed)
    {
        if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
        {
            taskbar->pNormalButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
        
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 0, true/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_MINIMIZED)
        {
            taskbar->pMinimizedButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
            
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 2, true/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_SELECTED)
        {
            taskbar->pActiveButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
            
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 1, true/*PaintBelow*/);
        }

        taskbar->pGroupedButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);

        OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 5/*Grouped*/, true/*PaintBelow*/);

        if ( taskbar->bShowIcon )
        {
            if (taskbar->bUseGroupedIconSettings && taskbar->pGroupedIcon)
            {   
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigGroupedIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pGroupedIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
            else if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigNormalIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pNormalIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
            else if (iCurrentState == STATE_MINIMIZED)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigMinimizedIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pMinimizedIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
            else if (iCurrentState == STATE_SELECTED)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigActiveIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pActiveIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
        }

        if (taskbar->bShowText)
        {
            CString strText = strUsedCaption;
        
            int found = strText.Find(" - ");
            if ( found != -1 && strText.Find(" - ", found+3) != -1)
                strText = strText.Mid(found+3, strText.Find(" - ", found+3)-found-3);
            else if ( found != -1 )
                strText = strText.Left(strText.Find(" - "));
            strText += " ( ";
            strText += IntToString( listhWndGroup.size() ).c_str();
            strText += " )";

            if ( m_pProcessStatsRequest && !taskbar->strGroupedText.empty() )
            {
                CString strCurrentCaption = taskbar->strGroupedText.c_str();
                CString strTmp = strText;

                strTmp.Replace("[", "[[");
                strTmp.Replace("]", "]]");
                strCurrentCaption.Replace("%caption%", strTmp);
                strCurrentCaption.Replace("%executable%", strClassName.c_str());
            
                char tmp[4096] = {0};
                m_pProcessStatsRequest( tmp, strCurrentCaption );

                strText = tmp;
            }

            if ( bIsScrolling )
            {
                if (taskbar->bUseGroupedFontSettings && taskbar->pGroupedText)
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 5);
                else if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 0);
                else if (iCurrentState == STATE_MINIMIZED)
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 2);
                else if (iCurrentState == STATE_SELECTED)
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 1);
            }
            else
            {
                if (taskbar->bUseGroupedFontSettings && taskbar->pGroupedText)
                    taskbar->pGroupedText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
                else if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
                    taskbar->pNormalText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
                else if (iCurrentState == STATE_MINIMIZED)
                    taskbar->pMinimizedText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
                else if (iCurrentState == STATE_SELECTED)
                    taskbar->pActiveText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
            }
        }

        if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
        {
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 0, false/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_MINIMIZED)
        {
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 2, false/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_SELECTED)
        {
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 1, false/*PaintBelow*/);
        }

        OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 5/*Grouped*/, false/*PaintBelow*/);
    }
    else if (bIsPressed)
    {
        if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
        {
            taskbar->pNormalButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
        
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 0, true/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_MINIMIZED)
        {
            taskbar->pMinimizedButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
            
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 2, true/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_SELECTED)
        {
            taskbar->pActiveButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
            
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 1, true/*PaintBelow*/);
        }

        taskbar->pPressedButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
        
        OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 6/*Pressed*/, true/*PaintBelow*/);

        if (taskbar->bShowIcon)
        {
            if (taskbar->bUsePressedIconSettings && taskbar->pPressedIcon)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigPressedIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pPressedIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
            else if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigNormalIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pNormalIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
            else if (iCurrentState == STATE_MINIMIZED)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigMinimizedIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pMinimizedIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
            else if (iCurrentState == STATE_SELECTED)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigActiveIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pActiveIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
        }

        if (taskbar->bShowText)
        {
            CString strText = strUsedCaption;
            
            if (taskbar->bUseGrouping && listhWndGroup.size() > 1)
            {
                int found = strText.Find(" - ");
                if ( found != -1 && strText.Find(" - ", found+3) != -1)
                    strText = strText.Mid(found+3, strText.Find(" - ", found+3)-found-3);
                else if ( found != -1 )
                    strText = strText.Left(strText.Find(" - "));
                strText += " ( ";
                strText += IntToString( listhWndGroup.size() ).c_str();
                strText += " )";
            }

            if ( m_pProcessStatsRequest && !taskbar->strPressedText.empty() )
            {
                CString strCurrentCaption = taskbar->strPressedText.c_str();
                CString strTmp = strText;

                strTmp.Replace("[", "[[");
                strTmp.Replace("]", "]]");
                strCurrentCaption.Replace("%caption%", strTmp);
                strCurrentCaption.Replace("%executable%", strClassName.c_str());
            
                char tmp[4096] = {0};
                m_pProcessStatsRequest( tmp, strCurrentCaption );

                strText = tmp;
            }

            if ( bIsScrolling )
            {
                if (taskbar->bUsePressedFontSettings && taskbar->pPressedText)
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 6);
                else if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 0);
                else if (iCurrentState == STATE_MINIMIZED)
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 2);
                else if (iCurrentState == STATE_SELECTED)
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 1);
            }
            else
            {
                if (taskbar->bUsePressedFontSettings && taskbar->pPressedText)
                    taskbar->pPressedText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
                else if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
                    taskbar->pNormalText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
                else if (iCurrentState == STATE_MINIMIZED)
                    taskbar->pMinimizedText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
                else if (iCurrentState == STATE_SELECTED)
                    taskbar->pActiveText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
            }
        }

        if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
        {
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 0, false/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_MINIMIZED)
        {
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 2, false/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_SELECTED)
        {
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 1, false/*PaintBelow*/);
        }

        OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 6/*Pressed*/, false/*PaintBelow*/);
    }
    else if (bIsHover || bIsScrolling)
    {
        if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
        {
            taskbar->pNormalButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
        
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 0, true/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_MINIMIZED)
        {
            taskbar->pMinimizedButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
            
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 2, true/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_SELECTED)
        {
            taskbar->pActiveButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
            
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 1, true/*PaintBelow*/);
        }

        if ( taskbar->pHoverButtonTexture )
        {
            if ( CheckState(taskbar->iSpecialHoverState) )
                taskbar->pHoverButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
        
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 4/*Hover*/, true/*PaintBelow*/);
        }

        if (taskbar->bShowIcon)
        {
            if (taskbar->bUseHoverIconSettings && taskbar->pHoverIcon && CheckState(taskbar->iSpecialHoverState))
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigHoverIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pHoverIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
            else if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigNormalIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pNormalIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
            else if (iCurrentState == STATE_MINIMIZED)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigMinimizedIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pMinimizedIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
            else if (iCurrentState == STATE_SELECTED)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigActiveIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pActiveIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }
        }

        if (taskbar->bShowText)
        {
            CString strText = strUsedCaption;
            
            if (taskbar->bUseGrouping && listhWndGroup.size() > 1)
            {
                int found = strText.Find(" - ");
                if ( found != -1 && strText.Find(" - ", found+3) != -1)
                    strText = strText.Mid(found+3, strText.Find(" - ", found+3)-found-3);
                else if ( found != -1 )
                    strText = strText.Left(strText.Find(" - "));
                strText += " ( ";
                strText += IntToString( listhWndGroup.size() ).c_str();
                strText += " )";
            }

            if ( m_pProcessStatsRequest && !taskbar->strHoverText.empty() )
            {
                CString strCurrentCaption = taskbar->strHoverText.c_str();
                CString strTmp = strText;

                strTmp.Replace("[", "[[");
                strTmp.Replace("]", "]]");
                strCurrentCaption.Replace("%caption%", strTmp);
                strCurrentCaption.Replace("%executable%", strClassName.c_str());
            
                char tmp[4096] = {0};
                m_pProcessStatsRequest( tmp, strCurrentCaption );

                strText = tmp;
            }

            if ( bIsScrolling )
            {
                if (taskbar->bUseHoverFontSettings && taskbar->pHoverText && CheckState(taskbar->iSpecialHoverState))
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 4);
                else if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 0);
                else if (iCurrentState == STATE_MINIMIZED)
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 2);
                else if (iCurrentState == STATE_SELECTED)
                    ScrollPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, (string) strText, 1);
            }
            else
            {
                if (taskbar->bUseHoverFontSettings && taskbar->pHoverText && CheckState(taskbar->iSpecialHoverState))
                    taskbar->pHoverText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
                else if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
                    taskbar->pNormalText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
                else if (iCurrentState == STATE_MINIMIZED)
                    taskbar->pMinimizedText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
                else if (iCurrentState == STATE_SELECTED)
                    taskbar->pActiveText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strText, taskbar->bMultiLineUsed);
            }
        }

        if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
        {
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 0, false/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_MINIMIZED)
        {
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 2, false/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_SELECTED)
        {
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 1, false/*PaintBelow*/);
        }

        if ( taskbar->pHoverButtonTexture )
        {
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 4/*Hover*/, false/*PaintBelow*/);
        }
    }
    else
    {
        if (iCurrentState != STATE_SELECTED && iCurrentState != STATE_MINIMIZED)
        {
            taskbar->pNormalButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
        
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 0, true/*PaintBelow*/);

            if(taskbar->bShowIcon)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigNormalIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pNormalIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }

            if (taskbar->bShowText)
                taskbar->pNormalText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strUsedCaption, taskbar->bMultiLineUsed);

            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 0, false/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_MINIMIZED)
        {
            taskbar->pMinimizedButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
            
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 2, true/*PaintBelow*/);

            if(taskbar->bShowIcon)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigMinimizedIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pMinimizedIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }

            if (taskbar->bShowText)
                taskbar->pMinimizedText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strUsedCaption, taskbar->bMultiLineUsed);

            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 2, false/*PaintBelow*/);
        }
        else if (iCurrentState == STATE_SELECTED)
        {
            taskbar->pActiveButtonTexture->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight);
            
            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 1, true/*PaintBelow*/);

            if(taskbar->bShowIcon)
            {
                /*if (*/ GetIconOnCustomList( hWndButtonTask, hIcon, taskbar->CustomIconList, taskbar->bUseBigActiveIcon); /*)*/
                    //hIcon = (HICON)CopyImage((HANDLE)hIcon, IMAGE_ICON, 0, 0, 0);

                taskbar->pActiveIcon->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, hIcon);
            }

            if (taskbar->bShowText)
                taskbar->pActiveText->apply(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, strUsedCaption, taskbar->bMultiLineUsed);

            OverlayButtonPainter(desthDC, iButtonX, iButtonY, iButtonWidth, iButtonHeight, 1, false/*PaintBelow*/);
        }
    }
}

void TaskbarButton::OverlayButtonPainter(HDC desthDC, int x, int y, int width, int height, int buttonType, bool bPaintBelow)
{
    if ( !taskbar->listGhostButton.empty() )
    {
        for(GhostButtonListIterator glit = taskbar->listGhostButton.begin(); glit != taskbar->listGhostButton.end(); glit++)
        {
            if ( *glit != NULL )
            {
                if ( (bPaintBelow && (*glit)->bOnTop) || (!bPaintBelow && !(*glit)->bOnTop) )
                    continue;

                if ( !(*glit)->bDoPaint )
                    continue;

                int gbx = x + xParseCoord((*glit)->strOrgX.c_str(), 0, width, false);
                int gby =  y + xParseCoord((*glit)->strOrgY.c_str(), 0, height, true);
                int gbwidth = xParseDimen((*glit)->strOrgWidth.c_str(), width, width, false);
                int gbheight = xParseDimen((*glit)->strOrgHeight.c_str(), height, height, true);

                if (buttonType == 1 && (*glit)->pActiveButtonTexture)
                    (*glit)->pActiveButtonTexture->apply(desthDC, gbx, gby, gbwidth, gbheight);
                else if (buttonType == 2 && (*glit)->pMinimizedButtonTexture)
                    (*glit)->pMinimizedButtonTexture->apply(desthDC, gbx, gby, gbwidth, gbheight);
                else if (buttonType == 3 && (*glit)->pFlashingButtonTexture)
                    (*glit)->pFlashingButtonTexture->apply(desthDC, gbx, gby, gbwidth, gbheight);
                else if (buttonType == 4 && (*glit)->pHoverButtonTexture && CheckState((*glit)->iSpecialHoverState) )
                    (*glit)->pHoverButtonTexture->apply(desthDC, gbx, gby, gbwidth, gbheight);
                else if (buttonType == 5 && (*glit)->pGroupedButtonTexture)
                    (*glit)->pGroupedButtonTexture->apply(desthDC, gbx, gby, gbwidth, gbheight);
                else if (buttonType == 6 && (*glit)->pPressedButtonTexture)
                    (*glit)->pPressedButtonTexture->apply(desthDC, gbx, gby, gbwidth, gbheight);
                else if (buttonType == 0 && (*glit)->pNormalButtonTexture)
                    (*glit)->pNormalButtonTexture->apply(desthDC, gbx, gby, gbwidth, gbheight);
            }
            else
            {
                break;
            }
        }
    }
}

void TaskbarButton::ScrollPainter( HDC desthDC, int x, int y, int width, int height, const string &drawText, int buttonType )
{
    //--------------------------------------------------------------------------------
    // start rendering text
    //--------------------------------------------------------------------------------
    POINT textSize;
    RECT textBorder;
    RECT textFade;
    int textAlphaTransparency;
    PaintText *painttext;

    if (buttonType == 1)
        painttext = taskbar->pActiveText;
    else if (buttonType == 2)
        painttext = taskbar->pMinimizedText;
    else if (buttonType == 3)
        painttext = taskbar->pFlashingText;
    else if (buttonType == 4)
        painttext = taskbar->pHoverText;
    else if (buttonType == 5)
        painttext = taskbar->pGroupedText;
    else if (buttonType == 6)
        painttext = taskbar->pPressedText;
    else
        painttext = taskbar->pNormalText;

    textSize = painttext->measure(desthDC, 0, 0, drawText.c_str());
    textBorder = painttext->getBorder();
    textFade = painttext->getTextFade();
    textAlphaTransparency = painttext->getAlphaTransparency();

    //Create Region in which text drawing is allowed
    HRGN allowedArea = CreateRectRgn( x+textBorder.left, y+textBorder.top, x+width-textBorder.right, y+height-textBorder.bottom );
    SelectClipRgn(desthDC, allowedArea);
    DeleteObject(allowedArea);

    if ( taskbar->bAlphaMap || textFade.left != 0 || textFade.right != 0 || textFade.top != 0 || textFade.bottom != 0 || textAlphaTransparency != 255 )
    {       
        BITMAPINFO alpha;
        ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
        alpha.bmiHeader.biWidth=width;      // Set size you need
        alpha.bmiHeader.biHeight=height;    // Set size you need
        alpha.bmiHeader.biPlanes=1;
        alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
        alpha.bmiHeader.biSizeImage=0;
        alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
        alpha.bmiHeader.biClrUsed= 0;
        alpha.bmiHeader.biClrImportant= 0;
        VOID *pvBits;

        HDC darkDC = CreateCompatibleDC(NULL);

        HBITMAP darkBitmap = CreateDIBSection( darkDC,
                                                &alpha,
                                                DIB_RGB_COLORS,
                                                &pvBits,
                                                NULL,
                                                0 );

        HBITMAP olddarkBitmap = (HBITMAP) SelectObject(darkDC, darkBitmap);

        HDC lightDC = CreateCompatibleDC(NULL);

        HBITMAP lightBitmap = CreateDIBSection( lightDC,
                                                &alpha,
                                                DIB_RGB_COLORS,
                                                &pvBits,
                                                NULL,
                                                0 );

        HBITMAP oldlightBitmap = (HBITMAP) SelectObject(lightDC, lightBitmap);

        BITMAP bm_dark;
        GetObject(darkBitmap, sizeof(BITMAP), &bm_dark);

        BYTE *pPixel_dark = (LPBYTE)bm_dark.bmBits;

        int iSize = width * height;

        int j = 0;

        for (j = 0; j < iSize; ++j)
        {
            pPixel_dark[0] = 0;
            pPixel_dark[1] = 0;
            pPixel_dark[2] = 0;
            pPixel_dark[3] = 255;

            pPixel_dark+= 4;
        }

        BITMAP bm_light;
        GetObject(lightBitmap, sizeof(BITMAP), &bm_light);

        BYTE *pPixel_light = (LPBYTE)bm_light.bmBits;

        for (j = 0; j < iSize; ++j)
        {
            pPixel_light[0] = 255;
            pPixel_light[1] = 255;
            pPixel_light[2] = 255;
            pPixel_light[3] = 255;

            pPixel_light+= 4;
        }

        if ( textSize.x >= width-textBorder.left-textBorder.right )
        {           
            if(iScrollPosition <= 0)
                iScrollPosition = textSize.x + taskbar->iScrollPadLength;

            if(iScrollPosition > textSize.x + taskbar->iScrollPadLength)
                iScrollPosition = 0;
            
            painttext->apply(darkDC,
                                iScrollPosition,
                                0,
                                2*textSize.x,
                                height,
                                drawText.c_str(),
                                taskbar->bMultiLineUsed,
                                TRUE,
                                TRUE/*disableExtras*/);

            painttext->apply(darkDC,
                                iScrollPosition - taskbar->iScrollPadLength - textSize.x,
                                0,
                                2*textSize.x,
                                height,
                                drawText.c_str(),
                                taskbar->bMultiLineUsed,
                                TRUE,
                                TRUE/*disableExtras*/);

            painttext->apply(lightDC,
                                iScrollPosition,
                                0,
                                2*textSize.x,
                                height,
                                drawText.c_str(),
                                taskbar->bMultiLineUsed,
                                TRUE,
                                TRUE/*disableExtras*/);

            painttext->apply(lightDC,
                                iScrollPosition - taskbar->iScrollPadLength - textSize.x,
                                0,
                                2*textSize.x,
                                height,
                                drawText.c_str(),
                                taskbar->bMultiLineUsed,
                                TRUE,
                                TRUE/*disableExtras*/);
        }
        else
        {
            painttext->apply(darkDC, 0, 0, width, height, drawText.c_str(), taskbar->bMultiLineUsed, FALSE, TRUE/*disableExtras*/);

            painttext->apply(lightDC, 0, 0, width, height, drawText.c_str(), taskbar->bMultiLineUsed, FALSE, TRUE/*disableExtras*/);
        }

        GetObject(darkBitmap, sizeof(BITMAP), &bm_dark);
        pPixel_dark = (LPBYTE)bm_dark.bmBits;

        GetObject(lightBitmap, sizeof(BITMAP), &bm_light);
        pPixel_light = (LPBYTE)bm_light.bmBits;

        int tmpx = 0;
        int tmpy = 0;
        int xval = 255;
        int yval = 255;

        int iLeftFade = (textFade.left != 0 ? 255/textFade.left : 0);
        int iRightFade = (textFade.right != 0 ? 255/textFade.right : 0);
        int iTopFade = (textFade.top != 0 ? 255/textFade.top : 0);
        int iBottomFade = (textFade.bottom != 0 ? 255/textFade.bottom : 0);

        for (j = 0; j < iSize; ++j)
        {
            if (pPixel_dark[3] == 0)
            {
                xval = 255;
                yval = 255;

                if ( textFade.left != 0 || textFade.right != 0 || textFade.top != 0 || textFade.bottom != 0 )
                {
                    tmpx = (j+1)%width;
                    tmpy = j/width;

                    if ( tmpx == 0 )
                        tmpx = width-1;

                    if ( textFade.left != 0 && tmpx > textBorder.left && tmpx <= textBorder.left+textFade.left )
                        xval = iLeftFade*(tmpx-textBorder.left);
                    else if ( textFade.right != 0 && tmpx > width-textBorder.right-textFade.right && tmpx <= width-textBorder.right )
                        xval = iRightFade*(width-textBorder.right-tmpx);
                        
                    //Order is reverted Top<->Bottom
                    if ( textFade.bottom != 0 && tmpy > textBorder.bottom && tmpy < textBorder.bottom+textFade.bottom )
                        yval = iBottomFade*(tmpy-textBorder.bottom);
                    else if ( textFade.top != 0 && tmpy >= height-textBorder.top-textFade.top && tmpy < height-textBorder.top )
                        yval = iTopFade*(height-textBorder.top-tmpy);
                        
                    if ( xval == 255 )
                        pPixel_dark[3] = yval;
                    else if ( yval == 255 )
                        pPixel_dark[3] = xval;
                    else
                        pPixel_dark[3] = (xval*yval)/255;

                    pPixel_dark[3] = pPixel_dark[3] - ((pPixel_light[0] - pPixel_dark[0]) * pPixel_dark[3] / 255);
                }
                else
                {
                    if (pPixel_dark[0] == pPixel_light[0] && pPixel_dark[1] == pPixel_light[1] && pPixel_dark[2] == pPixel_light[2])
                        pPixel_dark[3] = 255;
                    else if (pPixel_dark[0] == 0 && pPixel_dark[1] == 0 && pPixel_dark[2] == 0 && pPixel_light[0] == 255 && pPixel_light[1] == 255 && pPixel_light[2] == 255 )
                        pPixel_dark[3] = 0;
                    else
                        pPixel_dark[3] = 255 - (pPixel_light[0] - pPixel_dark[0]);
                }

                //PreMultiply
                pPixel_dark[0] = pPixel_dark[0] * pPixel_dark[3] / 255;
                pPixel_dark[1] = pPixel_dark[1] * pPixel_dark[3] / 255;
                pPixel_dark[2] = pPixel_dark[2] * pPixel_dark[3] / 255;
            }
            else
            {
                pPixel_dark[0] = 0;
                pPixel_dark[1] = 0;
                pPixel_dark[2] = 0;
                pPixel_dark[3] = 0;
            }
                            
            pPixel_dark+= 4;    
            pPixel_light+= 4;   
        }

        BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, textAlphaTransparency, AC_SRC_ALPHA };

        AlphaBlend(desthDC, x, y, width, height, darkDC, 0, 0, width, height, blendFunction);

        darkBitmap = (HBITMAP) SelectObject(darkDC, olddarkBitmap);
        DeleteDC(darkDC);
        DeleteObject(darkBitmap);
        DeleteObject(olddarkBitmap);

        lightBitmap = (HBITMAP) SelectObject(lightDC, oldlightBitmap);
        DeleteDC(lightDC);
        DeleteObject(lightBitmap);
        DeleteObject(oldlightBitmap);
    }
    else
    {
        if ( textSize.x >= width-textBorder.left-textBorder.right )
        {   
            if(iScrollPosition <= 0)
                iScrollPosition = textSize.x + taskbar->iScrollPadLength;

            if(iScrollPosition > textSize.x + taskbar->iScrollPadLength)
                iScrollPosition = 0;
            
            painttext->apply(desthDC,
                                x + iScrollPosition,
                                y,
                                2*textSize.x,
                                height,
                                drawText.c_str(),
                                taskbar->bMultiLineUsed,
                                TRUE,
                                TRUE/*disableExtras*/);

            painttext->apply(desthDC,
                                x + iScrollPosition - taskbar->iScrollPadLength - textSize.x,
                                y,
                                2*textSize.x,
                                height,
                                drawText.c_str(),
                                taskbar->bMultiLineUsed,
                                TRUE,
                                TRUE/*disableExtras*/);
        }
        else
        {
            painttext->apply(desthDC, x, y, width, height, drawText.c_str(), taskbar->bMultiLineUsed);
        }
    }
    
    //Reset Region in which text drawing is allowed
    SelectClipRgn(desthDC, NULL);
    //--------------------------------------------------------------------------------
    // stop rendering text
    //--------------------------------------------------------------------------------
}


bool TaskbarButton::CheckState(int iDisplayState)
{
    if ( ( iCurrentState == STATE_NORMAL 
            && (iDisplayState < 0 || iDisplayState == 0 || iDisplayState == 3 || iDisplayState == 4) )
                || ( iCurrentState == STATE_SELECTED 
                        && (iDisplayState < 0 || iDisplayState == 1 || iDisplayState == 3 || iDisplayState == 5) )
                    || ( iCurrentState == STATE_MINIMIZED 
                            && (iDisplayState < 0 || iDisplayState == 2 || iDisplayState == 4 || iDisplayState == 5) )
                        || ( iDisplayState == 6 && bIsFlashing ) )
        return true;
    else
        return false;
}