/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2009 The LS-Universe Team

    Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "Utilities.h"

#include <shellapi.h>

#define INVALID_ICON(x) ((NULL == (x)) || (1 == (int)(x)) || (-1 == (int)(x)))

#define LM_GETDESKTOPOF 9361

string GetTaskClass( HWND hTask )
{
    char windowClass[MAX_LINE_LENGTH] = {0};

    if (hTask)
        GetClassName( hTask, windowClass, MAX_LINE_LENGTH );

    return (string) windowClass;
}

string GetTaskWindowText( HWND hTask )
{
    char windowText[MAX_LINE_LENGTH] = {0};

    if (hTask)
        GetWindowText( hTask, windowText, MAX_LINE_LENGTH );

    return (string) windowText;
}

string GetTaskFileName( HWND hTask )
{
    HINSTANCE hLib = LoadLibraryA("PSAPI.DLL");
    if (!hLib) 
        return GetTaskClass( hTask );

    BOOL (WINAPI *lpfEnumProcessModules)( HANDLE, HMODULE *, DWORD, LPDWORD );
    DWORD (WINAPI *lpfGetModuleFileNameEx)( HANDLE, HMODULE, LPTSTR, DWORD );

    lpfEnumProcessModules = (BOOL(WINAPI *)(HANDLE, HMODULE *, DWORD, LPDWORD)) GetProcAddress( hLib, "EnumProcessModules" ) ;
    lpfGetModuleFileNameEx =(DWORD (WINAPI *)(HANDLE, HMODULE, LPTSTR, DWORD )) GetProcAddress( hLib, "GetModuleFileNameExA" ) ;
    
    if( lpfEnumProcessModules == NULL || lpfGetModuleFileNameEx == NULL )
    {
        // some of the functions did not load properly... return out
        FreeLibrary(hLib);
        return GetTaskClass( hTask );
    }
    else
    {
        // get the exe names
        HANDLE hProcess;
        HMODULE hModule;
        DWORD dwSize2;
        char szFilename[MAX_LINE_LENGTH] = {0};

        // open the process so we can get a handle
        DWORD dwPID;
        GetWindowThreadProcessId(hTask, &dwPID);
        hProcess = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, false, dwPID);
        if (hProcess)
        {
            if (lpfEnumProcessModules(hProcess, &hModule, sizeof(hModule), &dwSize2))
            {
                // get path name
                if (!lpfGetModuleFileNameEx(hProcess, hModule, szFilename, sizeof(szFilename)))
                {
                    CloseHandle(hProcess); // be nice. :)
                    return GetTaskClass( hTask );
                }
            }
            CloseHandle(hProcess); // be nice. :)
            
            return (string) szFilename;
        }
        
        return GetTaskClass( hTask );
    }
}

// trim leading and trailing whitespace
string trim(const string &toTrim)
{
    int start = 0;
    int end = toTrim.length() - 1;
    
    while(start <= end && isspace(toTrim[start]))
        start++;
    
    while(start <= end && isspace(toTrim[end]))
        end--;
    
    return toTrim.substr(start, end - start + 1);
}

string IntToString(int input)
{
    char buffer[32] = {0};
    itoa( input, buffer, 10 );
    return (string) buffer;
}

int SplitString(const string &process, string &one, string &two, string &three, string &four, string &five, string &six, string &seven)
{
    int count = 0;
    one = "";
    two = "";
    three = "";
    four = "";
    five = "";
    six = "";
    seven = "";
    char cone[256] = {0};
    char ctwo[256] = {0};
    char cthree[256] = {0};
    char cfour[256] = {0};
    char cfive[256] = {0};
    char csix[256] = {0};
    char cseven[256] = {0};
    char *cbuffers[] = {cone, ctwo, cthree, cfour, cfive, csix, cseven};
        
    int numTokens = LCTokenize( process.c_str(), cbuffers, 7, NULL);
    if( numTokens >= 1 )
    {
        one = cone;
        count++;
        
        if ( numTokens >= 2 )
        {
            two = ctwo;
            count++;
        
            if ( numTokens >= 3 )
            {
                three = cthree;
                count++;
            
                if ( numTokens >= 4 )
                {
                    four = cfour;
                    count++;

                    if ( numTokens >= 5 )
                    {
                        five = cfive;
                        count++;

                        if ( numTokens >= 6 )
                        {
                            six = csix;
                            count++;
        
                            if ( numTokens >= 7 )
                            {
                                seven = cseven;
                                count++;
                            }
                        }
                    }
                }
            }
        }
    }

    return count;
}

string iGetxRCString(const string &uniqueKey, const string &baseName, const string &defaultVal)
{
    char szValue[MAX_LINE_LENGTH] = { 0 };
    GetxRCString(szValue, uniqueKey.c_str(), baseName.c_str(), defaultVal.c_str());
    return (string) szValue;
}

string iGetx2RCString(const string &uniqueKey, const string &baseName, const string &defaultVal)
{
    char szValue[MAX_LINE_LENGTH] = { 0 };
    Getx2RCString(szValue, uniqueKey.c_str(), baseName.c_str(), defaultVal.c_str());
    return (string) szValue;
}

string iGetxRCLine(const string &uniqueKey, const string &baseName, const string &defaultVal)
{
    char szValue[MAX_LINE_LENGTH] = { 0 };
    GetxRCLine(szValue, uniqueKey.c_str(), baseName.c_str(), defaultVal.c_str());
    return (string) szValue;
}

string iGetx2RCLine(const string &uniqueKey, const string &baseName, const string &defaultVal)
{
    char szValue[MAX_LINE_LENGTH] = { 0 };
    Getx2RCLine(szValue, uniqueKey.c_str(), baseName.c_str(), defaultVal.c_str());
    return (string) szValue;
}

void PrefixedLSSetVariable(string prefix, string var, string value)
{
    if ( !var.empty() )
        prefix.append(var);

    LSSetVariable(prefix.c_str(), value.c_str());
}

StringList split(string str, string sep)
{
    StringList substrs;
    substrs.clear();

    if ( str.empty() )
        return substrs;

    int start = 0;
    int end = str.find(sep, start);
    
    while(end != string::npos)
    {
        substrs.push_back(str.substr(start, end - start - sep.length() + 1));
        start = end + sep.length();
        end = str.find(sep, start);
    }
    
    substrs.push_back(str.substr(start));

    return substrs;
}

bool IsSubString(const string &big, const string &small)
{
    if( match( small.c_str(), big.c_str() ) )
        return TRUE;
    else
        return FALSE;
}

bool IsOnTaskbarList(HWND hWnd, StringList theList, bool emptyValue)
{   
    //Matching Lists
    if ( theList.empty() )
        return emptyValue;

    string tmpTitle = GetTaskWindowText( hWnd );
    string tmpClass = GetTaskClass( hWnd );
    
    for (StringListIterator it = theList.begin(); it != theList.end(); ++it)
    {
        if (IsSubString(tmpTitle, (*it).c_str()) || IsSubString(tmpClass, (*it).c_str()))
            return TRUE;
    }
    return FALSE;
}

void GetTaskIcon(HWND hTask, HICON &hIcon, bool bBigIcon)
{
    int nTimeOut = 1000;

    if(bBigIcon)
    {
        if(SendMessageTimeout(   hTask
                                ,WM_GETICON
                                ,ICON_BIG
                                ,0
                                ,/*SMTO_BLOCK|*/SMTO_ABORTIFHUNG
                                ,nTimeOut
                                ,(LPDWORD)&hIcon
                                ) && !INVALID_ICON(hIcon) )
            return;

        hIcon = (HICON)GetClassLong(hTask, GCL_HICON);
        if(!INVALID_ICON(hIcon))
            return;

        if(SendMessageTimeout(   hTask
                                ,WM_QUERYDRAGICON
                                ,ICON_BIG
                                ,0
                                ,/*SMTO_BLOCK|*/SMTO_ABORTIFHUNG
                                ,nTimeOut
                                ,(LPDWORD)&hIcon
                            ) && !INVALID_ICON(hIcon))
            return;

        if(SendMessageTimeout(   hTask
                            ,WM_GETICON
                            ,ICON_SMALL
                            ,0
                            ,/*SMTO_BLOCK|*/SMTO_ABORTIFHUNG
                            ,nTimeOut
                            ,(LPDWORD)&hIcon
                        ) && !INVALID_ICON(hIcon))
            return;

        hIcon = (HICON)GetClassLong(hTask, GCL_HICONSM);
        if(!INVALID_ICON(hIcon))
            return;

        if(SendMessageTimeout(   hTask
                                ,WM_QUERYDRAGICON
                                ,ICON_SMALL
                                ,0
                                ,/*SMTO_BLOCK|*/SMTO_ABORTIFHUNG
                                ,nTimeOut
                                ,(LPDWORD)&hIcon
                            ) && !INVALID_ICON(hIcon))
            return; 
    }
    else
    {
        if(SendMessageTimeout(   hTask
                                ,WM_GETICON
                                ,ICON_SMALL
                                ,0
                                ,/*SMTO_BLOCK|*/SMTO_ABORTIFHUNG
                                ,nTimeOut
                                ,(LPDWORD)&hIcon
                            ) && !INVALID_ICON(hIcon))
            return;

        hIcon = (HICON)GetClassLong(hTask, GCL_HICONSM);
        if(!INVALID_ICON(hIcon))
            return;

        if(SendMessageTimeout(   hTask
                                ,WM_QUERYDRAGICON
                                ,ICON_SMALL
                                ,0
                                ,/*SMTO_BLOCK|*/SMTO_ABORTIFHUNG
                                ,nTimeOut
                                ,(LPDWORD)&hIcon
                            ) && !INVALID_ICON(hIcon))
            return;
        
        if(SendMessageTimeout(   hTask
                                ,WM_GETICON
                                ,ICON_BIG
                                ,0
                                ,/*SMTO_BLOCK|*/SMTO_ABORTIFHUNG
                                ,nTimeOut
                                ,(LPDWORD)&hIcon
                                ) && !INVALID_ICON(hIcon) )
            return;

        hIcon = (HICON)GetClassLong(hTask, GCL_HICON);
        if(!INVALID_ICON(hIcon))
            return;

        if(SendMessageTimeout(   hTask
                                ,WM_QUERYDRAGICON
                                ,ICON_BIG
                                ,0
                                ,/*SMTO_BLOCK|*/SMTO_ABORTIFHUNG
                                ,nTimeOut
                                ,(LPDWORD)&hIcon
                            ) && !INVALID_ICON(hIcon))
            return;
    }

    //All Failed
    hIcon = LoadIcon(NULL, MAKEINTRESOURCE(IDI_APPLICATION));
}

int ParseIconLocation(LPSTR szIconFile)
{
    LPSTR pszComma = strrchr(szIconFile, ',');

    if (!pszComma)
        return -1;

    *pszComma = 0;
    return atoi(pszComma + 1);
}

bool GetIconOnCustomList(HWND hWnd, HICON &hIcon, StringList theList, bool bBigIcon)
{
    if ( !theList.empty() )
    {
        string tmpTitle = GetTaskWindowText( hWnd );
        string tmpClass = GetTaskClass( hWnd );

        for (StringListIterator it = theList.begin(); it != theList.end(); ++it)
        {
            char search[64], iconPath[MAX_LINE_LENGTH];
            char *buffers[] = {search, iconPath};

            LCTokenize((*it).c_str(), buffers, 2, NULL);
            
            if (IsSubString(tmpTitle, search) || IsSubString(tmpClass, search))
            {
                int nIconIndex = ParseIconLocation(iconPath);

                if (nIconIndex >= 0)
                {
                    ExtractIconEx(iconPath, nIconIndex, (!bBigIcon ? NULL : &hIcon), (!bBigIcon ? &hIcon : NULL), 1);
                }
                else
                {
                    SHFILEINFO shFileInfo;

                    SHGetFileInfo(iconPath, NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_ICON | (bBigIcon ? SHGFI_LARGEICON : SHGFI_SMALLICON) );

                    hIcon = shFileInfo.hIcon;
                }
                return false;
            }
        }
    }

    GetTaskIcon(hWnd, hIcon, bBigIcon);
    return true;
}

bool IsValidWindow(const HWND &hWnd, int iHomeDesk)
{
    LRESULT lResult = -1;
    /* Ask VWM if the hWnd is on the current desktop */
    if ( iHomeDesk == -1 )
        lResult = SendMessage(GetLitestepWnd(), LM_GETDESKTOPOF, (WPARAM)hWnd, 0x00000001);
    /* Ask VWM if the hWnd is on a specified desktop */
    else
        lResult = SendMessage(GetLitestepWnd(), LM_GETDESKTOPOF, (WPARAM)hWnd, 0x00000000);
    
    /*
     * Some VWMs return -1 for any window that it doesn't 'manage', so we should accept those
     * windows as always being on the current desktop.  Otherwise, we need to check the 'active'
     * bit in the uppper nible of the return value.  If it is set, then the window was on the
     * current desktop.
    */
    if (0 > lResult)
        return true;

    if ( iHomeDesk != -1 )
    {
        if ( (int) lResult == iHomeDesk )
            return true;
        else 
            return false;
    }
    else
    {
        if ( 0x01000000 == (lResult & 0xFF000000) || lResult == 0 )
            return true;
        else 
            return false;
    }
}

void GetxRCList(char *text, StringList &theList, bool trim)
{
    LPVOID f = LCOpen(NULL);

    if(f != NULL)
    {
        char lineBuffer[MAX_LINE_LENGTH] = {0};

        while(LCReadNextConfig(f, text, lineBuffer, MAX_LINE_LENGTH))
        {
            char temp[64], configline[MAX_LINE_LENGTH];
            char *buffers[] = {temp};
            
            if(LCTokenize(lineBuffer, buffers, 1, configline) >= 1)
            {
                CString worker = configline;
                if ( trim )
                    worker.Replace("\"", "");

                theList.insert(theList.end(), (string) worker);
            }
        }
        
        LCClose(f);
    }
}

StringList GetxRCLineList(const string &uniqueKey)
{
    StringList resultList;
    LPVOID f = LCOpen(NULL);

    if(f != NULL)
    {
        char lineBuffer[MAX_LINE_LENGTH];

        while(LCReadNextConfig(f, uniqueKey.c_str(), lineBuffer, MAX_LINE_LENGTH))
        {
            char temp[64], configline[MAX_LINE_LENGTH];
            char *buffers[] = {temp};
            
            if(LCTokenize(lineBuffer, buffers, 1, configline) >= 1)
            {
                resultList.insert(resultList.end(), configline);
            }
        }
        
        LCClose(f);
    }

    return resultList;
}
