/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2009 The LS-Universe Team

    Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "Taskbar.h"

#include "VerInfo.h"

#include "WindowUtilities.h"

#pragma comment(lib, "../xPaintClass_SDK/SDK/xPaintClass-1.0.lib")

//xStatsClass
ProcessStatsRequestFunc m_pProcessStatsRequest = NULL;

string str_LoadedxStatsClass = "";

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ModulesCallback
//
BOOL __stdcall ModulesCallback(LPCSTR pszPath, DWORD dwFlags, LPARAM lParam)
{
    CString strCheck = pszPath;
    strCheck.MakeLower();
    int iPos = strCheck.Find("xstatsclass-");
    if ( iPos != -1 )
    {
        str_LoadedxStatsClass = strCheck.Right(strCheck.GetLength()-iPos);
        return FALSE;
    }

    return TRUE;
}

typedef list<Taskbar *> xTaskbarList;
typedef list<Taskbar *>::iterator xTaskbarListIterator;

// Bang commands
void BangAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs);
void BangCreate(HWND hwndCaller, LPCSTR pszArgs);
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs);
void BangDisplay(HWND hwndCaller, LPCSTR pszArgs);
void BangHide(HWND hwndCaller, LPCSTR pszArgs);
void BangLSBoxHook(HWND hwndCaller, LPCSTR pszArgs);
void BangMove(HWND hwndCaller, LPCSTR pszArgs);
void BangMoveBy(HWND hwndCaller, LPCSTR pszArgs);
void BangRefresh(HWND hwndCaller, LPCSTR pszArgs);
void BangReposition(HWND hwndCaller, LPCSTR pszArgs);
void BangRepositionBy(HWND hwndCaller, LPCSTR pszArgs);
void BangResize(HWND hwndCaller, LPCSTR pszArgs);
void BangResizeBy(HWND hwndCaller, LPCSTR pszArgs);
void BangScrollButtons(HWND hwndCaller, LPCSTR pszArgs);
void BangSetAlpha(HWND hwndCaller, LPCSTR pszArgs);
void BangShow(HWND hwndCaller, LPCSTR pszArgs);
void BangSwitch(HWND hwndCaller, LPCSTR pszArgs);
void BangToggle(HWND hwndCaller, LPCSTR pszArgs);

void BangMinimize(HWND hwndCaller, LPCSTR pszArgs);
void BangRestore(HWND hwndCaller, LPCSTR pszArgs);
void BangCascade(HWND hwndCaller, LPCSTR pszArgs);
void BangTileWindows(HWND hwndCaller, LPCSTR pszArgs);

#define StringCchPrintf _snprintf
#define StringCchLength(a, b, c) *(c) = strlen((a)), S_OK

BOOL WINAPI EnumWindowsProc(HWND hWnd, LPARAM lParam)
{
    if ( IsAppWindow(hWnd) )
    {
        if ( lParam == -1 )
        {
            MinimizeWindow(hWnd);
        }
        else
        {
            RestoreWindow(hWnd);
        }
    }
    else if ( lParam == -1 && (GetWindowLong(hWnd, GWL_USERDATA) != MAGIC_DWORD) )
    {
        if ( IsWindow(hWnd) && IsWindowVisible(hWnd) )
        {
            if ( GetParent( hWnd ) )
                return TRUE;
            
            if ( GetWindow(hWnd, GW_OWNER) )
                return TRUE;
    
            long Style = GetWindowLong(hWnd, GWL_STYLE);
            if ( Style & WS_MINIMIZEBOX )
            {   
                POINT pt;
                GetCursorPos(&pt);
                ::SendMessage(hWnd, WM_SYSCOMMAND, (WPARAM) SC_MINIMIZE, MAKELPARAM(pt.x, pt.y));
            }
        }
    }

    return TRUE;
}


HINSTANCE hInstance;
HWND messageHandler;

xTaskbarList taskbarList;

int lsMessages[] = {
    LM_GETREVID,
    LM_REFRESH,
    0
};

int bitsperpixel = QueryPixelDepth();

Taskbar *lookupTaskbar( const string &name = "")
{
    if (!name.empty())
    {
        for(xTaskbarListIterator tblit = taskbarList.begin(); tblit != taskbarList.end(); ++tblit)
        {
            if(_stricmp(name.c_str(), (*tblit)->GetName().c_str()) == NULL)
                return *tblit;
        }
        return NULL;
    }
    else if (taskbarList.size() != 0)
    {
        return *taskbarList.begin();
    }
    else
    {
        return NULL;
    }
}

#define LM_UPDATEBG (WM_USER + 1)

LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
    {
        case LM_GETREVID:
        {
            UINT uLength;
            StringCchPrintf((char*)lParam, 64, "%s.dll: %s", V_NAME, V_VERSION);
            
            if (SUCCEEDED(StringCchLength((char*)lParam, 64, &uLength)))
                return uLength;

            lParam = NULL;
            return 0;
        }

        case LM_REFRESH:
        {
            for(xTaskbarListIterator tblit = taskbarList.begin(); tblit != taskbarList.end(); ++tblit)
            {
                (*tblit)->Reconfigure(true);
            }

            return 0;
        }

        case LM_UPDATEBG:
        {
            PaintDesktopEx(0, 0, 0, 0, 0, 0, 0, TRUE);

            for(xTaskbarListIterator tblit = taskbarList.begin(); tblit != taskbarList.end(); ++tblit)
            {
                (*tblit)->Repaint();
            }

            return 0;
        }

        case WM_DISPLAYCHANGE:
        case WM_SETTINGCHANGE:
        {
            PostMessage(hWnd, LM_UPDATEBG, 0, 0);

            return 0;
        }
    }

    return DefWindowProc(hWnd, message, wParam, lParam);
}

// Why are we advoiding the crt startup code? The start up code is not large
// enough to cause bloat (except possibly a long system call, and I suspect that
// would need done eventually anyway) (default _DllMainCRTStartup is in
// (MSVCDIR\crt\src\dllcrt0.c) The elimination of the call to _CRT_INIT is 
// also causing warnings with MSVC7
BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes)
{
    // We don't need thread notifications for what we're doing.  Thus, get
    // rid of them, thereby eliminating some of the overhead of this DLL
    DisableThreadLibraryCalls(hInst);
    return TRUE;
}

int initModuleEx(HWND hParent, HINSTANCE hInstance, const char *lsPath)
{
    WNDCLASSEX wc;

    wc.cbSize = sizeof(WNDCLASSEX);
    wc.style = CS_GLOBALCLASS;
    wc.lpfnWndProc = Taskbar::WindowProcedure;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = sizeof(Taskbar *);
    wc.hInstance = hInstance;
    wc.hbrBackground = 0;
    wc.hCursor = LoadCursor(0, IDC_ARROW);
    wc.hIcon = 0;
    wc.lpszMenuName = 0;
    wc.lpszClassName = "xTaskbar";
    wc.hIconSm = 0;

    RegisterClassEx(&wc);

    wc.cbSize = sizeof(WNDCLASSEX);
    wc.style = CS_GLOBALCLASS;
    wc.lpfnWndProc = MessageHandlerProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hbrBackground = 0;
    wc.hCursor = 0;
    wc.hIcon = 0;
    wc.lpszMenuName = 0;
    wc.lpszClassName = "xTaskbarMessageHandler";
    wc.hIconSm = 0;

    RegisterClassEx(&wc);

    wc.cbSize = sizeof(WNDCLASSEX);
    wc.style = CS_GLOBALCLASS;
    wc.lpfnWndProc = DefWindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hbrBackground = 0;
    wc.hCursor = 0;
    wc.hIcon = 0;
    wc.lpszMenuName = 0;
    wc.lpszClassName = "xTaskbarOwner";
    wc.hIconSm = 0;

    RegisterClassEx(&wc);

    messageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
        "xTaskbarMessageHandler",
        0,
        0,
        0, 0, 0, 0, 
        HWND_MESSAGE,
        0,
        hInstance,
        0);

    if (!messageHandler)
        return 1;
    
    SendMessage(GetLitestepWnd(),
        LM_REGISTERMESSAGE,
        (WPARAM) messageHandler,
        (LPARAM) lsMessages);

    ::hInstance = hInstance;

    HRESULT hr = EnumLSData(2 /*ELD_MODULES*/, (FARPROC)ModulesCallback, (LPARAM) NULL);

    if ( !m_pProcessStatsRequest )
    {
        m_pProcessStatsRequest = (ProcessStatsRequestFunc) GetProcAddress(GetModuleHandle(str_LoadedxStatsClass.c_str()), "processStatsRequest");
    } 

    AddBangCommand("!xTaskbarAlwaysOnTop", BangAlwaysOnTop);
    AddBangCommand("!xTaskbarCreate", BangCreate);
    AddBangCommand("!xTaskbarDestroy", BangDestroy);
    AddBangCommand("!xTaskbarDisplay", BangDisplay);
    AddBangCommand("!xTaskbarHide", BangHide);
    AddBangCommand("!xTaskbarLSBoxHook", BangLSBoxHook);
    AddBangCommand("!xTaskbarMove", BangMove);
    AddBangCommand("!xTaskbarMoveBy", BangMoveBy);
    AddBangCommand("!xTaskbarRefresh", BangRefresh);
    AddBangCommand("!xTaskbarReposition", BangReposition);
    AddBangCommand("!xTaskbarRepositionBy", BangRepositionBy);
    AddBangCommand("!xTaskbarResize", BangResize);
    AddBangCommand("!xTaskbarResizeBy", BangResizeBy);
    AddBangCommand("!xTaskbarScrollButtons", BangScrollButtons);
    if(IsOS(OS_2KXP))
        AddBangCommand("!xTaskbarSetAlpha", BangSetAlpha);
    AddBangCommand("!xTaskbarShow", BangShow);
    AddBangCommand("!xTaskbarSwitch", BangSwitch);
    AddBangCommand("!xTaskbarToggle", BangToggle);

    AddBangCommand("!xTaskbarMinimizeAll", BangMinimize);
    AddBangCommand("!xTaskbarRestoreAll", BangRestore);
    AddBangCommand("!xTaskbarCascadeAll", BangCascade);
    AddBangCommand("!xTaskbarTileAll", BangTileWindows);

    StringList taskbarNames;

    LPVOID f = LCOpen(NULL);
    if(f != NULL)
    {
        char lineBuffer[MAX_LINE_LENGTH];

        while(LCReadNextConfig(f, "*xtaskbar", lineBuffer, MAX_LINE_LENGTH))
        {
            char temp[64], name[64];
            char *buffers[] = {temp, name};
            
            if(LCTokenize(lineBuffer, buffers, 2, NULL) == 2)
            {
                taskbarNames.insert(taskbarNames.end(), name);
            }
        }

        LCClose(f);
    }

    taskbarList.clear();

    if ( !taskbarNames.empty() )
    {
        for(StringListIterator sit = taskbarNames.begin(); sit != taskbarNames.end(); ++sit)
        {
            if ( !Getx2RCBool((*sit).c_str(), "LoadInBox") && !Getx2RCBool((*sit).c_str(), "NoCreateAtStartup") )
            {
                PrefixedLSSetVariable((*sit), "HoverTask", "");

                Taskbar *taskbar = new Taskbar((*sit));
            
                taskbarList.insert(taskbarList.end(), taskbar);
                
                taskbar->Load(hInstance);
            }
        }
    }
    else if ( !Getx2RCBool("xTaskbar", "LoadInBox") && !Getx2RCBool("xTaskbar", "NoCreateAtStartup") )
    {
        PrefixedLSSetVariable("xTaskbar", "HoverTask", "");

        Taskbar *taskbar = new Taskbar("xTaskbar");
        
        taskbarList.insert(taskbarList.end(), taskbar);
            
        taskbar->Load(hInstance);
    }

    return 0;
}

void quitModule(HINSTANCE hInstance)
{
    RemoveBangCommand("!xTaskbarAlwaysOnTop");
    RemoveBangCommand("!xTaskbarCreate");
    RemoveBangCommand("!xTaskbarDestroy");
    RemoveBangCommand("!xTaskbarDisplay");
    RemoveBangCommand("!xTaskbarHide");
    RemoveBangCommand("!xTaskbarLSBoxHook");
    RemoveBangCommand("!xTaskbarMove");
    RemoveBangCommand("!xTaskbarMoveBy");
    RemoveBangCommand("!xTaskbarRefresh");
    RemoveBangCommand("!xTaskbarReposition");
    RemoveBangCommand("!xTaskbarRepositionBy");
    RemoveBangCommand("!xTaskbarResize");
    RemoveBangCommand("!xTaskbarResizeBy");
    RemoveBangCommand("!xTaskbarScrollButtons");
    if(IsOS(OS_2KXP))
        RemoveBangCommand("!xTaskbarSetAlpha");
    RemoveBangCommand("!xTaskbarShow");
    RemoveBangCommand("!xTaskbarSwitch");
    RemoveBangCommand("!xTaskbarToggle");

    RemoveBangCommand("!xTaskbarMinimizeAll");
    RemoveBangCommand("!xTaskbarRestoreAll");
    RemoveBangCommand("!xTaskbarCascadeAll");
    RemoveBangCommand("!xTaskbarTileAll");

    for(xTaskbarListIterator tblit = taskbarList.begin(); tblit != taskbarList.end(); ++tblit)
    {
        (*tblit)->Hide();
        delete (*tblit);
        (*tblit) = NULL;
    }
    taskbarList.clear();

    SendMessage(GetLitestepWnd(),
        LM_UNREGISTERMESSAGE,
        (WPARAM) messageHandler,
        (LPARAM) lsMessages);

    DestroyWindow(messageHandler);

    UnregisterClass("xTaskbar", hInstance);
    UnregisterClass("xTaskbarMessageHandler", hInstance);
    UnregisterClass("xTaskbarOwner", hInstance);
}

// !xTaskbarCreate <taskbar>
void BangCreate(HWND hwndCaller, LPCSTR pszArgs)
{
    char name[64];
    char *buffers[] = {name};

    LCTokenize(pszArgs, buffers, 1, NULL);

    string args = name;
    if ( args.empty() )
        args = "xTaskbar";

    if ( !Getx2RCBool(args.c_str(), "LoadInBox") )
    {
        Taskbar *taskbar = new Taskbar(args);
        taskbarList.insert(taskbarList.end(), taskbar);
        taskbar->Load(hInstance);
    }
    else
    {
        MessageBox(NULL, ("You DON'T need to create a xTaskbar, when you want to hook it!\nRemove \"" + args + "LoadInBox\"!").c_str(), "User Information", MB_OK | MB_SETFOREGROUND | MB_ICONEXCLAMATION);
    }
}

// !xTaskbarDestroy <taskbar>
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs)
{
    char name[64];
    char *buffers[] = {name};

    LCTokenize(pszArgs, buffers, 1, NULL);

    Taskbar *taskbar = lookupTaskbar(name);

    if(taskbar)
    {
        taskbar->Hide();
        taskbarList.remove(taskbar);
        delete taskbar;
        taskbar = NULL;
    }
}

// !xTaskbarAlwaysOnTop <on|true, off|false, toggle|switch>
void BangAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs)
{
    char name[64], option[16];
    char *buffers[] = {name, option};

    int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

    if (numTokens == 1)
    {
        Taskbar *taskbar = lookupTaskbar();

        if(taskbar)
        {
            if (_stricmp(name, "false") == NULL || _stricmp(name, "off") == NULL)
                taskbar->SetAlwaysOnTop(false);
            else if (_stricmp(name, "toggle") == NULL || _stricmp(name, "switch") == NULL)
                taskbar->SetAlwaysOnTop(!taskbar->IsAlwaysOnTop());
            else
                taskbar->SetAlwaysOnTop(true);
        }
    }
    else if (numTokens == 2)
    {
        Taskbar *taskbar = lookupTaskbar(name);

        if(taskbar)
        {
            if (_stricmp(option, "false") == NULL || _stricmp(option, "off") == NULL)
                taskbar->SetAlwaysOnTop(false);
            else if (_stricmp(option, "toggle") == NULL || _stricmp(option, "switch") == NULL)
                taskbar->SetAlwaysOnTop(!taskbar->IsAlwaysOnTop());
            else 
                taskbar->SetAlwaysOnTop(true);
        }
    }
}

// !xTaskbarShow
void BangShow(HWND hwndCaller, LPCSTR pszArgs)
{
    char name[64];
    char *buffers[] = {name};

    LCTokenize(pszArgs, buffers, 1, NULL);

    Taskbar *taskbar = lookupTaskbar(name);

    if(taskbar)
    {
        if ( taskbar->IsShowNotAllowed() )
            return;

        taskbar->Show();
    }
}

// !xTaskbarHide
void BangHide(HWND hwndCaller, LPCSTR pszArgs)
{
    char name[64];
    char *buffers[] = {name};

    LCTokenize(pszArgs, buffers, 1, NULL);

    Taskbar *taskbar = lookupTaskbar(name);

    if(taskbar)
    {
        if ( taskbar->IsHideNotAllowed() )
            return;

        taskbar->Hide();
    }
}

// !xTaskbarToggle
void BangToggle(HWND hwndCaller, LPCSTR pszArgs)
{
    char name[64];
    char *buffers[] = {name};

    LCTokenize(pszArgs, buffers, 1, NULL);

    Taskbar *taskbar = lookupTaskbar(name);

    if(taskbar)
    {
        if(taskbar->IsVisible())
        {
            if ( taskbar->IsHideNotAllowed() )
                return;

            taskbar->Hide();
        }
        else
        {
            if ( taskbar->IsShowNotAllowed() )
                return;

            taskbar->Show();
        }
    }
}

// !xTaskbarMove <x> <y> [<steps> <time>]
void BangMove(HWND hwndCaller, LPCSTR pszArgs)
{
    int x = 0;
    int y = 0;
    char name[64], A[16], B[16], C[16], D[16];
    char *buffers[] = {name, A, B, C, D};

    int tokens = LCTokenize(pszArgs, buffers, 5, NULL);

    if (tokens == 2 || tokens == 4)
    {
        Taskbar *taskbar = lookupTaskbar();

        if(taskbar)
        {
            if ( !taskbar->GetBox() )
            {
                x = xParseCoord(name, 0, iScreenWidth, false);
                y = xParseCoord(A, 0, iScreenHeight, true);
            }
            else
            {
                RECT rc;
                GetClientRect(taskbar->GetBox(), &rc);
                x = xParseCoord(name, 0, rc.right, false);
                y = xParseCoord(A, 0, rc.bottom, true);
            }

            if ( tokens == 4 )
            {
                int step = atoi( B );
                int time = atoi( C );

                taskbar->Reposition(1, x, y, 0, 0, step, time);
            }
            else
            {
                taskbar->Reposition(1, x, y, 0, 0);
            }
        }
    }
    else if (tokens == 3 || tokens == 5)
    {
        Taskbar *taskbar = lookupTaskbar(name);

        if(taskbar)
        {
            if ( !taskbar->GetBox() )
            {
                x = xParseCoord(A, 0, iScreenWidth, false);
                y = xParseCoord(B, 0, iScreenHeight, true);
            }
            else
            {
                RECT rc;
                GetClientRect(taskbar->GetBox(), &rc);
                x = xParseCoord(A, 0, rc.right, false);
                y = xParseCoord(B, 0, rc.bottom, true);
            }

            if ( tokens == 5 )
            {
                int step = atoi( C );
                int time = atoi( D );

                taskbar->Reposition(1, x, y, 0, 0, step, time);
            }
            else
            {
                taskbar->Reposition(1, x, y, 0, 0);
            }
        }
    }
}

// !xTaskbarMoveBy <x> <y> [<steps> <time>]
void BangMoveBy(HWND hwndCaller, LPCSTR pszArgs)
{
    int x = 0;
    int y = 0;
    char name[64], A[16], B[16], C[16], D[16];
    char *buffers[] = {name, A, B, C, D};

    int tokens = LCTokenize(pszArgs, buffers, 5, NULL);

    if (tokens == 2 || tokens == 4)
    {
        Taskbar *taskbar = lookupTaskbar();

        if(taskbar)
        {
            x = atoi( name );
            y = atoi( A );
        
            if ( tokens == 4 )
            {
                int step = atoi( B );
                int time = atoi( C );

                taskbar->Reposition(1, x, y, 0, 0, step, time, true);
            }
            else
            {
                taskbar->Reposition(1, x, y, 0, 0, 0, 0, true);
            }
        }
    }
    else if (tokens == 3 || tokens == 5)
    {
        Taskbar *taskbar = lookupTaskbar(name);

        if(taskbar)
        {
            x = atoi( A );
            y = atoi( B );
        
            if ( tokens == 5 )
            {
                int step = atoi( C );
                int time = atoi( D );

                taskbar->Reposition(1, x, y, 0, 0, step, time, true);
            }
            else
            {
                taskbar->Reposition(1, x, y, 0, 0, 0, 0, true);
            }
        }
    }
}

// !xTaskbarReposition <x> <y> <width> <height> [<steps>] [<time>]
void BangReposition(HWND hwndCaller, LPCSTR pszArgs)
{
    int x = 0;
    int y = 0;
    int w = 64;
    int h = 64;
    char name[64], A[16], B[16], C[16], D[16], E[16], F[16];
    char *buffers[] = {name, A, B, C, D, E, F};

    int tokens = LCTokenize(pszArgs, buffers, 7, NULL);

    if (tokens == 4 || tokens == 6)
    {
        Taskbar *taskbar = lookupTaskbar();

        if(taskbar)
        {
            if ( !taskbar->GetBox() )
            {
                x = xParseCoord(name, 0, iScreenWidth, false);
                y = xParseCoord(A, 0, iScreenHeight, true);
                w = xParseDimen(B, 0, iScreenWidth, false);
                h = xParseDimen(C, 0, iScreenHeight, true);
            }
            else
            {
                RECT rc;
                GetClientRect(taskbar->GetBox(), &rc);
                x = xParseCoord(name, 0, rc.right, false);
                y = xParseCoord(A, 0, rc.bottom, true);
                w = xParseDimen(B, 0, rc.right, false);
                h = xParseDimen(C, 0, rc.bottom, true);
            }

            if ( tokens == 6 )
            {
                int step = atoi( D );
                int time = atoi( E );

                taskbar->Reposition(3, x, y, w, h, step, time);
            }
            else
            {
                taskbar->Reposition(3, x, y, w, h);
            }
        }
    }
    else if (tokens == 5 || tokens == 7)
    {
        Taskbar *taskbar = lookupTaskbar(name);

        if(taskbar)
        {
            if ( !taskbar->GetBox() )
            {
                x = xParseCoord(A, 0, iScreenWidth, false);
                y = xParseCoord(B, 0, iScreenHeight, true);
                w = xParseDimen(C, 0, iScreenWidth, false);
                h = xParseDimen(D, 0, iScreenHeight, true);
            }
            else
            {
                RECT rc;
                GetClientRect(taskbar->GetBox(), &rc);
                x = xParseCoord(A, 0, rc.right, false);
                y = xParseCoord(B, 0, rc.bottom, true);
                w = xParseDimen(C, 0, rc.right, false);
                h = xParseDimen(D, 0, rc.bottom, true);
            }

            if ( tokens == 7 )
            {
                int step = atoi( E );
                int time = atoi( F );

                taskbar->Reposition(3, x, y, w, h, step, time);
            }
            else
            {
                taskbar->Reposition(3, x, y, w, h);
            }
        }
    }
}

// !xTaskbarRepositionBy <cx> <cy> <cwidth> <cheight> [<steps>] [<time>]
void BangRepositionBy(HWND hwndCaller, LPCSTR pszArgs)
{
    int x = 0;
    int y = 0;
    int w = 64;
    int h = 64;
    char name[64], A[16], B[16], C[16], D[16], E[16], F[16];
    char *buffers[] = {name, A, B, C, D, E, F};

    int tokens = LCTokenize(pszArgs, buffers, 7, NULL);

    if (tokens == 4 || tokens == 6)
    {
        Taskbar *taskbar = lookupTaskbar();

        if(taskbar)
        {
            x = atoi( name );
            y = atoi( A );
            w = atoi( B );
            h = atoi( C );
            
            if ( tokens == 6 )
            {
                int step = atoi( D );
                int time = atoi( E );

                taskbar->Reposition(3, x, y, w, h, step, time, true);
            }
            else
            {
                taskbar->Reposition(3, x, y, w, h, 0, 0, true);
            }
        }
    }
    else if (tokens == 5 || tokens == 7)
    {
        Taskbar *taskbar = lookupTaskbar(name);

        if(taskbar)
        {
            x = atoi( A );
            y = atoi( B );
            w = atoi( C );
            h = atoi( D );
        
            if ( tokens == 7 )
            {
                int step = atoi( E );
                int time = atoi( F );

                taskbar->Reposition(3, x, y, w, h, step, time, true);
            }
            else
            {
                taskbar->Reposition(3, x, y, w, h, 0, 0, true);
            }
        }
    }
}


// !xTaskbarResize <width> <height> [<steps> <time>]
void BangResize(HWND hwndCaller, LPCSTR pszArgs)
{
    int w = 64;
    int h = 64;
    char name[64], A[16], B[16], C[16], D[16];
    char *buffers[] = {name, A, B, C, D};

    int tokens = LCTokenize(pszArgs, buffers, 5, NULL);
    
    if (tokens == 2 || tokens == 4)
    {
        Taskbar *taskbar = lookupTaskbar();

        if(taskbar)
        {
            if ( !taskbar->GetBox() )
            {
                w = xParseDimen(name, 0, iScreenWidth, false);
                h = xParseDimen(A, 0, iScreenHeight, true);
            }
            else
            {
                RECT rc;
                GetClientRect(taskbar->GetBox(), &rc);
                w = xParseDimen(name, 0, rc.right, false);
                h = xParseDimen(A, 0, rc.bottom, true);
            }

            if ( tokens == 4 )
            {
                int step = atoi( B );
                int time = atoi( C );

                taskbar->Reposition(2, 0, 0, w, h, step, time);
            }
            else
            {
                taskbar->Reposition(2, 0, 0, w, h);
            }
        }
    }
    else if (tokens == 3 || tokens == 5)
    {
        Taskbar *taskbar = lookupTaskbar(name);

        if(taskbar)
        {
            if ( !taskbar->GetBox() )
            {
                w = xParseDimen(A, 0, iScreenWidth, false);
                h = xParseDimen(B, 0, iScreenHeight, true);
            }
            else
            {
                RECT rc;
                GetClientRect(taskbar->GetBox(), &rc);
                w = xParseDimen(A, 0, rc.right, false);
                h = xParseDimen(B, 0, rc.bottom, true);
            }

            if ( tokens == 5 )
            {
                int step = atoi( C );
                int time = atoi( D );

                taskbar->Reposition(2, 0, 0, w, h, step, time);
            }
            else
            {
                taskbar->Reposition(2, 0, 0, w, h);
            }
        }
    }
}

// !xTaskbarResizeBy <cx> <cy> [<steps> <time>]
void BangResizeBy(HWND hwndCaller, LPCSTR pszArgs)
{
    int w = 64;
    int h = 64;
    char name[64], A[16], B[16], C[16], D[16];
    char *buffers[] = {name, A, B, C, D};

    int tokens = LCTokenize(pszArgs, buffers, 5, NULL);
    
    if (tokens == 2 || tokens == 4)
    {
        Taskbar *taskbar = lookupTaskbar();

        if(taskbar)
        {
            w = atoi( name );
            h = atoi( A );
            
            if ( tokens == 4 )
            {
                int step = atoi( B );
                int time = atoi( C );

                taskbar->Reposition(2, 0, 0, w, h, step, time, true);
            }
            else
            {
                taskbar->Reposition(2, 0, 0, w, h, 0, 0, true);
            }
        }
    }
    else if (tokens == 3 || tokens == 5)
    {
        Taskbar *taskbar = lookupTaskbar(name);

        if(taskbar)
        {
            w = atoi( A );
            h = atoi( B );
        
            if ( tokens == 5 )
            {
                int step = atoi( C );
                int time = atoi( D );

                taskbar->Reposition(2, 0, 0, w, h, step, time, true);
            }
            else
            {
                taskbar->Reposition(2, 0, 0, w, h, 0, 0, true);
            }
        }
    }
}

// !xTaskbarScrollButtons <taskbarname> next/prev [amount]
void BangScrollButtons(HWND hwndCaller, LPCSTR pszArgs)
{
    char name[32], direction[32], amount[16];
    char *buffers[] = {name, direction, amount};

    int numTokens = LCTokenize(pszArgs, buffers, 3, NULL);

    if ( numTokens == 1 )
    {
        Taskbar *taskbar = lookupTaskbar();

        if(taskbar)
        {
            if ( _stricmp(name, "next") == NULL )
            {
                taskbar->ScrollButtons(true, 1);
            }
            else 
            {
                taskbar->ScrollButtons(false, 1);
            }
        }
    }
    else if (numTokens == 2 || numTokens == 3)
    {
        Taskbar *taskbar = lookupTaskbar(name);

        if(taskbar)
        {
            int delta = 1;
            if ( numTokens == 3 )
                delta = atoi(amount);

            if ( delta < 1 )
                delta = 1;

            if ( _stricmp(direction, "next") == NULL )
            {
                taskbar->ScrollButtons(true, delta);
            }
            else 
            {
                taskbar->ScrollButtons(false, delta);
            }
        }
        else
        {
            Taskbar *taskbar = lookupTaskbar();

            if(taskbar)
            {
                int delta = 1;
                if ( numTokens == 2 )
                    delta = atoi(direction);

                if ( delta < 1 )
                    delta = 1;

                if ( _stricmp(name, "next") == NULL )
                {
                    taskbar->ScrollButtons(true, delta);
                }
                else 
                {
                    taskbar->ScrollButtons(false, delta);
                }
            }
        }
    }
}

// !xTaskbarLSBoxHook <taskbarname>
void BangLSBoxHook(HWND hwndCaller, LPCSTR pszArgs)
{
    char name[64] = {0};
    char rest[MAX_LINE_LENGTH] = {0};
    char *buffers[] = {name};

    LCTokenize(pszArgs, buffers, 1, rest);

    const char *handle = strrchr(pszArgs,' ');

    HWND hwndBox = (HWND)atoi(handle+1);

    string tmpname = name;

    if ( !Getx2RCBool(tmpname.c_str(), "LoadInBox") )
    {
        MessageBox(NULL, ("You need to specify the TaskbarName after !xTaskbarLsBoxHook,\nand you must set \n\"" + tmpname + "LoadInBox\",\nif you want to hook this xTaskbar in another Module!").c_str(), "User Information", MB_OK | MB_SETFOREGROUND | MB_ICONEXCLAMATION);
        return;
    }

    Taskbar *taskbar = lookupTaskbar(tmpname);

    if (taskbar == NULL)
    {
        taskbar = new Taskbar(name);
        
        taskbarList.insert(taskbarList.end(), taskbar);
                
        taskbar->Load(hInstance, hwndBox);
    }
    else
    {
        MessageBox(NULL, ("This Taskbar already exists!\nUse the setting\n\"" + tmpname + "LoadInBox\",\nif you want to hook this xTaskbar in another Module!").c_str(), "User Information", MB_OK | MB_SETFOREGROUND | MB_ICONEXCLAMATION);
    }
}

// !xTaskbarSetAlpha <value>
void BangSetAlpha(HWND hwndCaller, LPCSTR pszArgs)
{   
    char name[64], value[16], step[16], delay[16];
    char *buffers[] = {name, value, step, delay};

    int tokens = LCTokenize(pszArgs, buffers, 4, NULL);

    if (tokens == 1 || tokens == 3)
    {
        Taskbar *taskbar = lookupTaskbar();
        
        if(taskbar)
        {   
            if ( tokens == 1 )
            {       
                taskbar->SetAlpha(atoi( name ));    
            }
            else if ( tokens == 3 )
            {
                taskbar->SetAlphaFade(atoi( name ), atoi( value ), atoi( step ));
            }
        }
    }
    else if (tokens == 2 || tokens == 4)
    {
        Taskbar *taskbar = lookupTaskbar(name);
        
        if(taskbar)
        {
            if ( tokens == 2 )
            {       
                taskbar->SetAlpha(atoi( value ));   
            }
            else if ( tokens == 4 )
            {
                taskbar->SetAlphaFade(atoi( value ), atoi( step ), atoi( delay ));
            }
        }
    }

}

// !xTaskbarSwitch <next|prev>
void BangSwitch(HWND hwndCaller, LPCSTR pszArgs)
{
    char name[64], value[32];
    char *buffers[] = {name, value};

    int tokens = LCTokenize(pszArgs, buffers, 2, NULL);

    if (tokens == 1)
    {
        Taskbar *taskbar = lookupTaskbar();

        if(taskbar)
        {
            taskbar->SwitchToButton( name );
        }
    }
    else if (tokens == 2)
    {
        Taskbar *taskbar = lookupTaskbar(name);

        if(taskbar)
        {
            taskbar->SwitchToButton( value );
        }
    }
}

// !xTaskbarDisplay <value>
void BangDisplay(HWND hwndCaller, LPCSTR pszArgs)
{
    const NameValuePair displayValues[] = {
            { "all", -1 },
            { "normal+active+minimized", -1 },
            { "normal", 0 },
            { "active", 1 },
            { "minimized", 2 },
            { "active+normal", 3 },
            { "normal+active", 3 },
            { "minimized+normal", 4 },
            { "normal+minimized", 4 },
            { "active+minimized", 5 },
            { "minimized+active", 5 },
            { "flashing", 6},
            { 0, 0 }
            };

    const NameValuePair *current = displayValues;

    int display = -1;

    char name[64], value[64];
    char *buffers[] = {name, value};

    int tokens = LCTokenize(pszArgs, buffers, 2, NULL);

    if (tokens == 1)
    {
        while(current->name)
        {
            if(_stricmp(current->name, name) == 0)
            {
                display = current->value;
                break;
            }

            current++;
        }

        Taskbar *taskbar = lookupTaskbar();
        
        if(taskbar)
        {
            taskbar->SetDisplayButtonType( display );
            taskbar->CreateLayout();
        }
    }
    else if (tokens == 2)
    {
        while(current->name)
        {
            if( _stricmp(current->name, value) == 0)
            {
                display = current->value;
                break;
            }

            current++;
        }

        Taskbar *taskbar = lookupTaskbar(name);
        
        if(taskbar)
        {
            taskbar->SetDisplayButtonType( display );
            taskbar->CreateLayout();
        }
    }
}

// !xTaskbarRefresh <setting> <newsettings>
void BangRefresh(HWND hwndCaller, LPCSTR pszArgs)
{
    char name[MAX_LINE_LENGTH], setting[MAX_LINE_LENGTH], value[MAX_LINE_LENGTH];
    char *buffers[] = {name, setting};

    LCTokenize(pszArgs, buffers, 2, value);

    Taskbar *taskbar = lookupTaskbar(name);

    if (taskbar)
    {
        if ( taskbar->IsLayoutLocked() )
            return;

        PrefixedLSSetVariable( (string)setting, "", (string)value );
        taskbar->Reconfigure(true);
    }
    else
    {
        char setting1[MAX_LINE_LENGTH], value1[MAX_LINE_LENGTH];
        char *buffers1[] = {setting1};

        LCTokenize(pszArgs, buffers1, 1, value1);

        taskbar = lookupTaskbar();

        if (taskbar)
        {
            if ( taskbar->IsLayoutLocked() )
                return;

            PrefixedLSSetVariable( (string)setting1, "", (string)value1 );
            taskbar->Reconfigure(true);
        }
    }
}

void BangMinimize(HWND hwndCaller, LPCSTR pszArgs)
{
    EnumWindows(EnumWindowsProc, (LPARAM) -1);
}

void BangRestore(HWND hwndCaller, LPCSTR pszArgs)
{
    EnumWindows(EnumWindowsProc, (LPARAM) 1);
}

void BangCascade(HWND hwndCaller, LPCSTR pszArgs)
{
    CascadeWindows(NULL, MDITILE_SKIPDISABLED, NULL, 0, NULL);
}

void BangTileWindows(HWND hwndCaller, LPCSTR pszArgs)
{
    TileWindows(NULL, MDITILE_SKIPDISABLED | MDITILE_HORIZONTAL, NULL, 0, NULL);
}

